
import java.io.*;
import java.util.*;
import java.util.zip.*;
import junit.framework.*;

public class AmbiguousTigerParser_Test extends TestCase {
    public static void main(String[] args) {
        junit.textui.TestRunner.main(new String[] { AmbiguousTigerParser_Test.class.getName() });
    }
    
    public AmbiguousTigerParser_Test(String name) {
        super(name);
    }
    
    private static final Set SKIPED_NAMES = new HashSet(Arrays.asList(new String[] {
        //"ORBUtilSystemException.java",
    }));
    private static void parse(AmbiguousTigerParser parser, String fullname, String shortname, Reader reader) throws IOException, AmbiguousTigerParser.ParseException {
        if (SKIPED_NAMES.contains(shortname)) {
            System.out.println(fullname + " -- skipped.");
        } else if (!shortname.endsWith(".java")) {
            boolean parsed = false;
            try {
                parser.parseCompilationUnit(fullname, reader, 8);
                parsed = true;
            } catch (AmbiguousTigerParser.ParseException x) {
            }
            if (parsed) {
                System.out.println("Error: " + fullname);
                throw new RuntimeException("test failed.");
            }
        } else {
            boolean parsed = false;
            try {
                parser.parseCompilationUnit(fullname, reader, 8);
                parsed = true;
            //} catch (Exception x) {
            //    x.printStackTrace();
            } finally {
                if (!parsed)
                    System.out.println("Error: " + fullname);
            }
        }
    }
    
    private static void parse(AmbiguousTigerParser parser, File file) throws IOException, AmbiguousTigerParser.ParseException {
        if (file.isDirectory()) {
            //System.out.println(file);
            File[] files = file.listFiles();
            for (int i = 0; i < files.length; i++) {
                parse(parser, files[i]);
            }
        } else {
            Reader reader = new InputStreamReader(new BufferedInputStream(new FileInputStream(file)), "ISO8859_1");
            parse(parser, file.getPath(), file.getName(), reader);
            reader.close();
        }
    }
    
    public void test() throws IOException, AmbiguousTigerParser.ParseException {
        AmbiguousTigerParser parser = new AmbiguousTigerParser(true, true);
        
        String jdkHome = System.getProperty("JAVA_HOME");
        assertTrue("set JAVA_HOME", jdkHome != null);
        File file = new File(jdkHome, "src.zip");
        assertTrue(file.getName() + " is not found in " + jdkHome, file.exists());
        System.out.println(file);
        //System.out.println("quick.test=" + System.getProperty("quick.test", "not defined"));
        boolean quickTest = "true".equals(System.getProperty("quick.test", "false"));
        ZipFile zip = new ZipFile(file);
        int count = 0;
        Enumeration e = zip.entries();
        while (e.hasMoreElements()) {
            ZipEntry entry = (ZipEntry) e.nextElement();
            count++;
            if (!entry.isDirectory()) {
                Reader reader = new InputStreamReader(new BufferedInputStream(zip.getInputStream(entry)), "ISO8859_1");
                String name = entry.getName();
                parse(parser, name, name.substring(name.lastIndexOf('/') + 1), reader);
                reader.close();
            }
            if (quickTest) {
                if (count > 100)
                    break;
            }
        }
        
        /*
        String buildFile = System.getProperty("ant.file");
        file = new File(new File(new File(buildFile).getParentFile(), "tmp"), "src");
        System.out.println(file);
        parse(parser, file);
        
        file = new File(new File(new File(buildFile).getParentFile(), "_private"), "examples");
        if (file.exists()) {
            parser = new AmbiguousTigerParser(true, true);
            System.out.println(file);
            parse(parser, file);
        }
        */
        
    }
}
