
/* 
 * TODO:
 *  EnumDeclaration is not implemented.
 * 
 * 
 */

import java.util.*;
import java.io.*;
import parser.*;
import parser.Syntax.*;
import parser.Parser.*;

public class JavaMerger {
    private static final boolean REPLACE_ENHANCED_FOR = false;
    private static final boolean REPLACE_ENHANCED_IF = true;
    private static final boolean SUPPLY_ASSERTION_MESSAGE = true;
    private static final boolean SUPPORT_ONCE_METHOD = true;
    private static final boolean SUPPORT_GETTEE_FIELD = true;
    
    public static void main(String[] args) {
        try {
            if (args.length == 0) {
                System.out.println("usage: JavaMerger files...");
                return;
            }
            
            Writer writer = new OutputStreamWriter(System.out);
            mergeFiles(asList(args), null, writer, false);
            writer.close();
        } catch (Exception x) {
            x.printStackTrace();
        }
    }
    
    private static List asList(Object[] array) {
        return Arrays.asList(array);
    }
    private static List asList(List list) {
        return list;
    }
    
    private static class EditableCompilationUnit extends Default.Node implements CompilationUnit {
        public EditableCompilationUnit() {
            super(new LinkedList());
        }
    }
    public static void mergeFiles(Collection sources, String encoding, Writer writer, boolean fileNameOutput) throws IOException, ParseException {
        final PrintWriter original = (writer instanceof PrintWriter) ? (PrintWriter) writer : new PrintWriter(writer);
        writer = new Writer() {
            boolean pendingCR = false;
            public void write(int c) throws IOException {
                if (pendingCR) {
                    pendingCR = false;
                    if (c == '\n') {
                        // CR LF --> println
                        original.println();
                        return;
                    } else {
                        // CR not-LF --> println not-LF
                        original.println();
                    }
                }
                if (c == '\r')
                    pendingCR = true;
                else
                    original.write(c);
            }
            public void write(char[] cbuf, int off, int len) throws IOException {
                for (int i = 0; i < len; i++)
                    this.write(cbuf[off + i]);
            }
            public void flush() throws IOException {
                if (pendingCR)
                    original.println();
                original.flush();
            }
            public void close() throws IOException {
                this.flush();
                original.close();
            }
        };
        Parser parser = new Parser(true, true);
        CompilationUnit unit = new EditableCompilationUnit();
        Iterator it = sources.iterator();
        while (it.hasNext()) {
            Object source = (Object) it.next();
            File f = (source instanceof File) ? (File) source : new File((String) source);
            if (fileNameOutput)
                System.out.println("    " + f);
            CompilationUnit unit2 = parser.parseCompilationUnit(f, encoding);
            merge(unit, unit2);
        }
        new JavaMerger().output(unit, writer);
    }
    
    private static void merge(Node node1, Node node2) {
        // merge node2 into node1
        if (node1 instanceof CompilationUnit) {
            // merge PackageDeclaration
            merge(new ClassChecker(PackageDeclaration.class), new ClassChecker(Node.class), new IndexFinder() {
                public int getIndex(Node node) {
                    int result = 0;
                    Iterator it = node.getChildNodes().iterator();
                    for (int i = 0; it.hasNext(); i++) {
                        Node child = (Node) it.next();
                        if (child instanceof PackageDeclaration)
                            result = i + 1;
                    }
                    return result;
                }
            }, node1, node2);
            
            // merge import declarations
            IndexFinder finder = new IndexFinder() {
                public int getIndex(Node node) {
                    int result = 0;
                    Iterator it = node.getChildNodes().iterator();
                    for (int i = 0; it.hasNext(); i++) {
                        Node child = (Node) it.next();
                        if (child instanceof PackageDeclaration || child instanceof ImportDeclaration)
                            result = i + 1;
                    }
                    return result;
                }
            };
            merge(new ClassChecker(SingleTypeImportDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(TypeImportOnDemandDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(SingleStaticImportDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            merge(new ClassChecker(StaticImportOnDemandDeclaration.class), new ClassChecker(Node.class), finder, node1, node2);
            
            // merge type declaration
            merge(new ClassChecker(TypeDeclaration.class), new ClassChecker(Identifier.class), null, node1, node2);
        } else if (node1 instanceof TypeDeclaration) {
            // TODO: class と interface をマージしようとしてないかチェック
            // merge modifiers
            merge(new ClassChecker(Token.class), null, null, find(node1, Modifiers.class), find(node2, Modifiers.class));
            
            // merge inheritance
            merge(find(node1, Extendees.class), find(node2, Extendees.class));
            merge(find(node1, Implementees.class), find(node2, Implementees.class));
            
            // merge nested type declarations
            merge(new ClassChecker(TypeDeclaration.class), new ClassChecker(Identifier.class), null, find(node1, TypeDeclarationBody.class), find(node2, TypeDeclarationBody.class));
            
            // merge the other members
            merge(new Checker() {
                public boolean hit(Node node) {
                    return !(node instanceof TypeDeclaration);
                }
            }, new Checker() {
                public boolean hit(Node node) {
                    return false;
                }
            }, null, find(node1, TypeDeclarationBody.class), find(node2, TypeDeclarationBody.class));
        } else if (node1 instanceof Extendees) {
            Extendees n1 = (Extendees) node1;
            Extendees n2 = (Extendees) node2;
            Iterator it = asList(n2.extendees()).iterator();
            while (it.hasNext()) {
                ClassOrInterfaceType extendee = (ClassOrInterfaceType) it.next();
                if (n1.getChildNodes().isEmpty()) {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "extends"));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                } else {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ","));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                }
                n1.getChildNodes().add(extendee);
                n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
            }
        } else if (node1 instanceof Implementees) {
            Implementees n1 = (Implementees) node1;
            Implementees n2 = (Implementees) node2;
            Iterator it = asList(n2.implementees()).iterator();
            while (it.hasNext()) {
                ClassOrInterfaceType implementee = (ClassOrInterfaceType) it.next();
                if (n1.getChildNodes().isEmpty()) {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "implements"));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                } else {
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ","));
                    n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
                }
                n1.getChildNodes().add(implementee);
                n1.getChildNodes().add(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "));
            }
        }
    }
    private static Node find(Node node, Class clazz) {
        Iterator it = node.getChildNodes().iterator();
        while (it.hasNext()) {
            Node child = (Node) it.next();
            if (clazz.isInstance(child))
                return child;
        }
        return null;
    }
    
    private static interface Checker {
        public boolean hit(Node node);
    }
    private static class ClassChecker implements Checker {
        private Class clazz;
        public ClassChecker(Class clazz) {
            this.clazz = clazz;
        }
        public boolean hit(Node node) {
            return clazz.isInstance(node);
        }
    }
    private static interface IndexFinder {
        public int getIndex(Node node);
    }
    
    private static void merge(Checker targetChecker, Checker identifierChecker, IndexFinder finder, Node node1, Node node2) {
        Map idToNode = new LinkedHashMap();
        {
            Iterator it = node1.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                if (targetChecker.hit(child)) {
                    StringBuffer id = new StringBuffer();
                    if (identifierChecker == null) {
                        id.append(toEssentialString(child));
                    } else {
                        Iterator it4 = child.getChildNodes().iterator();
                        while (it4.hasNext()) {
                            Node grandchild = (Node) it4.next();
                            if (identifierChecker.hit(grandchild))
                                id.append(toEssentialString(grandchild));
                        }
                    }
                    idToNode.put(id.toString(), child);
                }
            }
        }
        
        {
            Token lastJavadoc = null;
            Iterator it = node2.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                if (child instanceof Token && ((Token) child).isWhite()) {
                    Token token = (Token) child;
                    if (token.getSymbolID() == Syntax.TOKEN_JAVADOC_COMMENT) {
                        lastJavadoc = token;
                    }
                } else if (!targetChecker.hit(child)) {
                    lastJavadoc = null;
                } else {
                    StringBuffer id = new StringBuffer();
                    if (identifierChecker == null) {
                        id.append(toEssentialString(child));
                    } else {
                        Iterator it2 = child.getChildNodes().iterator();
                        while (it2.hasNext()) {
                            Node grandchild = (Node) it2.next();
                            if (identifierChecker.hit(grandchild))
                                id.append(toEssentialString(grandchild));
                        }
                    }
                    Node earlierVisitor = (id.length() == 0) ? null : (Node) idToNode.get(id.toString());
                    if (earlierVisitor == null) {
                        int index = node1.getChildNodes().size();
                        if (finder != null) {
                            index = finder.getIndex(node1);
                        }
                        Object[] list;
                        if (lastJavadoc != null)
                            list = new Object[] { lastJavadoc, child, new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n") };
                        else
                            list = new Object[] { new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "), child, new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n") };
                        node1.getChildNodes().addAll(index, asList(list));
                        if (id.length() != 0)
                            idToNode.put(id.toString(), child);
                    } else {
                        merge(earlierVisitor, child);
                    }
                }
            }
        }
    }
    
    private static String toEssentialString(Node node) {
        StringBuffer buffer = new StringBuffer();
        toEssentialString(node, buffer);
        return buffer.toString();
    }
    private static void toEssentialString(Node node, StringBuffer buffer) {
        if (node instanceof Token) {
            Token token = (Token) node;
            if (!token.isWhite())
                buffer.append(((Token) node).getOriginalImage());
            return;
        }
        
        Iterator it = node.getChildNodes().iterator();
        while (it.hasNext()) {
            Node child = (Node) it.next();
            toEssentialString(child, buffer);
        }
    }
    
    private boolean outputPosition = true;
    private void output(Node node, Writer writer) throws IOException {
        writer.write("/*\n");   // */
        outputPosition = true;
        outputNode(node, writer);
        if (outputPosition) {
            writer.write("*/");
        }
        writer.write('\n');
        writer.flush();
    }
    private void outputNode(Node node, Writer writer) throws IOException {
        if (node instanceof Token) {
            Token token = (Token) node;
            if (outputPosition) {
                try {
                    writer.write(token.getSourceName());
                    writer.write(':');
                    writer.write(Integer.toString(token.getLine()));
                    writer.write(':');
                } catch (UnsupportedOperationException x) {
                }
                writer.write("*/");
                outputPosition = false;
            }
            if (token.getSymbolID() == Parser.TOKEN_LINE_TERMINATOR) {
                writer.write("/*\n");   // */
                outputPosition = true;
            } else if (token.getSymbolID() == Parser.TOKEN_END_OF_LINE_COMMENT) {
                writer.write(token.getOriginalImage());
                writer.write("\n/*\n");   // */
                outputPosition = true;
            } else {
                writer.write(token.getOriginalImage());
            }
            return;
        }
        
        if (REPLACE_ENHANCED_FOR) {
            if (node instanceof EnhancedForStatement) {
                outputNode((EnhancedForStatement) node, writer);
                return;
            }
        }
        
        if (REPLACE_ENHANCED_IF) {
            if (node instanceof EnhancedIfThenElseStatement) {
                replaceEnhancedIf((EnhancedIfThenElseStatement) node, writer);
            } else if (node instanceof EnhancedTypeComparison) {
                outputNode((EnhancedTypeComparison) node, writer);
                return;
            }
        }
        
        if (SUPPLY_ASSERTION_MESSAGE) {
            if (node instanceof AssertStatement) {
                AssertStatement as = (AssertStatement) node;
                if (as.message() == null) {
                    outputExpandedAssetion(as, writer);
                    return;
                }
            }
        }
        
        if (SUPPORT_ONCE_METHOD) {
            if (node instanceof MethodDeclaration) {
                MethodDeclaration md = (MethodDeclaration) node;
                if (md.once() != null) {
                    outputExpandedOnceMethod(md, writer);
                    return;
                }
            }
        }
        
        if (SUPPORT_GETTEE_FIELD) {
            if (node instanceof FieldDeclaration) {
                FieldDeclaration fd = (FieldDeclaration) node;
                if (fd.gettee() != null) {
                    Iterator it = fd.getChildNodes().iterator();
                    while (it.hasNext()) {
                        Node child = (Node) it.next();
                        if (child != fd.gettee())
                            outputNode(child, writer);
                    }
                    outputGetterMethod(fd, writer);
                    return;
                }
            }
        }
        
        Iterator it = node.getChildNodes().iterator();
        while (it.hasNext()) {
            Node child = (Node) it.next();
            outputNode(child, writer);
        }
        
    }
    private int counter = 0;
    private void outputNode(EnhancedForStatement node, Writer writer) throws IOException {
        counter++;
        outputNode(node.forKeyword(), writer);
        outputNode(node.leftParen(), writer);
        {
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "java.util.Iterator $$i" + counter + "$$ = ("), writer);
            outputNode(node.expression(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ").iterator(); $$i" + counter + "$$.hasNext();"), writer);
        }
        outputNode(node.rightParen(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "{"), writer);
        
        if (node.modifier() != null) {
            outputNode(node.modifier(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " "), writer);
        }
        outputNode(node.type(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " "), writer);
        outputNode(node.identifier(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " = ("), writer);
        outputNode(node.type(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ") $$i" + counter + "$$.next();"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        
        outputNode(node.statement(), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "}"), writer);
    }
    
    private void replaceEnhancedIf(EnhancedIfThenElseStatement node, Writer writer) throws IOException {
        Map simpleNameToCastType = new HashMap();
        Expression condition = node.expression();
        while (condition instanceof EnhancedConditionalAndExpression) {
            EnhancedConditionalAndExpression c = (EnhancedConditionalAndExpression) condition;
            EnhancedTypeComparison etc = (EnhancedTypeComparison) c.operand1();
            simpleNameToCastType.put(((SimpleExpressionName) etc.operand()).identifier().getImage(), etc.type());
            condition = c.operand2();
        }
        if (condition instanceof EnhancedTypeComparison) {
            EnhancedTypeComparison etc = (EnhancedTypeComparison) condition;
            simpleNameToCastType.put(((SimpleExpressionName) etc.operand()).identifier().getImage(), etc.type());
        } else {
            replaceEnhancedIfCast(simpleNameToCastType, condition);
        }
        replaceEnhancedIfCast(simpleNameToCastType, node.statementThen());
    }
    private void replaceEnhancedIfCast(Map simpleNameToCastType, Node node) {
        Iterator it0 = node.getChildNodes().iterator();
        while (it0.hasNext()) {
            Node child = (Node) it0.next();
            if (child instanceof SimpleExpressionName || child instanceof SimpleAmbiguousName) {
                SimpleName sn = (SimpleName) child;
                Node type = (Node) simpleNameToCastType.get(sn.identifier().getImage());
                if (type != null) {
                    List list = new LinkedList();
                    list.add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "("));
                    list.add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "("));
                    list.add(type);
                    list.add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ")"));
                    child.getChildNodes().addAll(0, list);
                    child.getChildNodes().add(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ")"));
                    continue;
                }
            }
            replaceEnhancedIfCast(simpleNameToCastType, child);
        }
    }
    private void outputNode(EnhancedTypeComparison node, Writer writer) throws IOException {
        Iterator it = node.getChildNodes().iterator();
        while (it.hasNext()) {
            Node child = (Node) it.next();
            if (child instanceof Token && ((Token) child).getImage().equals(":")) {
                Token token = (Token) child;
                outputNode(new Default.Token(token.getSymbolID(), "instanceof", "instanceof", token), writer);
            } else {
                outputNode(child, writer);
            }
        }
    }
    
    private void outputExpandedAssetion(AssertStatement node, Writer writer) throws IOException {
        outputNode(node.assertKeyword(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " "), writer);
        outputNode(node.expression(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " : \"[\""), writer);
        
        outputAssetionExpression(node.expression(), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " + \"] is not satisfied.\";"), writer);
    }
    private void outputAssetionExpression(Node node, Writer writer) throws IOException {
        if (node instanceof ExpressionName) {
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " + "), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "("), writer);
            outputBare(node, writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ")"), writer);
        } else if (node instanceof Token) {
            Token token = (Token) node;
            StringBuffer buffer = new StringBuffer();
            buffer.append('"');
            String image = token.getOriginalImage();
            for (int i = 0; i < image.length(); i++) {
                char ch = image.charAt(i);
                if (Character.isWhitespace(ch)) {
                    buffer.append(" ");
                } else {
                    switch (ch) {
                        default:
                            buffer.append(ch);
                            break;
                        case '\\':
                            buffer.append("\\");
                            break;
                        case '"':
                            buffer.append("\"");
                            break;
                    }
                }
            }
            buffer.append('"');
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " + "), writer);
            outputNode(new Default.Token(Syntax.TOKEN_STRING, buffer.toString()), writer);
        } else {
            Iterator it = node.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                outputAssetionExpression(child, writer);
            }
        }
    }
    
    private void outputExpandedOnceMethod(MethodDeclaration node, Writer writer) throws IOException {
        counter++;
        
        {
            Iterator it = node.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                if (child != node.once() && child != node.body()) {
                    outputNode(child, writer);
                }
            }
        }
        
        outputNode((Node) node.body().getChildNodes().get(0), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "java.util.List $$key" + counter + "$$ = java.util.Arrays.asList(new java.lang.Object[] { "), writer);
        {
            Iterator it = asList(node.formalParameters()).iterator();
            while (it.hasNext()) {
                FormalParameter param = (FormalParameter) it.next();
                outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, param.identifier().getImage() + ", "), writer);
            }
        }
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "});"), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "if ($$memo" + counter + "$$.containsKey($$key" + counter + "$$)) return ("), writer);
        outputNode(node.resultType(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ") $$memo" + counter + "$$.get($$key" + counter + "$$);"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        outputNode(node.resultType(), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " $$result" + counter + "$$ = $$impl" + counter + "$$("), writer);
        {
            Iterator it = asList(node.formalParameters()).iterator();
            while (it.hasNext()) {
                FormalParameter param = (FormalParameter) it.next();
                outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, param.identifier().getImage()), writer);
                if (it.hasNext())
                    outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ", "), writer);
            }
        }
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ");"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "$$memo" + counter + "$$.put($$key" + counter + "$$, $$result" + counter + "$$);"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "return $$result" + counter + "$$;"), writer);
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        
        outputNode((Node) node.body().getChildNodes().get(node.body().getChildNodes().size() - 1), writer);
        
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        if (node.resultType() instanceof Syntax.PrimitiveType) {
            String type = ((Syntax.Token) node.resultType().getChildNodes().get(0)).getImage();
            if (type.equals("int")) {
                type = "java.lang.Integer";
            } else if (type.equals("char")) {
                type = "java.lang.Character";
            } else {
                type = "java.lang." + Character.toUpperCase(type.charAt(0)) + type.substring(1);
            }
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "private java.util.Map<java.util.List, " + type + "> $$memo" + counter + "$$ = new java.util.HashMap<java.util.List, " + type + ">();"), writer);
        } else {
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "private java.util.Map<java.util.List, "), writer);
            outputNode(node.resultType(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "> $$memo" + counter + "$$ = new java.util.HashMap<java.util.List, "), writer);
            outputNode(node.resultType(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, ">();"), writer);
        }
        outputNode(new Default.Token(Syntax.TOKEN_LINE_TERMINATOR, "\n"), writer);
        
        {
            Iterator it = node.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                if (child == node.identifier()) {
                    outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "$$impl" + counter + "$$"), writer);
                } else if (child != node.once()) {
                    outputNode(child, writer);
                }
            }
        }
    }
    
    private void outputGetterMethod(FieldDeclaration node, Writer writer) throws IOException {
        counter++;
        
        Iterator it = asList(node.variableDeclarators()).iterator();
        while (it.hasNext()) {
            VariableDeclarator decl = (VariableDeclarator) it.next();
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, " public "), writer);
            outputNode(node.type(), writer);
            String image = decl.identifier().getImage();
            image = "get" + Character.toUpperCase(image.charAt(0)) + image.substring(1);
            outputNode(new Default.Token(Syntax.TOKEN_WHITE_SPACES, " "), writer);
            outputNode(new Default.Token(Syntax.TOKEN_IDENTIFIER, image), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "() { return this."), writer);
            outputNode(decl.identifier(), writer);
            outputNode(new Default.Token(Syntax.TOKEN_MULTIPURPOSE, "; }"), writer);
        }
    }
    
    private void outputBare(Node node, Writer writer) throws IOException {
        if (node instanceof Token) {
            Token token = (Token) node;
            outputNode(token, writer);
        } else {
            Iterator it = node.getChildNodes().iterator();
            while (it.hasNext()) {
                Node child = (Node) it.next();
                outputBare(child, writer);
            }
        }
    }
}

