
package jp.gr.java_conf.koto.notavacc.types;

import java.util.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;
import jp.gr.java_conf.koto.notavacc.lrg.*;
import jp.gr.java_conf.koto.notavacc.types.*;
import jp.gr.java_conf.koto.notavacc.parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Original.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.Root;

public class TypeSystemGenerator extends Module {
    private final boolean arrayRatherThanList;
    private final boolean tiger;
    public TypeSystemGenerator(Environment env, boolean tiger, boolean arrayRatherThanList) {
        super(env);
        this.arrayRatherThanList = arrayRatherThanList;
        this.tiger = tiger;
    }
    
    private Root root;
    public TypeSystem generate(Root root, Set usedTypeNonterminals) {
        this.root = root;
        
        logger.verbose(environment.processingFile, "generating types.");
        
        // nameToType
        
        Map definitionToType = new LinkedHashMap();
        
        // create objects.
        Iterator it0 = root.typeDefinitions().iterator();
        while (it0.hasNext()) {
            TypeDefinition definition = (TypeDefinition) it0.next();
            String name = root.getCanonicalName(definition);
            TypeNonterminal nonterminal = new TypeNonterminal(name);
            if (!usedTypeNonterminals.contains(nonterminal))
                nonterminal = null;
            GeneratedNodeType type = new GeneratedNodeType(definition.identifier(), name, nonterminal);
            type.isProtected = (definition.protectedKeyword() != null);
            type.isPrivate = (definition.privateKeyword() != null);
            type.isProtectedParsable = (definition.protectedParsableKeyword() != null);
            type.isAbstract = (definition.abstractKeyword() != null);
            type.superTypes = new LinkedHashSet();
            definitionToType.put(definition, type);
        }
        
        Iterator it5 = root.absorberDefinitions2().iterator();
        while (it5.hasNext()) {
            AbsorberDefinition2 definition = (AbsorberDefinition2) it5.next();
            String name = root.getCanonicalName(definition);
            GeneratedAbsorberType type = new GeneratedAbsorberType(definition.identifier(), name, definition);
            type.isProtected = (definition.protectedKeyword() != null);
            type.isPrivate = (definition.privateKeyword() != null);
            type.isAbstract = (definition.abstractKeyword() != null);
            type.superTypes = new LinkedHashSet();
            definitionToType.put(definition, type);
        }
        
        // build inheritance tree.
        Iterator it1 = definitionToType.entrySet().iterator();
        while (it1.hasNext()) {
            Map.Entry entry = (Map.Entry) it1.next();
            TypeGenerateDefinition definition = (TypeGenerateDefinition) entry.getKey();
            Type type = (Type) entry.getValue();
            //System.out.println("# " + type.getName() + " " + type.superTypes);
            
            Set superTypes;
            if (definition instanceof TypeDefinition) {
                TypeDefinition def = (TypeDefinition) definition;
                superTypes = ((GeneratedNodeType) type).superTypes;
                if (def.superTypes().isEmpty()) {
                    superTypes.add(NodeRootType.INSTANCE);
                    continue;
                }
            } else {
                AbsorberDefinition2 def = (AbsorberDefinition2) definition;
                superTypes = ((GeneratedAbsorberType) type).superTypes;
                
                boolean absorberFound = false;
                Iterator it28 = def.superTypes().iterator();
                while (it28.hasNext()) {
                    AbsorberOrTypeName superName = (AbsorberOrTypeName) it28.next();
                    TypeGenerateDefinition superDefinition = root.getDefinition(superName);
                    if (superDefinition instanceof AbsorberDefinition2) {
                        absorberFound = true;
                        break;
                    }
                }
                if (!absorberFound) {
                    superTypes.add(AbsorberRootType.INSTANCE);
                }
            }
            
            Iterator it2 = definition.superTypes().iterator();
            while (it2.hasNext()) {
                AbsorberOrTypeName superName = (AbsorberOrTypeName) it2.next();
                TypeGenerateDefinition superDefinition = root.getDefinition(superName);
                if (getSuperTypeDefinitions(superDefinition).contains(definition)) {
                    logger.error("{0} should not extend {1} extending {2}.", definition.identifier(), superName.identifier(), definition.identifier().getImage());
                } else {
                    superTypes.add(definitionToType.get(superDefinition));
                }
            }
        }
        if (logger.hasError())
            logger.fatal();
        
        
        // bind fields
        Iterator it3 = definitionToType.entrySet().iterator();
        while (it3.hasNext()) {
            Map.Entry entry = (Map.Entry) it3.next();
            TypeGenerateDefinition definition = (TypeGenerateDefinition) entry.getKey();
            Type type = (Type) entry.getValue();
            if (definition instanceof TypeDefinition) {
                bindFields((TypeDefinition) definition, definitionToType);
            } else {
                bindFields((AbsorberDefinition2) definition, definitionToType);
            }
        }
        
        // build result object.
        Map nameToType = new LinkedHashMap();
        Iterator it27 = definitionToType.values().iterator();
        while (it27.hasNext()) {
            UserDefinedType type = (UserDefinedType) it27.next();
            nameToType.put(type.getName(), type);
        }
        
        logger.verbose(environment.processingFile, "generated.  {0} types.", new Integer(nameToType.size()));
        
        return new GeneratedTypeSystem(nameToType);
    }
    private Set getSuperTypeDefinitions(TypeGenerateDefinition definition) {
        Set result = new LinkedHashSet();
        getSuperTypeDefinitions(result, definition);
        return result;
    }
    private void getSuperTypeDefinitions(Set result, TypeGenerateDefinition definition) {
        if (result.add(definition)) {
            Iterator it2 = definition.superTypes().iterator();
            while (it2.hasNext()) {
                AbsorberOrTypeName name = (AbsorberOrTypeName) it2.next();
                TypeGenerateDefinition superDefinition = root.getDefinition(name);
                getSuperTypeDefinitions(result, superDefinition);
            }
        }
    }
    
    private static class GeneratedNodeType extends UserDefinedNodeType {
        public final Token hint;
        
        public GeneratedNodeType(Token hint, String name, TypeNonterminal typeNonterminal) {
            this.hint = hint;
            this.name = name;
            this.typeNonterminal = typeNonterminal;
        }
        
        private final String name;
        public String getName() {
            return name;
        }
        
        private final TypeNonterminal typeNonterminal;
        public TypeNonterminal getTypeNonterminal() {
            return typeNonterminal;
        }
        
        private Set superTypes = null;
        public Set getDirectSuperTypes() {
            return superTypes;
        }
        public boolean isAbstract;
        public boolean isAbstract() {
            return isAbstract;
        }
        public boolean isProtected;
        public boolean isProtected() {
            return isProtected;
        }
        public boolean isPrivate;
        public boolean isPrivate() {
            return isPrivate;
        }
        public boolean isProtectedParsable;
        public boolean isProtectedParsable() {
            return isProtectedParsable;
        }
        private Set declaredFields = null;
        public Set getDeclaredFields() {
            return declaredFields;
        }
        private Map nameToField = null;
        public Map getMapNameToField() {
            return nameToField;
        }
    }
    
    private static class GeneratedAbsorberType extends UserDefinedAbsorberType {
        public final Token hint;
        
        public GeneratedAbsorberType(Token hint, String name, Parser.AbsorberDefinition2 absorberDefinition) {
            this.hint = hint;
            this.name = name;
            //this.absorberDefinition = absorberDefinition;
        }
        
        private final String name;
        public String getName() {
            return name;
        }
        
        //private final Parser.AbsorberDefinition2 absorberDefinition;
        //public Parser.AbsorberDefinition2 getAbsorberDefinition() {
        //    return absorberDefinition;
        //}
        
        private Set superTypes = null;
        public Set getDirectSuperTypes() {
            return superTypes;
        }
        public boolean isAbstract;
        public boolean isAbstract() {
            return isAbstract;
        }
        public boolean isProtected;
        public boolean isProtected() {
            return isProtected;
        }
        public boolean isPrivate;
        public boolean isPrivate() {
            return isPrivate;
        }
        private Set declaredFields = null;
        public Set getDeclaredFields() {
            return declaredFields;
        }
        private Map nameToField = null;
        public Map getMapNameToField() {
            return nameToField;
        }
    }
    
    private static class GeneratedField extends Field {
        public final Token hint;
        public GeneratedField(Type owner, String name, Type type, Token hint) {
            super(owner, name, type);
            this.hint = hint;
        }
    }
    
    private static class GeneratedTypeSystem implements TypeSystem {
        private Map nameToType;
        public GeneratedTypeSystem(Map nameToType) {
            this.nameToType = nameToType;
        }
        public Map getMapNameToType() {
            return nameToType;
        }
    }
    
    private Set reportedError = new HashSet();
    private UserDefinedType bindFields(TypeGenerateDefinition definition, final Map definitionToType) {
        if (definition instanceof TypeDefinition) {
            return bindFields((TypeDefinition) definition, definitionToType);
        } else {
            return bindFields((AbsorberDefinition2) definition, definitionToType);
        }
    }
    private UserDefinedNodeType bindFields(TypeDefinition definition, final Map definitionToType) {
        GeneratedNodeType target = (GeneratedNodeType) definitionToType.get(definition);
        if (target.declaredFields != null) {
            // already bound.
            return target;
        }
        
        // bind all supertypes before the process.
        Iterator it2 = definition.superTypes().iterator();
        while (it2.hasNext()) {
            TypeName name = (TypeName) it2.next();
            TypeDefinition superDefinition = root.getDefinition(name);
            bindFields(superDefinition, definitionToType);
        }
        
        // find labels in the right hand side.
        Map nameToHint = new LinkedHashMap();
        {
            new ExpressionVisitor(root, 1) {
                protected void visit(Context context, LabeledExpression expr, Object argument) {
                    Map nameToHint = (Map) argument;
                    String label = expr.label().getImage();
                    if (!nameToHint.containsKey(label))
                        nameToHint.put(label, expr.label());
                    super.visit(context, expr, argument);
                }
            }.visitAll(definition.expression(), nameToHint);
        }
        
        MultiMap nameToTypes = new MultiMap();
        {
            new ExpressionVisitor(root, 1) {
                protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
                    MultiMap nameToTypes = (MultiMap) argument;
                    
                    Set types = new LinkedHashSet();
                    
                    GeneratedNodeType type = (GeneratedNodeType) definitionToType.get(def);
                    types.add(type);
                    
                    Iterator it20 = context.typeRestrictors.iterator();
                    while (it20.hasNext()) {
                        TypeName typeRestrictor = (TypeName) it20.next();
                        TypeDefinition restrictorDefinition = root.getDefinition(typeRestrictor);
                        GeneratedNodeType restrictor = (GeneratedNodeType) definitionToType.get(restrictorDefinition);
                        if (restrictor.isSubTypeOf(type)) {
                            if (reportedError.add(Arrays.asList(new Object[] { typeRestrictor, expr }))) {
                                logger.error("The restrictor {0} should be a supertype of the restricted type {1}.", typeRestrictor.identifier(), expr.name().identifier());
                            }
                        }
                        
                        types.add(restrictor);
                    }
                    
                    Iterator it8 = context.labelStrings.iterator();
                    while (it8.hasNext()) {
                        String label = (String) it8.next();
                        nameToTypes.addAll(label, types);
                    }
                }
                protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
                    addToken(context, argument);
                }
                protected void visit(Context context, StringExpression expr, Object argument) {
                    addToken(context, argument);
                }
                protected void addToken(Context context, Object argument) {
                    MultiMap nameToTypes = (MultiMap) argument;
                    Iterator it24 = context.labelStrings.iterator();
                    while (it24.hasNext()) {
                        String label = (String) it24.next();
                        nameToTypes.add(label, TokenType.INSTANCE);
                    }
                }
            }.visitAll(definition.expression(), nameToTypes);
        }
        
        Map nameToCount = new LinkedHashMap();
        {
            new ExpressionVisitor(root, 2) {
                protected void visit(Context context, SelectiveExpression expr, Object argument) {
                    Map nameToCount = (Map) argument;
                    
                    Iterator it0 = expr.operands().iterator();
                    while (it0.hasNext()) {
                        Expression operand = (Expression) it0.next();
                        Map tmp = new LinkedHashMap();
                        visitAll(context, operand, tmp);
                        Iterator it23 = tmp.entrySet().iterator();
                        while (it23.hasNext()) {
                            Map.Entry entry = (Map.Entry) it23.next();
                            String label = (String) entry.getKey();
                            
                            int count1 = countOf((Integer) entry.getValue());
                            int count2 = countOf((Integer) nameToCount.get(label));
                            nameToCount.put(label, new Integer(Math.max(count1, count2)));
                        }
                    }
                }
                protected void visit(Context context, SequentialExpression expr, Object argument) {
                    Map nameToCount = (Map) argument;
                    
                    Iterator it0 = expr.operands().iterator();
                    while (it0.hasNext()) {
                        Expression operand = (Expression) it0.next();
                        Map tmp = new LinkedHashMap();
                        visitAll(context, operand, tmp);
                        Iterator it23 = tmp.entrySet().iterator();
                        while (it23.hasNext()) {
                            Map.Entry entry = (Map.Entry) it23.next();
                            String label = (String) entry.getKey();
                            
                            int count1 = countOf((Integer) entry.getValue());
                            int count2 = countOf((Integer) nameToCount.get(label));
                            nameToCount.put(label, new Integer(Math.min(2, count1 + count2)));
                        }
                    }
                }
                protected void visit(Context context, PlusExpression expr, Object argument) {
                    visitAll(context, expr.operand(), argument);
                    Map nameToCount = (Map) argument;
                    Iterator it22 = nameToCount.entrySet().iterator();
                    while (it22.hasNext()) {
                        Map.Entry entry = (Map.Entry) it22.next();
                        int count = countOf((Integer) entry.getValue());
                        count = Math.min(2, count * 2);
                        entry.setValue(new Integer(count));
                    }
                }
                
                protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
                    countUp(context.labelStrings, argument);
                }
                protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
                    countUp(context.labelStrings, argument);
                }
                protected void visit(Context context, StringExpression expr, Object argument) {
                    countUp(context.labelStrings, argument);
                }
                private void countUp(Collection labels, Object argument) {
                    Map nameToCount = (Map) argument;
                    Iterator it21 = labels.iterator();
                    while (it21.hasNext()) {
                        String label = (String) it21.next();
                        int count = countOf((Integer) nameToCount.get(label));
                        count = Math.min(2, count + 1);
                        nameToCount.put(label, new Integer(count));
                    }
                }
                private int countOf(Integer value) {
                    if (value == null)
                        return 0;
                    return value.intValue();
                }
            }.visitAll(definition.expression(), nameToCount);
        }
        
        
        // find labels in the supertypes.
        MultiMap nameToInheritingTypes = new MultiMap();
        MultiMap nameToInheritingFields = new MultiMap();
        Iterator it9 = target.superTypes.iterator();
        while (it9.hasNext()) {
            Type superType = (Type) it9.next();
            Iterator it10 = superType.getMapNameToField().values().iterator();
            while (it10.hasNext()) {
                GeneratedField field = (GeneratedField) it10.next();
                String name = field.getName();
                nameToInheritingFields.add(name, field);
                Type type = field.getType();
                if (type instanceof ArrayType) {
                    ArrayType arrayType = (ArrayType) type;
                    type = arrayType.getComponentType();
                    assert !(type instanceof ArrayType);
                    if (nameToTypes.containsKey(name)) {
                        nameToCount.put(name, new Integer(2));  // make it an array.
                    }
                } else {
                    Integer integer = (Integer) nameToCount.get(name);
                    if (integer != null && integer.intValue() > 1) {
                        // error: multiple children may be labeld in the type,
                        //        but the supertype(s) has a method for a single child.
                        Token hint = (Token) nameToHint.get(name);
                        Object superTypeHint;
                        if (superType instanceof GeneratedNodeType) {
                            GeneratedNodeType gt = (GeneratedNodeType) superType;
                            superTypeHint = gt.hint;
                        } else {
                            assert false;   // the only user-defined type has a field.
                            superTypeHint = "`" + superType + "'";
                        }
                        logger.error("The label {0} in {1}, overriding {2} in {3}, can not accept more than one node.", hint, target.hint, field.hint, superTypeHint);
                    }
                }
                nameToInheritingTypes.add(name, type);
            }
        }
        
        
        // fix the field types.
        target.declaredFields = new LinkedHashSet();
        
        assert nameToHint.keySet().containsAll(nameToTypes.keySet());
        Iterator it25 = nameToHint.keySet().iterator();
        while (it25.hasNext()) {
            String name = (String) it25.next();
            Set types = (Set) nameToTypes.get(name);
            if (types == null)
                continue;
            assert !types.isEmpty();
            
            final boolean debug = false;
            if (debug)
                System.out.println("* " + target.getName() + " " + name + "  : " + types);
            
            Integer countInteger = (Integer) nameToCount.get(name);
            // subclassable
            //          1.4-    1.5+
            // single   NG      OK
            // array    NG      OK
            // list     OK*     OK
            // * treat as subclassable, but outputted as List.
            boolean subclassable = tiger || (!arrayRatherThanList && (countInteger != null && countInteger.intValue() > 1));
            //the case of list with 1.5+ = NG
            //boolean subclassable = arrayRatherThanList ? tiger : (tiger ^ (countInteger != null && countInteger.intValue() > 1));
            
            // find the most specific type.
            do {
                if (debug)
                    System.out.println("types " + types);
                
                Set commonSuperTypes = null;
                Iterator it14 = types.iterator();
                while (it14.hasNext()) {
                    Type type = (Type) it14.next();
                    Set superTypes = new LinkedHashSet(type.getSuperTypes());
                    superTypes.add(type);
                    if (commonSuperTypes == null) {
                        commonSuperTypes = superTypes;
                    } else {
                        commonSuperTypes.retainAll(superTypes);
                    }
                }
                assert !commonSuperTypes.isEmpty();
                types = commonSuperTypes;
                
                if (debug)
                    System.out.println("common types " + types);
                
                Set restricters = nameToInheritingTypes.getValueSet(name);
                if (restricters != null) {
                    Iterator it12 = restricters.iterator();
                    while (it12.hasNext()) {
                        Type restricter = (Type) it12.next();
                        Iterator it15 = types.iterator();
                        while (it15.hasNext()) {
                            Type type = (Type) it15.next();
                            if (subclassable) {
                                if (!type.equals(restricter) && !type.isSubTypeOf(restricter))
                                    it15.remove();
                            } else {
                                if (!type.equals(restricter))
                                    it15.remove();
                            }
                        }
                    }
                }
                
                if (debug)
                    System.out.println("restricted " + types);
                if (types.isEmpty())
                    break;
                
                Set removableSuperTypes = new LinkedHashSet();
                Iterator it13 = types.iterator();
                while (it13.hasNext()) {
                    Type type = (Type) it13.next();
                    removableSuperTypes.addAll(type.getSuperTypes());
                }
                types.removeAll(removableSuperTypes);
                if (debug)
                    System.out.println("removed " + types);
            } while (types.size() > 1);
            
            if (types.isEmpty()) {
                // error: cannot select.
                Token hint = (Token) nameToHint.get(name);
                logger.error("The label {0} in {1} is", hint, target.hint);
                Iterator it16 = nameToInheritingFields.getValueSet(name).iterator();
                while (it16.hasNext()) {
                    GeneratedField field = (GeneratedField) it16.next();
                    Type type = field.getType();
                    if (type instanceof ArrayType)
                        type = ((ArrayType) type).getComponentType();
                    Object ownerTypeHint;
                    if (field.getOwner() instanceof GeneratedNodeType) {
                        GeneratedNodeType gt = (GeneratedNodeType) field.getOwner();
                        ownerTypeHint = gt.hint;
                    } else {
                        assert false;   // the only user-defined type has a field.
                        ownerTypeHint = field.getOwner();
                    }
                    if (it16.hasNext()) {
                        logger.error("        overriding {0} in {1} so should accept `{2}'',", field.hint, ownerTypeHint, type);
                    } else {
                        logger.error("        overriding {0} in {1} so should accept `{2}''.", field.hint, ownerTypeHint, type);
                    }
                }
                target.declaredFields.add(new GeneratedField(target, name, NodeRootType.INSTANCE, hint));
            } else {
                Type type = (Type) types.iterator().next();
                Token hint = (Token) nameToHint.get(name);
                Integer integer = (Integer) nameToCount.get(name);
                if (integer != null && integer.intValue() > 1)
                    target.declaredFields.add(new GeneratedField(target, name, new ArrayType(type), hint));
                else
                    target.declaredFields.add(new GeneratedField(target, name, type, hint));
            }
        }
        
        
        target.nameToField = new LinkedHashMap();
        
        Iterator it19 = target.declaredFields.iterator();
        while (it19.hasNext()) {
            GeneratedField field = (GeneratedField) it19.next();
            target.nameToField.put(field.getName(), field);
        }
        
        Iterator it17 = nameToInheritingTypes.entrySet().iterator();
        while (it17.hasNext()) {
            Map.Entry entry = (Map.Entry) it17.next();
            String name = (String) entry.getKey();
            Set types = (Set) entry.getValue();
            if (!target.nameToField.containsKey(name)) {
                // not declared explicitly, so inherit types.
                if (types.size() > 1) {
                    logger.error("{0} can not inherit conflicted labels `{1}''.", target.hint, name);
                }
                GeneratedField field = (GeneratedField) nameToInheritingFields.getValueSet(name).iterator().next();
                target.nameToField.put(name, field);
            }
        }
        
        return target;
    }
    
    private UserDefinedAbsorberType bindFields(AbsorberDefinition2 definition, Map definitionToType) {
        GeneratedAbsorberType target = (GeneratedAbsorberType) definitionToType.get(definition);
        if (target.declaredFields != null) {
            // already bound.
            return target;
        }
        
        target.declaredFields = new LinkedHashSet();
        Iterator it11 = definition.elements().iterator();
        while (it11.hasNext()) {
            AbsorberElement element = (AbsorberElement) it11.next();
            Type type = (UserDefinedType) definitionToType.get(root.getDefinition(element.typeName()));
            if (element.plus() != null)
                type = new ArrayType(type);
            target.declaredFields.add(new GeneratedField(target, element.label().getImage(), type, element.label()));
        }
        
        target.nameToField = new LinkedHashMap();
        Iterator it18 = definition.superTypes().iterator();
        while (it18.hasNext()) {
            AbsorberOrTypeName superName = (AbsorberOrTypeName) it18.next();
            TypeGenerateDefinition superDef = root.getDefinition(superName);
            UserDefinedType superType = bindFields(superDef, definitionToType);
            
            target.nameToField.putAll(superType.getMapNameToField());
            // TODO: コンフリクトをチェックする。
        }
        
        Iterator it19 = target.declaredFields.iterator();
        while (it19.hasNext()) {
            GeneratedField field = (GeneratedField) it19.next();
            
            target.nameToField.put(field.getName(), field);
            // TODO: コンフリクトをチェックする。
        }
        
        return target;
    }
    
}

