
// TODO:
// 現状は、シフト時に $absorber 起源の nonterminal であれば、まとめる。
// これを、type 起源の nonterminal の場合と、常に1つしかヒットしない alias の場合でも行い、
// absorber 化する。


package jp.gr.java_conf.koto.notavacc.generator;

import java.util.*;
import java.io.*;
import java.math.*;
import jp.gr.java_conf.koto.io.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;
import jp.gr.java_conf.koto.notavacc.dfa.*;
import jp.gr.java_conf.koto.notavacc.dfa.DeterministicFiniteAutomaton.*;
import jp.gr.java_conf.koto.notavacc.lrg.*;
import jp.gr.java_conf.koto.notavacc.lrg.LRTable.*;
import jp.gr.java_conf.koto.notavacc.types.*;
import jp.gr.java_conf.koto.notavacc.parser.Original.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.Root;

public class JavaCodeGenerator extends CodeGenerator {
    private final boolean arrayRatherThanList;
    
    private Root root;
    private DeterministicFiniteAutomaton dfa;
    private LRTable lrTable;
    private double targetVersion;
    private boolean separateOutput;
    private String manualLocale;
    private boolean javadocForManual;
    public JavaCodeGenerator(Environment env, Root root, DeterministicFiniteAutomaton dfa, LRTable lrTable, TypeSystem typeSystem, String baseName, double targetVersion, boolean arrayRatherThanList) throws IOException {
        super(env, root, dfa, lrTable, typeSystem, getOutputFiles(baseName, env.separateOutput));
        this.arrayRatherThanList = arrayRatherThanList;
        this.root = root;
        this.dfa = dfa;
        this.lrTable = lrTable;
        this.targetVersion = targetVersion;
        this.separateOutput = env.separateOutput;
        this.manualLocale = env.manualLocale;
        if (this.manualLocale == null) {
            this.manualLocale = "en";
        }
        this.javadocForManual = env.outputForManual;
        
        if (separateOutput)
            storePrinted = new List[] { new LinkedList(), new LinkedList(), };
        else
            storePrinted = new List[] { new LinkedList(), };
    }
    private static File[] getOutputFiles(String baseName, boolean separateOutput) {
        File file1 = new File(baseName + ".java");
        if (!separateOutput) {
            return new File[] { file1 };
        }
        
        File f = new File(baseName);
        String name = f.getName();
        int index = name.indexOf('.');
        if (index > 0)
            name = name.substring(0, index);
        File file2 = new File(f.getParent(), name + "Tables.java");
        
        return new File[] { file1, file2 };
    }
    
    
    /*private static String toJavaConstantNameStyle(String name) {
        char prev = '_';
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < name.length(); i++) {
            char ch = name.charAt(i);
            
            if (Character.isUpperCase(ch)) {
                if (Character.isLowerCase(prev)) {
                    result.append('_');             // e.g. FooURL to FOO_URL
                } else if (Character.isUpperCase(prev)) {
                    if (i + 1 < name.length()) {
                        char nextChar = name.charAt(i + 1);
                        if (Character.isLowerCase(nextChar)) {
                            result.append('_');     // e.g. URLBar to URL_BAR
                        }
                    }
                }
            }
            
            result.append(Character.toUpperCase(ch));
            prev = ch;
        }
        return result.toString();
    }*/
    
    private static String encodeForHTML(Object object) {
        return encodeForHTML(object.toString());
    }
    private static String encodeForHTML(String image) {
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < image.length(); i++) {
            char ch = image.charAt(i);
            switch (ch) {
            default:
                if (ch <= ' ' || ch >= 0x7F) {
                    result.append("&#" + ((int) ch) + ";");
                } else {
                    result.append(ch);
                }
                break;
            case '<':
                result.append("&lt;");
                break;
            case '>':
                result.append("&gt;");
                break;
            case '&':
                result.append("&amp;");
                break;
            case '"':
                result.append("&quot;");
                break;
            }
        }
        return result.toString();
    }
    private static String toHexCode(char ch) {
        String hex = "000" + Integer.toString(ch, 16).toUpperCase(Locale.ENGLISH);
        hex = hex.substring(hex.length() - 4);
        return hex;
    }
    
    private static String quoted(char ch) {
        StringBuffer quoted = new StringBuffer("\'");
        quoted(quoted, ch);
        quoted.append("\'");
        return quoted.toString();
    }
    private static String quoted(String str) {
        return quoted(str, true);
    }
    private static String quoted(String str, boolean quote) {
        StringBuffer quoted = new StringBuffer();
        if (quote)
            quoted.append("\"");
        for (int i = 0; i < str.length(); i++) {
            char ch = str.charAt(i);
            quoted(quoted, ch);
        }
        if (quote)
            quoted.append("\"");
        return quoted.toString();
    }
    private static void quoted(StringBuffer result, char ch) {
        switch (ch) {
        default:
            if (0x0020 <= ch && ch <= 0x007E) {
                result.append(ch);
            } else {
                result.append("\\u").append(toHexCode(ch));
            }
            break;
        case '\n':   result.append("\\n");      break;
        case '\t':   result.append("\\t");      break;
        case '\b':   result.append("\\b");      break;
        case '\r':   result.append("\\r");      break;
        case '\f':   result.append("\\f");      break;
        case '\\':   result.append("\\\\");     break;
        case '\'':   result.append("\\\'");     break;
        case '\"':   result.append("\\\"");     break;
        }
    }
    private static String quotedAsComment(String str) {
        for (;;) {
            int index = str.indexOf("*/");
            if (index < 0)
                break;
            str = str.substring(0, index + 1) + " " + str.substring(index + 1);
        }
        return "/*" + str + "*/";
    }
    
    
    private Map replacements = new HashMap();
    private void addReplacement(String word, String replacement) {
        if (replacements.containsKey(word))
            throw new IllegalArgumentException(word);
        replacements.put(word, replacement /*+ "_unique"*/);
    }
    private String getReplacement(String word) {
        return (String) replacements.get(word);
    }
    private String replace(String str) {
        StringBuffer result = new StringBuffer();
        int index = 0;
        int mode = -1;  // -1 for initial state, 0 for identifiers, 1 for the other tokens.
        for (int i = 0;; i++) {
            char ch;
            if (i < str.length())
                ch = str.charAt(i);
            else
                ch = '\uFFFF';
            boolean identifierPart = Character.isJavaIdentifierPart(ch);
            int nextMode = identifierPart ? 0 : 1;
            if (mode != -1 && (ch == '\uFFFF' || nextMode != mode || (mode == 0 && ch == '$'))) {
                String sub = str.substring(index, i);
                if (mode == 0) {
                    String newSequence = (String) replacements.get(sub);
                    if (newSequence != null)
                        sub = newSequence;
                }
                result.append(sub);
                index = i;
            }
            mode = nextMode;
            if (i >= str.length())
                break;
        }
        
        return result.toString();
    }
    private String replaceTypeParameter(String str) {
        StringBuffer result = null;
        int start = 0;
        for (;;) {
            int index = str.indexOf("$<", start);
            if (index < 0)
                break;
            if (result == null)
                result = new StringBuffer();
            result.append(str.substring(start, index));
            index++;
            int level = 0;
            while (index < str.length()) {
                char ch = str.charAt(index++);
                if (ch == '<') {
                    level++;
                } else if (ch == '>') {
                    level--;
                }
                if (targetVersion >= 1.5)
                    result.append(ch);
                if (level == 0)
                    break;
            }
            start = index;
        }
        if (result == null)
            return str;
        result.append(str.substring(start));
        return result.toString();
    }
    
    
    private List[] storePrinted;
    private static class Line {
        public String sourceFileName;
        public int lineNumber;
        public String line;
        public Line(String sourceFileName, int lineNumber, String line) {
            this.sourceFileName = sourceFileName;
            this.lineNumber = lineNumber;
            this.line = line;
        }
    }
    private int storeChannel = 0;
    private void changeStoreChannel(int storeChannel) {
        this.storeChannel = storeChannel;
    }
    private void println(String line) {
        println(null, -1, line);
    }
    private void println(String sourceFileName, int lineNumber, String line) {
        storePrinted[storeChannel].add(new Line(sourceFileName, lineNumber, line));
    }
    
    private void flush() throws IOException {
        for (int i = 0; i < storePrinted.length; i++) {
            Formalizer formalizer = new Formalizer(writers[i]);
            Iterator it = storePrinted[i].iterator();
            while (it.hasNext()) {
                Line line = (Line) it.next();
                it.remove();
                
                line.line = replace(line.line);
                line.line = replaceTypeParameter(line.line);
                formalizer.outputLine(line.sourceFileName, line.lineNumber, line.line);
            }
        }
    }
    
    static class Formalizer {
        private final SafePrintWriter writer;
        public Formalizer(SafePrintWriter writer) {
            this.writer = writer;
        }
        
        private int indent = 0;
        private int temporalIndent = -1;
        private boolean inCStyleComment = false;
        public void outputLine(String sourceFileName, int lineNumber, String line) throws IOException {
            line = line.trim();
            
            boolean headInCStyleComment = inCStyleComment;
            boolean tailInLineComment = false;
            {
                char inLiteral = '\0';
            loop:
                for (int i = 0; i < line.length(); i++) {
                    if (inCStyleComment) {
                        if (line.startsWith("*/", i)) {
                            inCStyleComment = false;
                            i++;
                        }
                    } else if (inLiteral != '\0') {
                        char ch = line.charAt(i);
                        if (ch == inLiteral) {
                            inLiteral = '\0';
                        } else if (ch == '\\') {
                            i++;
                        }
                    } else {
                        char ch = line.charAt(i);
                        switch (ch) {
                        case '\'':
                        case '\"':
                            inLiteral = ch;
                            break;
                        case '/':
                            if (line.startsWith("*", i + 1)) {
                                inCStyleComment = true;
                                i++;
                            } else if (line.startsWith("/", i + 1)) {
                                tailInLineComment = true;
                                break loop;
                            }
                            break;
                        }
                    }
                }
            }
            boolean tailInCStyleComment = inCStyleComment;
            boolean tailInComment = tailInCStyleComment || tailInLineComment;
            String indentSpaces = "    ";
            
            if (headInCStyleComment) {
                for (int i = 0; i < indent; i++)
                    writer.print(indentSpaces);
                if (line.startsWith("*"))
                    writer.print(" ");
                writer.println(line);
            } else {
                if (line.startsWith("}"))
                    indent--;
                if (temporalIndent == -1)
                    temporalIndent = indent;
                
                if (line.startsWith("case ") || line.startsWith("default:") || line.startsWith("default "))
                    temporalIndent--;
                
                for (int i = 0; i < temporalIndent; i++)
                    writer.print(indentSpaces);
                int count = temporalIndent * indentSpaces.length();
                writer.print(line);
                count += line.length();
                if (!tailInCStyleComment) {
                    int max = Math.max(80, count + 4);
                    for (int i = 0; i + count < max; i++) {
                        writer.print(" ");
                    }
                    writer.print("// (" + sourceFileName + ":" + lineNumber + ")");
                }
                writer.println();
                
                temporalIndent = -1;
                if (!tailInComment && line.endsWith("{")) {
                    indent++;
                } else {
                    if (line.startsWith("if ") || line.startsWith("while ") || line.startsWith("do ") || line.startsWith("for ")
                                || line.startsWith("if(") || line.startsWith("while(") || line.startsWith("do(") || line.startsWith("for(")
                                || line.equals("else") || line.startsWith("else "))
                        temporalIndent = indent + 1;
                }
            }
        }
    }
    
    protected void generateImpl(Tables tables) throws IOException {
        generateCode(tables);
        flush();
    }
    
    private Map typeToName = new HashMap();
    private String getTypeName(Type type) {
        String result = (String) typeToName.get(type);
        if (result == null) {
            if (type instanceof ArrayType) {
                if (arrayRatherThanList) {
                    ArrayType arrayType = (ArrayType) type;
                    result = getTypeName(arrayType.getComponentType()) + "[]";
                } else {
                    if (targetVersion >= 1.5) {
                        ArrayType arrayType = (ArrayType) type;
                        result = "java.util.List<? extends " + getTypeName(arrayType.getComponentType()) + ">";
                    } else {
                        result = "java.util.List";
                    }
                }
            } else if (type instanceof NodeRootType) {
                result = "$Parser.$Node";
            } else if (type instanceof TokenType) {
                result = "$Parser.$Token";
            } else if (type instanceof ObjectType) {
                result = "java.lang.Object";
            } else if (type instanceof UserDefinedNodeType) {
                UserDefinedNodeType userDefinedType = (UserDefinedNodeType) type;
                result = "$Parser." + userDefinedType.getName();
            } else if (type instanceof AbsorberRootType) {
                result = "$Parser.$AmbiguityAbsorber";
            } else if (type instanceof UserDefinedAbsorberType) {
                UserDefinedAbsorberType userDefinedType = (UserDefinedAbsorberType) type;
                result = "$Parser." + userDefinedType.getName();
            } else {
                throw new AssertionError();
            }
            typeToName.put(type, result);
        }
        return result;
    }
    
    private Map typeToSimpleName = new HashMap();
    private String getSimpleTypeName(Type type) {
        String result = (String) typeToSimpleName.get(type);
        if (result == null) {
            if (type instanceof ArrayType) {
                if (arrayRatherThanList) {
                    ArrayType arrayType = (ArrayType) type;
                    result = getSimpleTypeName(arrayType.getComponentType()) + "[]";
                } else {
                    if (targetVersion >= 1.5) {
                        ArrayType arrayType = (ArrayType) type;
                        result = "java.util.List<? extends " + getSimpleTypeName(arrayType.getComponentType()) + ">";
                    } else {
                        result = "java.util.List";
                    }
                }
            } else if (type instanceof NodeRootType) {
                result = "Node";
            } else if (type instanceof TokenType) {
                result = "Token";
            } else if (type instanceof ObjectType) {
                result = "java.lang.Object";
            } else {
                UserDefinedNodeType userDefinedType = (UserDefinedNodeType) type;
                result = userDefinedType.getName();
            }
            typeToSimpleName.put(type, result);
        }
        return result;
    }
    
    private static final Object RESERVED_NAME = new Object();
    private void reserve(String name, MultiMap namespace) {
        namespace.add(name, RESERVED_NAME);
        addReplacement("$" + name, name);
    }
    private void allocate(String name, MultiMap namespace) {
        String alocatedName = "__" + name + "__notavacc_reserved";
        namespace.add(alocatedName, RESERVED_NAME);
        addReplacement("$" + name, alocatedName);
    }
    
    private void generateCode(Tables tables) throws IOException {
        logger.verbose(environment.processingFile, "{0} EBNF productions.", new Integer(tables.typeNonterminals().size() + tables.aliasNonterminals().size()));
        
        MultiMap topTypeNamespace = new MultiMap();
        MultiMap topFieldNamespace = new MultiMap();
        MultiMap topMethodNamespace = new MultiMap();
        MultiMap nodeFieldNamespace = new MultiMap();
        MultiMap nodeNoArgumentMethodNamespace = new MultiMap();
        MultiMap nodeMethodNamespace = new MultiMap();
        
        final boolean multipleShiftEntry = false;
        final boolean visitorSupport = true;
        final boolean ambiguitySupport = true;  //root.supportsAbsorber();
        final boolean resultTypeNarrowing = targetVersion >= 1.5;
        final boolean useMyList = targetVersion >= 1.5 && !arrayRatherThanList;
        final boolean whiteTokensMostLeft = true;
        
        String packageName = root.getPackageName();
        String topLevelClassName = root.getParserSimpleName();
        
        addReplacement("$packageName", packageName);
        addReplacement("$Parser", topLevelClassName);
        topTypeNamespace.add(topLevelClassName, topLevelClassName);
        topTypeNamespace.add("java", RESERVED_NAME);
        topFieldNamespace.add("java", RESERVED_NAME);
        topMethodNamespace.add("java", RESERVED_NAME);
        nodeFieldNamespace.add("java", RESERVED_NAME);
        nodeNoArgumentMethodNamespace.add("java", RESERVED_NAME);
        nodeMethodNamespace.add("java", RESERVED_NAME);
        if (targetVersion >= 1.5) {
            addReplacement("$StringBuffer", "java.lang.StringBuilder");
        } else {
            addReplacement("$StringBuffer", "java.lang.StringBuffer");
        }
        if (targetVersion >= 1.4) {
            addReplacement("$CharSequence", "java.lang.CharSequence");
            addReplacement("$AssertionError", "java.lang.AssertionError");
            addReplacement("$subSequence", "subSequence");
            addReplacement("$final_avoid_jdk12_bug", "final");
            addReplacement("$protected_avoid_jdk12_bug", "protected");
            addReplacement("$assert", "assert");
            addReplacement("$LinkedHashSet", "java.util.LinkedHashSet");
            addReplacement("$LinkedHashMap", "java.util.LinkedHashMap");
            //addReplacement("$LinkedHashSet", "java.util.HashSet");
            //addReplacement("$LinkedHashMap", "java.util.HashMap");
        } else {
            addReplacement("$CharSequence", "java.lang.StringBuffer");
            addReplacement("$AssertionError", "java.lang.RuntimeException");
            addReplacement("$subSequence", "substring");
            addReplacement("$final_avoid_jdk12_bug", " ");
            addReplacement("$protected_avoid_jdk12_bug", "public");
            addReplacement("$assert", "; // ");
            addReplacement("$LinkedHashSet", "java.util.HashSet");
            addReplacement("$LinkedHashMap", "java.util.HashMap");
        }
        
        final int labelWordCount = tables.labelWordCount();
        
        if (javadocForManual) {
            addReplacement("$Type", "Type");
            addReplacement("$parseType", "parseType");
            addReplacement("$label", "label");
        }
        
]]<<
        #Set ambiguousWrapperTypes = new LinkedHashSet();
        
        #if (packageName != null)
            package $packageName;
        
        #if (manualLocale.equals("ja")) {
            /**
             * ￢＜＞＜∪∪ (notavaCC) >>[[ encodeForHTML(Version.VERSION) ]]<< によって出力された構文表現です。通常 >>[[ encodeForHTML(Version.LICENSE) ]]<< にしたがってライセンスされます。
            #if (javadocForManual) {
                 * <p>
                 * このクラスの完全限定名は、<i><a href="../input.html#syntaxref:Root:ParserDeclaration" class="symbol">ParserDeclaration</a></i>で指定されます。このクラスは、<i><a href="../input.html#syntaxref:Root:ParserDeclaration" class="symbol">ParserDeclaration</a></i>が<code>$protected</code>と指定されていればデフォルトスコープ（パッケージスコープ）、そうでなければ<code>public</code>になります。
                 * <tt>--target</tt> オプションの値が 1.4 より小さい場合、{@link java.lang.CharSequence} の代わりに {@link java.lang.StringBuffer} が、{@link java.lang.AssertionError} の代わりに {@link java.lang.RuntimeException} が使用されます。
            #}
             * <p>
             * メソッド
            #{
                #List parsableTypes = new ArrayList(tables.userDefinedNodeTypes().size());
                #Iterator it1 = tables.userDefinedNodeTypes().iterator();
                #while (it1.hasNext()) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) it1.next();
                    #LRState initialState = (LRState) lrTable.symbolToInitialState().get(type.getTypeNonterminal());
                    #if (initialState != null) {
                        #parsableTypes.add(type);
                    #}
                #}
                #for (int i = 0; i < parsableTypes.size(); i++) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) parsableTypes.get(i);
                    #if (i == parsableTypes.size() - 1) {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<}
                    #} else if (i == parsableTypes.size() - 2) {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<} および
                    #} else {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<}、
                    #}
                #}
            #}
             * はテキストを構文解析し、抽象構文木を出力します。
             * ツリーのノードは、{@link $Node}のサブタイプのインスタンスになります。
             * <p>
             * このクラス、およびネストしたクラスは、<strong>同期化されません</strong>。
             * 
            #if (javadocForManual) {
                * @since    notavaCC 1.0
            #}
             */
        #} else {
            /**
             * A syntax representation, generated by &#xAC;&lt;&gt;&lt;&#x222A;&#x222A; (notavaCC) >>[[ encodeForHTML(Version.VERSION) ]]<<.  Licensed under >>[[ encodeForHTML(Version.LICENSE) ]]<< usually.
            #if (javadocForManual) {
                 * <p>
                 * The fully qualified name of the class is given by <i><a href="../input.html#syntaxref:Root:ParserDeclaration" class="symbol">ParserDeclaration</a></i>.  The class has the default scope if it is described as <code>$protected</code>.  Otherwise, the class is public.
                 * If the value of the <tt>--target</tt> option is less than 1.4, {@link java.lang.CharSequence} and {@link java.lang.AssertionError} are replaced by {@link java.lang.StringBuffer} and {@link java.lang.RuntimeException}.
            #}
             * <p>
            #{
                #List parsableTypes = new ArrayList(tables.userDefinedNodeTypes().size());
                #Iterator it1 = tables.userDefinedNodeTypes().iterator();
                #while (it1.hasNext()) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) it1.next();
                    #LRState initialState = (LRState) lrTable.symbolToInitialState().get(type.getTypeNonterminal());
                    #if (initialState != null) {
                        #parsableTypes.add(type);
                    #}
                #}
                #for (int i = 0; i < parsableTypes.size(); i++) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) parsableTypes.get(i);
                    #if (i == parsableTypes.size() - 1) {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<}
                    #} else if (i == parsableTypes.size() - 2) {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<} and
                    #} else {
                         * {@link $Parser#parse>>[[ type.getName() ]]<<},
                    #}
                #}
                #if (parsableTypes.size() == 1) {
                     * parses a text and generates an abstruct syntax tree.
                #} else if (parsableTypes.size() > 1) {
                     * parse a text and generates an abstruct syntax tree.
                #}
            #}
             * A node of the tree is an instance of sub-types of {@link $Node}.
             * <p>
             * This class and any nested classes in this class are <strong>not synchronized</strong>.
             * 
            #if (javadocForManual) {
                * @since    notavaCC 1.0
            #}
             */
        #}
        >>[[ root.isProtectedParser() ? "" : "public " ]]<<class $Parser >>[[ separateOutput ? ("extends " + topLevelClassName + "Tables ") : "" ]]<<{
            #if (manualLocale.equals("ja")) {
                /**
                 * パーザを構築します。
                #if (javadocForManual) {
                     * <i><a href="../input.html#syntaxref:Root:ConstructorScope" class="symbol">ConstructorScope</a></i> が与えられた場合、コンストラクタはデフォルトスコープになります。
                #}
                 * @since    notavaCC 1.0
                 */
            #} else {
                /**
                 * Constructs a parser.
                #if (javadocForManual) {
                     * <i><a href="../ja/input.html#syntaxref:Root:ConstructorScope" class="symbol">ConstructorScope</a></i> makes the constructor the default scope.
                #}
                 * @since    notavaCC 1.0
                 */
            #}
            #if (root.hasProtectedConstructor()) {
                $Parser() {
                }
            #} else {
                public $Parser() {
                }
            #}
            
            
            #Map symbolToJavaWord = new LinkedHashMap();
            #Map symbolToJavadocWord = new LinkedHashMap();
            #{
                #Terminal traditionalWhiteTerminal = new NamedTerminal("$WHITE_TOKEN");
                #Symbol[] symbols = tables.symbols();
                #for (int i = 0; i < symbols.length; i++) {
                    #Symbol symbol = symbols[i];
                    #int index = tables.getIndexOfSymbol(symbol);
                    #if (symbol instanceof EOFTerminal) {
                        #java.lang.String word = "EOF_TOKEN";
                        #symbolToJavaWord.put(symbol, "$Parser." + word);
                        #symbolToJavadocWord.put(symbol, "$Parser#" + word);
                        #if (manualLocale.equals("ja")) {
                            /**
                             * EOF をあらわします.
                             * @see     $Parser.$Token#getSymbolID()
                             * @since    notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Describes the EOF.
                             * @see     $Parser.$Token#getSymbolID()
                             * @since    notavaCC 1.0
                             */
                        #}
                        public static final int >>[[ word ]]<< = >>[[ index ]]<<;
                        #topFieldNamespace.add(word, symbol);
                    #} else if (symbol.equals(traditionalWhiteTerminal)) {
                        #java.lang.String word = "WHITE_TOKEN";
                        #symbolToJavaWord.put(symbol, "$Parser." + word);
                        #symbolToJavadocWord.put(symbol, "$Parser#" + word);
                        /** Describes the white tokens, which are typically comments or white spaces. */
                        public static final int >>[[ word ]]<< = >>[[ index ]]<<;
                        #topFieldNamespace.add(word, symbol);
                    #} else if (symbol instanceof NamedTerminal) {
                        #NamedTerminal terminal = (NamedTerminal) symbol;
                        #java.lang.String word = "TOKEN_" + terminal.getName();
                        #symbolToJavaWord.put(symbol, "$Parser." + word);
                        #symbolToJavadocWord.put(symbol, "$Parser#" + word);
                        #if (manualLocale.equals("ja")) {
                            /**
                             * 終端記号 <code>>>[[ encodeForHTML(symbol) ]]<<</code> をあらわします。
                            #if (javadocForManual) {
                                 * このフィールドは、<i><a href="../input.html#syntax:TokenDefinition" class="symbol">TokenDefinition</a></i> で定義された終端記号ごとに１つずつ出力されます。<code>>>[[ encodeForHTML(symbol) ]]<<</code> は <i><a href="../input.html#syntaxref:TokenDefinition:IDENTIFIER" class="symbol">IDENITIFIER</a></i> で与えられる名前です。
                            #}
                             * @see     $Parser.$Token#getSymbolID()
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Describes the terminal symbol <code>>>[[ encodeForHTML(symbol) ]]<<</code>.
                            #if (javadocForManual) {
                                 * A <i><a href="../ja/input.html#syntax:TokenDefinition" class="symbol">TokenDefinition</a></i> generates this field, and The word <code>>>[[ encodeForHTML(symbol) ]]<<</code> is given by <i><a href="../ja/input.html#syntaxref:TokenDefinition:IDENTIFIER" class="symbol">IDENITIFIER</a></i>.
                            #}
                             * @see     $Parser.$Token#getSymbolID()
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        public static final int >>[[ word ]]<< = >>[[ index ]]<<;
                        #topFieldNamespace.add(word, symbol);
                    #} else if (symbol instanceof TypeNonterminal) {
                        #TypeNonterminal nonterminal = (TypeNonterminal) symbol;
                        #if (separateOutput) {
                            #symbolToJavaWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)));
                        #} else {
                            #symbolToJavaWord.put(symbol, nonterminal.getName() + ".$ID");
                        #}
                        #symbolToJavadocWord.put(symbol, nonterminal.getName() + "#$ID");
                    #} else if (symbol instanceof StringTerminal || symbol instanceof AliasNonterminal) {
                        #symbolToJavaWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)) + quotedAsComment(symbol.toString()));
                        #symbolToJavadocWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)));
                    #} else {
                        #assert symbol instanceof AnonymousNonterminal;
                        #symbolToJavaWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)));
                        #symbolToJavadocWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)));
                    #}
                    #/*if (!environment.debug) {
                        #symbolToJavaWord.put(symbol, Integer.toString(tables.getIndexOfSymbol(symbol)));
                    #}*/
                #}
            #}
            
            
            
            /**
            #if (manualLocale.equals("ja")) {
                 * 抽象構文木のノードを構築します。
                 * <p>
                #if (javadocForManual) {
                     * このメソッドは、引数で与えられた情報からノードを構築するファクトリーメソッドです。このメソッドをオーバーライドすることで、ノードの構築をカスタマイズすることができます。
                #}
                 * <p>
                 * 構築すべきノードの型が、<code>symbolID</code>によって、次のように指定されます。
            #} else {
                 * Creates a node of an abstract syntax tree.
                 * <p>
                #if (javadocForManual) {
                     * This is a factory method to create a node from the information given by the arguments.  The user can override the method in order to customize the node creation.
                #}
                 * <p>
                 * The type of the node to create is described by <code>symbolID</code> as follows.
            #}
             * <table border=1>
             * <tr><th><code>symbolID</code></th><th>type to create</th></tr>
            #{
                #Iterator it0 = tables.userDefinedNodeTypes().iterator();
                #while (it0.hasNext()) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) it0.next();
                    #if (javadocForManual && type.isPrivate()) {
                        #// Don't output ListOrNode.
                    #} else {
                        #TypeNonterminal nonterminal = type.getTypeNonterminal();
                        #if (nonterminal != null) {
                             * <tr>
                             *   <td>{@link >>[[ symbolToJavadocWord.get(nonterminal) ]]<<}</td>
                             *   <td>{@link >>[[ getTypeName(type) ]]<<}</td>
                             * </tr>
                        #}
                    #}
                #}
            #}
             * </table>
            #if (manualLocale.equals("ja")) {
                 * <p>
                 * このメソッドのデフォルトの実装は、<code>$createNode(symbolID, parameters, false)</code> と等価です。
                 * 
                 * @param   symbolID
                 *      構築すべきノードの型。
                 * @param   parameters
                 *      ノード構築するのに必要な情報。
                 * @return
                 *      構築されたノード。
                 * @throws   $ParseException
                 *      構文解析対象のテキストが、満たすべき何らかの条件を満たしていない場合。
            #} else {
                 * <p>
                 * The default implementation by the <code>$Parser</code> is equivalent to <code>$createNode(symbolID, parameters, false)</code>.
                 * 
                 * @param   symbolID
                 *      the kind of the node to create.
                 * @param   parameters
                 *      the parameters to initialize the node.
                 * @return
                 *      a created node.
                 * @throws   $ParseException
                 *      if the text to parse doesn't satisfy some requirements.
            #}
             * @since    notavaCC 1.0
             */
            #reserve("createNode", topMethodNamespace);
            protected $Node $createNode(int symbolID, $NodeInitializationParameters parameters) throws $ParseException {
                return $createNode(symbolID, parameters, false);
            }
            
            #if (manualLocale.equals("ja")) {
                /**
                 * 抽象構文木のノードを構築します。
                 * 構築されたノードは、{@link $Parser.$Default} に含まれるクラス（{@link $Parser.$Default.$Node} のサブクラス）のインスタンスです。
                 * 
                 * @param   symbolID
                 *      構築すべきノードの型。{@link #$createNode(int, $Parser.$NodeInitializationParameters)} を参照。
                 * @param   parameters
                 *      ノード構築するのに必要な情報。
                 * @param   compact
                 *      ラベル付けされたノードだけを子に持つノードを構築するかどうか。このメソッドは <code>parameters</code>で指定された子を持つノードを構築しますが、<code>compact == true</code> の場合、その際にラベル付けされていないノードを除外します。
                 * @return
                 *      構築されたノード。
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * Creates a node of an abstract syntax tree.  The created node is a instance of a class in {@link $Parser.$Default}.
                 * 
                 * @param   symbolID
                 *      the kind of the node to create.  See {@link #$createNode(int, $Parser.$NodeInitializationParameters)}.
                 * @param   parameters
                 *      the parameters to initialize the node.
                 * @param   compact
                 *      if <code>true</code>, the method creates a compact node, which have only labeled children.
                 * @return  a created node.
                 * @since   notavaCC 1.0
                 */
            #}
            protected final $Parser.$Default.$Node $createNode(int symbolID, $NodeInitializationParameters parameters, boolean compact) {
                switch(symbolID) {
                    #{
                        #Iterator it0 = tables.userDefinedNodeTypes().iterator();
                        #while (it0.hasNext()) {
                            #UserDefinedNodeType type = (UserDefinedNodeType) it0.next();
                            #TypeNonterminal nonterminal = type.getTypeNonterminal();
                            #if (nonterminal != null) {
                                case >>[[ symbolToJavaWord.get(nonterminal) ]]<<:
                                    return new $Parser.$Default.>>[[ type.getName() ]]<<(parameters, compact);
                            #}
                        #}
                    #}
                }
                throw new $AssertionError("Node creation failed.");
            }
            
            #if (ambiguitySupport) {
                #reserve("createAmbiguityAbsorber", topMethodNamespace);
                protected $Parser.$AmbiguityAbsorber $createAmbiguityAbsorber(java.util.List$<$Node> nodes) throws $ParseException {
                    $Parser.$AmbiguityAbsorber result;
                    #Iterator it38 = tables.userDefinedAbsorberTypes().iterator();
                    #while (it38.hasNext()) {
                        #UserDefinedAbsorberType type = (UserDefinedAbsorberType) it38.next();
                        #if (!type.isAbstract()) {
                            result = $Default.>>[[ type.getName() ]]<<.tryToCreate(nodes);
                            if (result != null)
                                return result;
                        #}
                    #}
                    return null;
                }
                #allocate("throwAmbiguousGrammerError", topMethodNamespace);
                private void $throwAmbiguousGrammerError(int symbolID, java.util.List$<$Node> nodes) throws $ParseException.AmbiguousGrammarError {
                    java.lang.String[] symbolIDToName = {
                        #for (int i = 0; i < tables.symbols().length; i++) {
                            >>[[ quoted(tables.symbols()[i].toString()) ]]<<,
                        #}
                    };
                    java.io.StringWriter buffer = new java.io.StringWriter();
                    java.io.PrintWriter writer = new java.io.PrintWriter(buffer);
                    writer.println("DETECTED UNABSORBED AMBIGUITY OF " + >>[[ quoted(environment.processingFile.getName()) ]]<<);
                    writer.println("This message is for the developer of the parser generated by notavaCC.");
                    writer.println("If you are not the developer of the program that output this message, contact the developper.");
                    writer.println("----");
                    writer.println(symbolIDToName[symbolID] + " is ambiguous.  The following are the possible trees of the input, but there is no $absorber that absorbs this ambiguity.");
                    nodes = new java.util.ArrayList$<$Node>(nodes);
                    for (int i = 0; i < nodes.size(); i++) {
                        writer.println("[Tree " + i + "]");
                        writer.println(nodes.get(i));
                    }
                    writer.println("----");
                    writer.print("Perhaps, an easy way to avoid this error is to add the following line to " + >>[[ quoted(environment.processingFile.getName()) ]]<<);
                    #if (!root.absorberDefinitions2().isEmpty())
                        writer.print(" before " + >>[[ quoted(((AbsorberDefinition2) root.absorberDefinitions2().get(0)).identifier().toString()) ]]<<);
                    writer.println(".");
                    writer.println();
                    writer.print("$absorber AbsorberNameYouLike { ");
                    for (int i = 0; i < nodes.size(); i++) {
                        if (i != 0)
                            writer.print(" | ");
                        
                        java.lang.String className = nodes.get(i).getClass().getName();
                        className = className.substring(className.lastIndexOf('.') + 1);
                        className = className.substring(className.lastIndexOf('$') + 1);
                        writer.print("label" + i + ":" + className);
                    }
                    writer.println(" }");
                    
                    writer.close();
                    throw new $ParseException.AmbiguousGrammarError(buffer.toString());
                }
            #}
            
            #if (manualLocale.equals("ja")) {
                /**
                 * 構文木全体が構築されたときに呼び出され、構文木を編集することができます。
                 * このメソッドをオーバーライドすることで、構築された構文木がユーザに返される直前に、その構文木を編集することができます。
                 * 
                 * @param   root
                 *      {@link $Parser#$createNode} によって作られた構文木の根。
                 * @return  パーズ結果として扱われる構文木の根。
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * This method is called when an whole abstract syntax tree is built.
                 * The tree this method returns is regarded as the result of parsing.
                 * 
                 * <p>
                 * 
                 * @param   root
                 *      the root of the tree generated by {@link $Parser#$createNode}.
                 * @return  the result of parsing.
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("modifyWholeTree", topMethodNamespace);
            protected $Node $modifyWholeTree($Node root) throws $ParseException {
                return root;
            }
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * ファイルに格納されている文字列を表す <code>$CharSequence</code> を得ます。
                 * <p>
                 * 
                 * @param   file
                 *      ファイル。
                 * @param   charsetName
                 *      文字セットの名前。デフォルトの文字セットを指定する場合 <code>null</code>。
                 * @return
                 *      ファイルに格納されている文字列を表す <code>$CharSequence</code>。
                 * @throws   java.io.IOException
                 *      ファイルへのアクセス中にエラーが発生した場合。
                 * @since    notavaCC 1.0
                 */
            #} else {
                /**
                 * Gets the <code>$CharSequence</code> that represents the text in the given file.
                 * <p>
                 * 
                 * @param   file
                 *      a file.
                 * @param   charsetName
                 *      the charset name of the file, or <code>null</code> for the default charset.
                 * @return
                 *      the <code>$CharSequence</code> that represents the text in the <code>file</code>.
                 * @throws   java.io.IOException
                 *      if some error is occured for the access to the <code>file</code>.
                 * @since    notavaCC 1.0
                 */
            #}
            #reserve("toCharSequence", topMethodNamespace);
            #if (targetVersion >= 1.4) {
                protected final $CharSequence $toCharSequence(java.io.File file, java.lang.String charsetName) throws java.io.IOException {
                    java.io.FileInputStream fis = null;
                    java.io.InputStreamReader reader = null;
                    java.nio.channels.FileChannel fc = null;
                    try {
                        fis = new java.io.FileInputStream(file);
                        if (charsetName == null)
                            reader = new java.io.InputStreamReader(fis);
                        else
                            reader = new java.io.InputStreamReader(fis, charsetName);
                        java.nio.charset.Charset charset;
                        try {
                            charset = java.nio.charset.Charset.forName(reader.getEncoding());
                        } catch(java.nio.charset.UnsupportedCharsetException x) {
                            charset = null;
                        }
                        if (charset != null) {
                            java.nio.charset.CharsetDecoder decoder = charset.newDecoder();
                            fc = fis.getChannel();
                            int size = (int) fc.size();
                            java.nio.MappedByteBuffer bb = fc.map(java.nio.channels.FileChannel.MapMode.READ_ONLY, 0, size);
                            return decoder.decode(bb);
                        } else {
                            long length = file.length();
                            if (length <= 0)
                                length = 1;
                            else if (length > java.lang.Integer.MAX_VALUE)
                                length = java.lang.Integer.MAX_VALUE;
                            $StringBuffer result = new $StringBuffer((int) length);
                            char[] buffer = new char[1024];
                            for(;;) {
                                int size = reader.read(buffer);
                                if (size < 0)
                                    break;
                                result.append(buffer, 0, size);
                            }
                            return result;
                        }
                    } finally {
                        if (fc != null)
                            fc.close();
                        if (reader != null)
                            reader.close();
                        else if (fis != null)
                            fis.close();
                    }
                }
            #} else {
                protected final $CharSequence $toCharSequence(java.io.File file, java.lang.String charsetName) throws java.io.IOException {
                    java.io.FileInputStream fis = null;
                    java.io.InputStreamReader reader = null;
                    try {
                        fis = new java.io.FileInputStream(file);
                        if (charsetName == null)
                            reader = new java.io.InputStreamReader(fis);
                        else
                            reader = new java.io.InputStreamReader(fis, charsetName);
                        long length = file.length();
                        if (length <= 0)
                            length = 1;
                        else if (length > java.lang.Integer.MAX_VALUE)
                            length = java.lang.Integer.MAX_VALUE;
                        $StringBuffer result = new $StringBuffer((int) length);
                        char[] buffer = new char[1024];
                        for (;;) {
                            int size = reader.read(buffer);
                            if (size < 0)
                                break;
                            result.append(buffer, 0, size);
                        }
                        return result;
                    } finally {
                        if (reader != null)
                            reader.close();
                        else if (fis != null)
                            fis.close();
                    }
                }
            #}
            #if (manualLocale.equals("ja")) {
                /**
                 * 文字ストリームから得られる文字列を表す <code>$CharSequence</code> を得ます。
                 * <p>
                 * このメソッドは、文字ストリームの終端に達するまで文字列を読み出し、その文字列をあらわす <code>$CharSequence</code> を返します。
                 * 
                 * @param   reader
                 *      文字ストリーム。このストリームは <code>close()</code> されません。
                 * @return
                 *      文字ストリームから得られる文字列を表す <code>$CharSequence</code>。
                 * @throws   java.io.IOException
                 *      文字ストリームへのアクセス中にエラーが発生した場合。
                 * @since    notavaCC 1.0
                 */
            #} else {
                /**
                 * Gets the <code>$CharSequence</code> that represents the text from the given character stream.
                 * <p>
                 * This method reads characters from the stream until the end of the stream has been reached, and returns the <code>$CharSequence</code> that represents the characters.
                 * 
                 * @param   reader
                 *      a character stream.  This method will not <code>close</code> it.
                 * @return
                 *      the <code>$CharSequence</code> that represents the text from the <code>reader</code>.
                 * @throws   java.io.IOException
                 *      if some error is occured for the access to the <code>reader</code>.
                 * @since    notavaCC 1.0
                 */
            #}
            protected final $CharSequence $toCharSequence(java.io.Reader reader) throws java.io.IOException {
                $StringBuffer result = new $StringBuffer();
                char[] buffer = new char[1024];
                for (;;) {
                    int size = reader.read(buffer);
                    if (size < 0)
                        break;
                    result.append(buffer, 0, size);
                }
                return result;
            }
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * 字句解析器を構築します。
                 * <p>
                #if (javadocForManual) {
                     * このメソッドは、字句解析器を構築するファクトリーメソッドです。このメソッドをオーバーライドすることで、字句解析をカスタマイズすることができます。
                #}
                 * 
                 * @param   sourceName
                 *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                 * @param   text
                 *      解析すべきテキスト。
                 * @param   tabStop
                 *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                 * @return
                 *      与えられたテキストを対象に字句解析を行う字句解析器。
                 * @throws   $ParseException
                 *      構文解析対象のテキストが、満たすべき何らかの条件を満たしていない場合。
                 * @since    notavaCC 1.0
                 */
            #} else {
                /**
                 * Creates a lexical analyzer.
                 * <p>
                #if (javadocForManual) {
                     * This is a factory method to create a lexical analyzer.  The user can override the method in order to customize the lexical analysis.
                #}
                 * @param   sourceName
                 *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                 * @param   text
                 *      the text to analyze.
                 * @param   tabStop
                 *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                 * @return
                 *      the lexical analyzer that analyze the <code>text</code>.
                 * @throws   $ParseException
                 *      if the text to parse doesn't satisfy some requirements.
                 * @since    notavaCC 1.0
                 */
            #}
            #reserve("createLexicalAnalyzer", topMethodNamespace);
            protected $Parser.$LexicalAnalyzer $createLexicalAnalyzer(java.lang.String sourceName, $CharSequence text, int tabStop) throws $ParseException {
                return new $Parser.$Default.$LexicalAnalyzer(sourceName, text, tabStop);
            }
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * ノードの構築に必要な情報を <code>private</code>なフィールドで保持するオブジェクト。このクラスのコンストラクタは<code>private</code>です。
                 * <p>
                 * 
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * Contains <code>private</code> fields that is used in order to create a node.  The constructor is <code>private</code>.
                 * <p>
                 * 
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("NodeInitializationParameters", topTypeNamespace);
            $protected_avoid_jdk12_bug static class $NodeInitializationParameters {
                private $NodeInitializationParameters() {
                }
                private $Node[] childNodes;
                private java.util.List labelBitsList;
            }
            #if (visitorSupport) {
                #for (int numberOfExceptions = 0; numberOfExceptions <= ((targetVersion < 1.5 || environment.noParameterizedVisitor) ? 0 : 4); numberOfExceptions++) {
                    #if (manualLocale.equals("ja")) {
                        /**
                         * ツリーをトラバースしながら何らかの処理を行うオブジェクト。
                         * <p>
                         * 
                         * @see     $Parser.$Visitor#visitAll($Parser.$Node)
                         * @since   notavaCC 1.0
                         */
                    #} else {
                        /**
                         * A visitor.
                         * <p>
                         * 
                         * @see     $Parser.$Visitor#visitAll($Parser.$Node)
                         * @since   notavaCC 1.0
                         */
                    #}
                    #reserve("Visitor" + ((numberOfExceptions == 0) ? "" : ("" + numberOfExceptions)), topTypeNamespace);
                    #StringBuffer typeParameters = new StringBuffer();
                    #StringBuffer throwsExceptions = new StringBuffer();
                    #if (numberOfExceptions > 0) {
                        #typeParameters.append("<");
                        #throwsExceptions.append(" throws ");
                        #for (int i = 0; i < numberOfExceptions; i++) {
                            #if (i != 0) {
                                #typeParameters.append(", ");
                                #throwsExceptions.append(", ");
                            #}
                            #typeParameters.append("X" + i + " extends java.lang.Throwable");
                            #throwsExceptions.append("X" + i);
                        #}
                        #typeParameters.append(">");
                    #}
                    public static class $Visitor>>[[ (numberOfExceptions == 0) ? "" : ("" + numberOfExceptions + typeParameters) ]]<< {
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * （部分）木をトラバースします。
                             * <p>
                             * このメソッドのデフォルトの実装は、次のとおりです。ユーザはこのメソッドをオーバーライドし、挙動を自由に変えてかまいません。
                             * <pre>
                             * visitNode(node);
                             * 
                             * java.util.Iterator it = node.$getChildNodes().iterator();
                             * while (it.hasNext()) {
                             *     $Parser.$Node child = ($Parser.$Node) it.next();
                             *     visitAll(child);
                             * }
                             * </pre>
                             * 
                             * @param   node
                             *      トラバースする（部分）木の根。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Traverses a (sub-)tree.
                             * <p>
                             * The default implementation is as follows.  The user can override the method and change the behavior at will.
                             * <pre>
                             * visitNode(node);
                             * 
                             * java.util.Iterator it = node.$getChildNodes().iterator();
                             * while (it.hasNext()) {
                             *     $Parser.$Node child = ($Parser.$Node) it.next();
                             *     visitAll(child);
                             * }
                             * </pre>
                             * 
                             * @param   node
                             *      the root of the (sub-)tree to traverse.
                             * @since   notavaCC 1.0
                             */
                        #}
                        public void visitAll($Parser.$Node node)>>[[ throwsExceptions ]]<< {
                            visitNode(node);
                            
                            java.util.Iterator it = node.$getChildNodes().iterator();
                            while (it.hasNext()) {
                                $Parser.$Node child = ($Parser.$Node) it.next();
                                visitAll(child);
                            }
                        }
                        #if (manualLocale.equals("ja")) {
                            /**
                             * <code>visit</code> メソッドを呼び出します。
                             * <p>
                             * このメソッドのデフォルトの実装は、このクラスで定義されるメソッドで、名前が <code>visit</code> であり、<code>node</code> を引数に取ることのできるメソッドを全て呼び出します。例えば <code>node instanceof {@link $Parser.$Token}</code> であれば、{@link #visit($Parser.$Node)} と、{@link #visit($Parser.$Token)} が呼び出されます。呼び出される順番は、規定されません。ユーザはこのメソッドをオーバーライドし、挙動を自由に変えてかまいません。
                             * 
                             * @param   node
                             *      <code>visit</code> メソッドへ渡されるノード
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Call <code>visit</code>.
                             * <p>
                             * The default implementation is to call all the method, that is defined in this class, whose name is <code>visit</code>, and that can take <code>node</code> as the argument.  For example, if <code>node instanceof {@link $Parser.$Token}</code>, both of {@link #visit($Parser.$Node)} and {@link #visit($Parser.$Token)} is called.  The order of the call is undefined.  The user can override the method and change the behavior at will.
                             * 
                             * @param   node
                             *      a node.
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected void visitNode($Parser.$Node node)>>[[ throwsExceptions ]]<< {
                            visit(($Parser.$Node) node);
                            
                            if (node instanceof $Parser.$Token)
                                visit(($Parser.$Token) node);
                            #{
                                #final MultiMap parentToChildren = new MultiMap();
                                #Iterator it0 = tables.userDefinedNodeTypes().iterator();
                                #while (it0.hasNext()) {
                                    #UserDefinedNodeType type = (UserDefinedNodeType) it0.next();
                                    #Iterator it15 = type.getDirectSuperTypes().iterator();
                                    #while (it15.hasNext()) {
                                        #Type supertype = (Type) it15.next();
                                        #parentToChildren.add(supertype, type);
                                    #}
                                #}
                                #final Set checkedTypes = new HashSet();
                                #new Object() {
                                    #public void output(Type type) {
                                        #Set set = parentToChildren.getValueSet(type);
                                        #if (checkedTypes.add(type)) {
                                            #if (type instanceof NodeRootType) {
                                                #if (set != null) {
                                                    #Iterator it16 = set.iterator();
                                                    #while (it16.hasNext()) {
                                                        #UserDefinedNodeType child = (UserDefinedNodeType) it16.next();
                                                        
                                                        #output(child);
                                                    #}
                                                #}
                                            #} else {
                                                #UserDefinedNodeType udt = (UserDefinedNodeType) type;
                                                #if (!udt.isPrivate()) {
                                                    #if (set == null || set.isEmpty()) {
                                                        if (node instanceof >>[[ getTypeName(type) ]]<<)
                                                            visit((>>[[ getTypeName(type) ]]<<) node);
                                                    #} else {
                                                        if (node instanceof >>[[ getTypeName(type) ]]<<) {
                                                            visit((>>[[ getTypeName(type) ]]<<) node);
                                                            
                                                            #Iterator it16 = set.iterator();
                                                            #while (it16.hasNext()) {
                                                                #UserDefinedNodeType child = (UserDefinedNodeType) it16.next();
                                                                #output(child);
                                                            #}
                                                        }
                                                    #}
                                                #}
                                            #}
                                        #}
                                    #}
                                #}.output(NodeRootType.INSTANCE);
                            #}
                        }
                        
                        /**
                        #if (manualLocale.equals("ja")) {
                             * ノードへ訪れるたびに呼び出されます。
                             * <p>
                             * 
                             * @param   node
                             *      訪れたノード。
                        #} else {
                             * Called for a visited node.
                             * <p>
                             * 
                             * @param   node
                             *      a visited node.
                        #}
                         * @see      #visitAll
                         * @since    notavaCC 1.0
                         */
                        protected void visit($Parser.$Node node)>>[[ throwsExceptions ]]<< {
                        }
                        /**
                        #if (manualLocale.equals("ja")) {
                             * {@link $Parser.$Token}へ訪れるたびに呼び出されます。
                             * <p>
                             * 
                             * @param   node
                             *      訪れたノード。
                        #} else {
                             * Called for a visited {@link $Parser.$Token}.
                             * <p>
                             * 
                             * @param   node
                             *      a visited node.
                        #}
                         * @see      #visitAll
                         * @since    notavaCC 1.0
                         */
                        protected void visit($Parser.$Token node)>>[[ throwsExceptions ]]<< {
                        }
                        #{
                            #Iterator it0 = tables.userDefinedNodeTypes().iterator();
                            #while (it0.hasNext()) {
                                #UserDefinedNodeType type = (UserDefinedNodeType) it0.next();
                                #if (!type.isPrivate()) {
                                    /**
                                    #if (manualLocale.equals("ja")) {
                                         * {@link >>[[ getTypeName(type) ]]<<}、およびそのサブタイプのノードへ訪れるたびに呼び出されます。
                                         * <p>
                                        #if (javadocForManual) {
                                             * このメソッドは、<code>$private</code> ではない１つの <i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> 毎に１つ出力されます。
                                        #}
                                         * 
                                         * @param   node
                                         *      訪れたノード。
                                    #} else {
                                         * Called for a visited {@link >>[[ getTypeName(type) ]]<<}.
                                         * <p>
                                        #if (javadocForManual) {
                                             * A non-<code>$private</code> <i><a href="../ja/input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> generates this method.
                                        #}
                                         * 
                                         * @param   node
                                         *      a visited node.
                                    #}
                                     * @see      #visitAll
                                     * @since    notavaCC 1.0
                                     */
                                    protected void visit(>>[[ getTypeName(type) ]]<< node)>>[[ throwsExceptions ]]<< {
                                    }
                                #}
                            #}
                        #}
                    }
                #}
            #}
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * 構文解析の結果として得られる抽象構文木のノード。
                 * <p>
                 * 
                #if (javadocForManual) {
                     * @see     $Parser.$Type
                     * @see     $Parser.Token
                     * @see     #$parseType
                #}
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * A node of the abstract syntax tree that is the result of parsing.
                 * <p>
                 * 
                #if (javadocForManual) {
                     * @see     $Parser.$Type
                     * @see     $Parser.Token
                     * @see     #$parseType
                #}
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("Node", topTypeNamespace);
            public static interface $Node {
                #if (useMyList) {
                    public static class ConcreteList extends java.util.LinkedList<$Node> {
                        private static final long serialVersionUID = 0;
                        public ConcreteList() {
                        }
                        public ConcreteList(java.util.Collection<? extends $Node> c) {
                            super(c);
                        }
                    }
                #}
                #if (manualLocale.equals("ja")) {
                    /**
                     * 子のノードのリストを返します。
                    #if (javadocForManual) {
                         * <p>
                         * ￢＜＞＜∪∪が出力するコードによって作られる $Node 型のオブジェクトでは、リストは変更可能で、リストの要素は構文解析対象のテキストに出現した順番に並びます。しかし、ユーザが構築したノードがこの条件を満たす必要はありません。
                    #}
                     * 
                     * @return  子のノードのリスト。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the list of the child nodes of <code>this</code>.
                    #if (javadocForManual) {
                         * <p>
                         * For a $Node created by the program generated by &#xAC;&lt;&gt;&lt;&#x222A;&#x222A;, the list should be modifiable and the nodes should be ordered by its position in the parsed text.  However, the nodes constructed by the users of &#xAC;&lt;&gt;&lt;&#x222A;&#x222A; are not required to satisfy it.
                    #}
                     * 
                     * @return  the list of the child nodes of <code>this</code>.
                     * @since   notavaCC 1.0
                     */
                #}
                #reserve("getChildNodes", nodeNoArgumentMethodNamespace);
                public java.util.List$<$Parser.$Node> $getChildNodes();
                
                #if (ambiguitySupport) {
                    #reserve("getWrappedAmbiguityAbsorber", nodeNoArgumentMethodNamespace);
                    public $AmbiguityAbsorber $getWrappedAmbiguityAbsorber();
                #}
            } // interface $Node
            
            #if (manualLocale.equals("ja")) {
                /**
                 * トークン（構文解析対象のテキストの断片）。
                 * <p>
                 * トークンは構文解析対象のテキスト（ソース）の断片です。通常、トークンは字句解析器（{@link $Parser.$LexicalAnalyzer}）によって作られます。
                 * このクラスには、そのトークンがどのソースのどの位置の断片であるかをあらわすメソッド（{@link #getSourceName}、{@link #getIndex}、{@link #getLine}、{@link #getColumn}）が用意されていますが、これらが返す値の厳密な意味は、字句解析器などのそのトークンを構築するプログラムによって規定されます。また、これらのメソッドはサポートされていなくてもよく、その場合 {@link java.lang.UnsupportedOperationException} 例外が発生します。
                 * 
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * A token, which is a fragment of a parsed text.
                 * <p>
                 * A token is a fragment of a parsed text (a source).  Tokens are usually constructed by {@link $Parser.$LexicalAnalyzer}s.  This class introduces the methods that represent what is the source of the token and where the token is placed in the source ({@link #getSourceName}, {@link #getIndex}, {@link #getLine}, {@link #getColumn}), however, the strict specification of the returned value is defined by the generator of the token, e.g., <code>$LexicalAnalyzer</code>s.  Also, these methods may be unsuported and may throw {@link java.lang.UnsupportedOperationException}.
                 * 
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("Token", topTypeNamespace);
            public static interface $Token extends $Node {
                #if (useMyList) {
                    public static class ConcreteList extends java.util.LinkedList<$Token> {
                        private static final long serialVersionUID = 0;
                        public ConcreteList() {
                        }
                        public ConcreteList(java.util.Collection<? extends $Token> c) {
                            super(c);
                        }
                    }
                #}
                #if (manualLocale.equals("ja")) {
                    /**
                     * このトークンの種類を返します。
                     * <p>
                     * 戻り値は、次のいずれかです。
                     * 
                     * <table border=1>
                     * <tr><th>戻り値</th><th>トークンの種類</th></tr>
                    #{
                        #Symbol[] symbols = tables.symbols();
                        #for (int i = 0; i < tables.superiorIndexOfTerminals(); i++) {
                            # Symbol symbol = symbols[i];
                             * <tr>
                             *   <td>{@link >>[[ symbolToJavadocWord.get(symbol) ]]<<}</td>
                             *   <td><tt>>>[[ encodeForHTML(symbol) ]]<<</tt></td>
                             * </tr>
                        #}
                    #}
                    #if (javadocForManual) {
                         * <tr>
                         *   <td>その他の値</td>
                         *   <td><a href="../input.html#syntaxref:expression:STRING"><code>STRING</code></a>で定義される非終端記号</td>
                         * </tr>
                    #}
                     * </table>
                     * 
                     * @return  このトークンの種類。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the kind of <code>this</code> token.
                     * <p>
                     * The result value should be one of the following:</p>
                     * 
                     * <table border=1>
                     * <tr><th>value</th><th>described token</th></tr>
                    #{
                        #Symbol[] symbols = tables.symbols();
                        #for (int i = 0; i < tables.superiorIndexOfTerminals(); i++) {
                            # Symbol symbol = symbols[i];
                             * <tr>
                             *   <td>{@link >>[[ symbolToJavadocWord.get(symbol) ]]<<}</td>
                             *   <td><tt>>>[[ encodeForHTML(symbol) ]]<<</tt></td>
                             * </tr>
                        #}
                    #}
                    #if (javadocForManual) {
                         * <tr>
                         *   <td>others</td>
                         *   <td>an instance of a terminal defined by <a href="../ja/input.html#syntaxref:expression:STRING"><code>STRING</code></a></td>
                         * </tr>
                    #}
                     * </table>
                     * 
                     * @return  the kind of <code>this</code> token.
                     * @since   notavaCC 1.0
                     */
                #}
                public int getSymbolID();
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * このトークンがホワイト・トークンの場合、<code>true</code>を返します。
                     * <p>
                    #if (!javadocForManual) {
                        #if (tables.firstIndexOfWhiteTerminals() < tables.superiorIndexOfWhiteTerminals()) {
                             * This method returns <code>true</code> if and only if <code>getSymbolID()</code> is one of the following:
                             * <ul>
                            #{
                                #Symbol[] symbols = tables.symbols();
                                #for (int i = tables.firstIndexOfWhiteTerminals(); i < tables.superiorIndexOfWhiteTerminals(); i++) {
                                    # Symbol symbol = symbols[i];
                                     *   <td>{@link >>[[ symbolToJavadocWord.get(symbol) ]]<<}</td>
                                #}
                            #}
                             * </ul>
                        #} else {
                             * This method always returns <code>false</code> because the grammar represented by $Parser has no white token.
                        #}
                    #}
                     * 
                     * @return  このトークンがホワイト・トークンの場合 <code>true</code>。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returnds <code>true</code> if <code>this</code> is a white token.
                     * <p>
                    #if (!javadocForManual) {
                        #if (tables.firstIndexOfWhiteTerminals() < tables.superiorIndexOfWhiteTerminals()) {
                             * This method returns <code>true</code> if and only if <code>getSymbolID()</code> is one of the following:
                             * <ul>
                            #{
                                #Symbol[] symbols = tables.symbols();
                                #for (int i = tables.firstIndexOfWhiteTerminals(); i < tables.superiorIndexOfWhiteTerminals(); i++) {
                                    # Symbol symbol = symbols[i];
                                     *   <td>{@link >>[[ symbolToJavadocWord.get(symbol) ]]<<}</td>
                                #}
                            #}
                             * </ul>
                        #} else {
                             * This method always returns <code>false</code> because the grammar represented by $Parser has no white token.
                        #}
                    #}
                     * 
                     * @return  <code>true</code> if <code>this</code> is a while token.
                     * @since   notavaCC 1.0
                     */
                #}
                public boolean isWhite();
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * このトークンに対応する文字列を返します。
                     * <p>
                     * 文字列がユニコードエスケープのようなメタ表現を含む場合、戻り値は、表現されている文字列（例えば A）であり、表現（例えば \u005Cu0041）ではありません。そのようなメタ表現が含まれていない場合、このメソッドの戻り値は {@link #getOriginalImage()} と同じになります。
                     * 
                     * @return  このトークンに対応する文字列。
                     * @see     $Parser.$Token#getOriginalImage()
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the string <code>this</code> token corresponds to.
                     * <p>
                     * If the string contains some meta notation like Unicode escapes, the returned value is the represented string (e.g., A) rather than the representation (e.g., \u005Cu0041).  If not, the returned value is the same as {@link #getOriginalImage()}.
                     * 
                     * @return  the string <code>this</code> token corresponds to.
                     * @see     $Parser.$Token#getOriginalImage()
                     * @since   notavaCC 1.0
                     */
                #}
                public java.lang.String getImage();
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * このトークンに対応する文字列を返します。
                     * <p>
                     * 文字列がユニコードエスケープのようなメタ表現を含む場合、戻り値は、表現（例えば \u005Cu0041）であり、表現されている文字列（例えば A）ではありません。そのようなメタ表現が含まれていない場合、このメソッドの戻り値は {@link #getImage()} と同じになります。
                     * 
                     * @return  このトークンに対応する文字列。
                     * @see     $Parser.$Token#getImage()
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the string <code>this</code> token corresponds to.
                     * <p>
                     * If the string contains some meta notation like Unicode escapes, the returned value is the representation (e.g., \u005Cu0041) rather than the represented string (e.g., A).  If not, the returned value is the same as {@link #getImage()}.
                     * 
                     * @return  the string <code>this</code> token corresponds to.
                     * @see     $Parser.$Token#getImage()
                     * @since   notavaCC 1.0
                     */
                #}
                public java.lang.String getOriginalImage();
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * このトークンに対応するソースの名前を返します。
                     * 
                     * @return  ソースの名前。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the source name <code>this</code> token coressponds to.
                     * 
                     * @return  the source name <code>this</code> token coressponds to.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public java.lang.String getSourceName() throws java.lang.UnsupportedOperationException;
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * ソース上の、このトークンが始まった位置のインデックス（0 based）を返します。
                     * 
                     * @return  このトークンが始まった位置のインデックス。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the zero-based index number in the source where <code>this</code> token begins.
                     * 
                     * @return  the index number in the source where <code>this</code> token begins.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public int getIndex() throws java.lang.UnsupportedOperationException;
                #if (manualLocale.equals("ja")) {
                    /**
                     * ソース上の、このトークンが始まった位置の行番号（1 based）を返します。
                     * 
                     * @return  このトークンが始まった位置の行番号。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the one-based line number in the source where <code>this</code> token begins.
                     * 
                     * @return  the line number in the source where <code>this</code> token begins.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public int getLine() throws java.lang.UnsupportedOperationException;
                #if (manualLocale.equals("ja")) {
                    /**
                     * ソース上の、このトークンが始まった位置のカラム番号（1 based）を返します。
                     * 
                     * @return  このトークンが始まった位置のカラム番号。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the one-based column number in the source where <code>this</code> token begins.
                     * 
                     * @return  the column number in the source where <code>this</code> token begins.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public int getColumn() throws java.lang.UnsupportedOperationException;
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * トークンの {@linkplain #getOriginalImage イメージ}と、サポートされている場合その位置を返します。
                     * 例えば、<code>`identifier' (line 2, column 3)</code> や <code>`identifier'</code> のようになります。厳密なフォーマットは規定されません。
                     * 
                     * @return  トークンのイメージと位置
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the {@linkplain #getOriginalImage image} and the position of <code>this</code> token if supported.
                     * For instance, <code>`identifier' (line 2, column 3)</code> or <code>`identifier'</code>.
                     * The format is not specified strictly.
                     * 
                     * @return  the {@linkplain #getOriginalImage image} and the parenthesized position of <code>this</code> token.
                     * @since   notavaCC 1.0
                     */
                #}
                public java.lang.String toString();
            } // interface $Token
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * {@link $Parser.$Token} の列を生成するオブジェクト。
                 * 
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * Generator of a sequence of {@link $Parser.$Token}s.
                 * 
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("LexicalAnalyzer", topTypeNamespace);
            public static abstract class $LexicalAnalyzer {
                #if (manualLocale.equals("ja")) {
                    /**
                     * シーケンスの次のトークンを返します。
                     * <p>
                     * もしトークンが無い場合、{@link $Parser.$Token#getSymbolID()} が {@link $Parser#EOF_TOKEN} であるようなトークンを返します。この時の {@link $Parser.$Token#getImage()} や {@link $Parser.$Token#getIndex()} などの戻り値がどうなるかは規定されません。
                     * 
                     * @return  次のトークン。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the next token in the sequence.
                     * <p>
                     * If there is no tokens, the {@linkplain $Parser.$Token#getSymbolID() symbol ID} of the returned value should be {@link $Parser#EOF_TOKEN}, but the {@linkplain $Parser.$Token#getImage() image} or the {@linkplain $Parser.$Token#getIndex() index} etc. are not specified (may be {@linkplain java.lang.UnsupportedOperationException unsupported}).
                     * 
                     * @return  the next token in the sequence.
                     * @since   notavaCC 1.0
                     */
                #}
                public abstract $Token next() throws $ParseException;
            } // $LexicalAnalyzer
            
            
            #if (ambiguitySupport) {
                #reserve("AmbiguityAbsorber", topTypeNamespace);
                public static interface $AmbiguityAbsorber extends $Node {
                }
                
                #Iterator it38 = tables.userDefinedAbsorberTypes().iterator();
                #while (it38.hasNext()) {
                    #UserDefinedAbsorberType type = (UserDefinedAbsorberType) it38.next();
                    #String scope = "public";
                    #if (type.isProtected()) {
                        #scope = "protected";
                    #} else if (type.isPrivate()) {
                        #scope = "private";
                    #}
                    #StringBuffer extendedTypes = new StringBuffer();
                    #{
                        #Iterator inner = type.getDirectSuperTypes().iterator();
                        #while (inner.hasNext()) {
                            #Type superType = (Type) inner.next();
                            #if (extendedTypes.length() != 0)
                                #extendedTypes.append(", ");
                            #extendedTypes.append(getTypeName(superType));
                        #}
                    #}
                    
                    >>[[ scope ]]<< static interface >>[[ type.getName() ]]<< extends >>[[ extendedTypes ]]<< {
                        #Iterator it33 = type.getDeclaredFields().iterator();
                        #while (it33.hasNext()) {
                            #Field field = (Field) it33.next();
                            public >>[[ getTypeName(field.getType()) ]]<< >>[[ field.getName() ]]<<();
                        #}
                    }
                #}
            #}
            
            #int publicParsableTypeCount = 0;
            #{
                #reserve("ID", nodeFieldNamespace);
                #Iterator it1 = tables.userDefinedNodeTypes().iterator();
                #while (it1.hasNext()) {
                    #UserDefinedNodeType type = (UserDefinedNodeType) it1.next();
                    #String scope = type.isProtected() ? "protected" : type.isPrivate() ? "private" : "public";
                    #String methodScope = type.isProtectedParsable() ? "protected" : "public";
                    #LRState initialState = (LRState) lrTable.symbolToInitialState().get(type.getTypeNonterminal());
                    #if (initialState != null) {
                        #if (methodScope.equals("public"))
                            #publicParsableTypeCount++;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * ファイルに含まれる <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> を構文解析します。
                            #if (javadocForManual) {
                                 * <p>
                                 * ファイルに含まれるテキストを、非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> をゴール記号（開始記号）とする文法に基づき構文解析します。
                                 * このメソッドは、<code>$parsable</code> な非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> に対して出力されます。<code>>>[[ encodeForHTML(type.getName()) ]]<<</code> が <code>$parsable-protected</code> の場合、このメソッドは <code>protected</code> になります
                            #}
                             * <p>
                             * このメソッドの呼び出しは、<code>parse>>[[ encodeForHTML(type.getName()) ]]<<(file.getPath(), file, charsetName, 8)</code> の呼び出しと等価です。
                             * 
                             * @param   file
                             *      構文解析対象のテキストを含むファイル。
                             * @param   charsetName
                             *      文字セットの名前。デフォルトの文字セットを指定する場合 <code>null</code>。
                             * @return
                             *      抽象構文木
                             * @throws   java.io.IOException
                             *      フェイルへのアクセス中にエラーが発生した場合。
                             * @throws   $ParseException
                             *      ファイルに含まれるテキストが、満たすべき条件を満たしていなかった場合。
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Parses a <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> in the given file.
                            #if (javadocForManual) {
                                 * <p>
                                 * This method parses a text in the given file according to the grammar whose start symbol is <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                                 * A <code>$parsable</code> nonterminal symbol <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> generates this method.  If <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> is <code>$parsable-protected</code>, this method is <code>protected</code>.
                            #}
                             * <p>
                             * This is equivalent to <code>parse>>[[ encodeForHTML(type.getName()) ]]<<(file.getPath(), file, charsetName, 8)</code>.
                             * 
                             * @param   file
                             *      the file including the text to parse.
                             * @param   charsetName
                             *      the charset name of the file, or <code>null</code> for the default charset.
                             * @return
                             *      the abstract syntax tree.
                             * @throws   java.io.IOException
                             *      if some error is occured for the access to the file.
                             * @throws   $ParseException
                             *      if the text in the file doesn't satisfy some requirements.
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        #topMethodNamespace.add("parse" + type.getName(), type);
                        >>[[ methodScope ]]<< >>[[ getTypeName(type) ]]<< parse>>[[ type.getName() ]]<<(java.io.File file, java.lang.String charsetName) throws java.io.IOException, $ParseException {
                            return parse>>[[ type.getName() ]]<<(file.getPath(), file, charsetName, 8);
                        }
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * ファイルに含まれる <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> を構文解析します。
                            #if (javadocForManual) {
                                 * <p>
                                 * ファイルに含まれるテキストを、非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> をゴール記号（開始記号）とする文法に基づき構文解析します。
                                 * このメソッドは、<code>$parsable</code> な非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> に対して出力されます。<code>>>[[ encodeForHTML(type.getName()) ]]<<</code> が <code>$parsable-protected</code> の場合、このメソッドは <code>protected</code> になります
                            #}
                             * 
                             * @param   sourceName
                             *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                             * @param   file
                             *      構文解析対象のテキストを含むファイル。
                             * @param   charsetName
                             *      文字セットの名前。デフォルトの文字セットを指定する場合 <code>null</code>。
                             * @param   tabStop
                             *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                             * @return
                             *      抽象構文木
                             * @throws   java.io.IOException
                             *      フェイルへのアクセス中にエラーが発生した場合。
                             * @throws   $ParseException
                             *      ファイルに含まれるテキストが、満たすべき条件を満たしていなかった場合。
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Parses a <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> in the given file.
                            #if (javadocForManual) {
                                 * <p>
                                 * This method parses a text in the given file according to the grammar whose start symbol is <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                                 * A <code>$parsable</code> nonterminal symbol <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> generates this method.  If <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> is <code>$parsable-protected</code>, this method is <code>protected</code>.
                            #}
                             * 
                             * @param   sourceName
                             *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                             * @param   file
                             *      the file including the text to parse.
                             * @param   charsetName
                             *      the charset name of the file, or <code>null</code> for the default charset.
                             * @param   tabStop
                             *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                             * @return
                             *      the abstract syntax tree.
                             * @throws   java.io.IOException
                             *      if some error is occured for the access to the file.
                             * @throws   $ParseException
                             *      if the text in the file doesn't satisfy some requirements.
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        >>[[ methodScope ]]<< >>[[ getTypeName(type) ]]<< parse>>[[ type.getName() ]]<<(java.lang.String sourceName, java.io.File file, java.lang.String charsetName, int tabStop) throws java.io.IOException, $ParseException {
                            return parse>>[[ type.getName() ]]<<(sourceName, $toCharSequence(file, charsetName), tabStop);
                        }
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * 文字ストリームに含まれる <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> を構文解析します。
                            #if (javadocForManual) {
                                 * <p>
                                 * 文字ストリームの終端に達するまで文字列を読み出し、その文字列を 非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> をゴール記号（開始記号）とする文法に基づき構文解析します。
                                 * このメソッドは、<code>$parsable</code> な非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> に対して出力されます。<code>>>[[ encodeForHTML(type.getName()) ]]<<</code> が <code>$parsable-protected</code> の場合、このメソッドは <code>protected</code> になります
                            #}
                             * 
                             * @param   sourceName
                             *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                             * @param   reader
                             *      解析すべきテキストを含む文字ストリーム。このストリームは <code>close()</code> されません。
                             * @param   tabStop
                             *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                             * @return
                             *      抽象構文木
                             * @throws   java.io.IOException
                             *      文字ストリームへのアクセス中にエラーが発生した場合。
                             * @throws   $ParseException
                             *      文字ストリームに含まれるテキストが、満たすべき条件を満たしていなかった場合。
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Parses a <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> in the given character stream.
                            #if (javadocForManual) {
                                 * <p>
                                 * This method reads characters from the stream until the end of the stream has been reached,
                                 * and parses the text according to the grammar whose start symbol is <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                                 * A <code>$parsable</code> nonterminal symbol <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> generates this method.  If <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> is <code>$parsable-protected</code>, this method is <code>protected</code>.
                            #}
                             * 
                             * @param   sourceName
                             *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                             * @param   reader
                             *      a character stream.  This method will not <code>close</code> it.
                             * @param   tabStop
                             *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                             * @return
                             *      the abstract syntax tree.
                             * @throws   java.io.IOException
                             *      if some error is occured for the access to the file.
                             * @throws   $ParseException
                             *      if the text in the file doesn't satisfy some requirements.
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        >>[[ methodScope ]]<< >>[[ getTypeName(type) ]]<< parse>>[[ type.getName() ]]<<(java.lang.String sourceName, java.io.Reader reader, int tabStop) throws java.io.IOException, $ParseException {
                            return parse>>[[ type.getName() ]]<<(sourceName, $toCharSequence(reader), tabStop);
                        }
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * <code>$CharSequence</code> に含まれる <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> を構文解析します。
                            #if (javadocForManual) {
                                 * <p>
                                 * <code>$CharSequence</code> を 非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> をゴール記号（開始記号）とする文法に基づき構文解析します。
                                 * このメソッドは、<code>$parsable</code> な非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> に対して出力されます。<code>>>[[ encodeForHTML(type.getName()) ]]<<</code> が <code>$parsable-protected</code> の場合、このメソッドは <code>protected</code> になります
                            #}
                             * 
                             * @param   sourceName
                             *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                             * @param   text
                             *      解析すべきテキストを含む <code>$CharSequence</code>。
                             * @param   tabStop
                             *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                             * @return
                             *      抽象構文木
                             * @throws   $ParseException
                             *      <code>text</code> に含まれるテキストが、満たすべき条件を満たしていなかった場合。
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Parses a <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> in the given text.
                            #if (javadocForManual) {
                                 * <p>
                                 * This method parses the <code>text</code> according to the grammar whose start symbol is <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                                 * A <code>$parsable</code> nonterminal symbol <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> generates this method.  If <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> is <code>$parsable-protected</code>, this method is <code>protected</code>.
                            #}
                             * 
                             * @param   sourceName
                             *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                             * @param   text
                             *      the text to parse.
                             * @param   tabStop
                             *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                             * @return
                             *      the abstract syntax tree.
                             * @throws   $ParseException
                             *      if the text doesn't satisfy some requirements.
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        >>[[ methodScope ]]<< >>[[ getTypeName(type) ]]<< parse>>[[ type.getName() ]]<<(java.lang.String sourceName, $CharSequence text, int tabStop) throws $ParseException {
                            $Parser.$LexicalAnalyzer analyzer = $createLexicalAnalyzer(sourceName, text, tabStop);
                            return parse>>[[ type.getName() ]]<<(analyzer);
                        }
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * 字句解析器が出力するトークンの列を構文解析します。
                            #if (javadocForManual) {
                                 * <p>
                                 * 字句解析器が出力するトークンの列を 非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> をゴール記号（開始記号）とする文法に基づき構文解析します。
                                 * このメソッドは、<code>$parsable</code> な非終端記号 <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> に対して出力されます。<code>>>[[ encodeForHTML(type.getName()) ]]<<</code> が <code>$parsable-protected</code> の場合、このメソッドは <code>protected</code> になります
                            #}
                             * 
                             * @param   analyzer
                             *      トークンの列を与える字句解析器
                             * @return
                             *      抽象構文木
                             * @throws   $ParseException
                             *      <code>text</code> に含まれるテキストが、満たすべき条件を満たしていなかった場合。
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #} else {
                            /**
                             * Parses the sequence of tokens.
                            #if (javadocForManual) {
                                 * <p>
                                 * This method parses the sequence of the tokens given by the lexical analyzer according to the grammar whose start symbol is <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                                 * A <code>$parsable</code> nonterminal symbol <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> generates this method.  If <code>>>[[ encodeForHTML(type.getName()) ]]<<</code> is <code>$parsable-protected</code>, this method is <code>protected</code>.
                            #}
                             * 
                             * @param   analyzer
                             *      the lexical analyzer that gives the sequence of tokens to parse.
                             * @return
                             *      the abstract syntax tree.
                             * @throws   $ParseException
                             *      if the sequence of tokens doesn't satisfy some requirements.
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                             */
                        #}
                        >>[[ methodScope ]]<< >>[[ getTypeName(type) ]]<< parse>>[[ type.getName() ]]<<($Parser.$LexicalAnalyzer analyzer) throws $ParseException {
                            int initialState = >>[[ tables.getIndexOfState(initialState) ]]<<;
                            
                            #if (ambiguitySupport) {
                                >>[[ getTypeName(type) ]]<< result;
                                $Parser.$Node node = $parse(analyzer, initialState);
                                if (node instanceof >>[[ getTypeName(type) ]]<<) {
                                    result = (>>[[ getTypeName(type) ]]<<) node;
                                } else {
                                    #//$assert node instanceof $Parser.$AmbiguityAbsorber : node;
                                    #ambiguousWrapperTypes.add(type);
                                    #if (targetVersion >= 1.3)
                                        result = (>>[[ getTypeName(type) ]]<<) $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) node, >>[[ getTypeName(type) ]]<<.class);
                                    #else
                                        result = $create>>[[ getSimpleTypeName(type) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) node);
                                }
                            #} else {
                                >>[[ getTypeName(type) ]]<< result = (>>[[ getTypeName(type) ]]<<) $parse(analyzer, initialState);
                            #}
                            
                            return (>>[[ getTypeName(type) ]]<<) $modifyWholeTree(result);
                        }
                    #}
                    #StringBuffer extendedTypes = new StringBuffer();
                    #{
                        #boolean first = true;
                        #Iterator inner = type.getDirectSuperTypes().iterator();
                        #while (inner.hasNext()) {
                            #Type superType = (Type) inner.next();
                            #if (first)
                                #first = false;
                            #else
                                #extendedTypes.append(", ");
                            #extendedTypes.append(getTypeName(superType));
                        #}
                    #}
                    #if (manualLocale.equals("ja")) {
                        /**
                         * 抽象構文木のノード <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                        #if (javadocForManual) {
                             * <p>
                             * このクラスは、１つの <i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> 毎に１つ出力されます。>>[[ encodeForHTML(type.getName()) ]]<< は、<i><a href="../input.html#syntaxref:TypeDefinition:IDENTIFIER" class="symbol">IDENITIFIER</a></i> で与えられます。<i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> が <code>$protected</code> か <code>$private</code> の場合、このクラスは <code>protected</code> か <code>private</code> になります。そうでない場合 <code>public</code> になります。
                        #}
                         * 
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                         */
                    #} else {
                        /**
                         * An abstract-syntax-tree node <code>>>[[ encodeForHTML(type.getName()) ]]<<</code>.
                         * 
                        #if (javadocForManual) {
                             * <p>
                             * A <i><a href="../ja/input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> generates this class.  The word >>[[ encodeForHTML(type.getName()) ]]<< is given by <i><a href="../ja/input.html#syntaxref:TypeDefinition:IDENTIFIER" class="symbol">IDENITIFIER</a></i>.  If <i><a href="../ja/input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> is <code>$protected</code> or <code>$private</code>, this class is <code>protected</code> or <code>private</code>.  Otherwise, it is <code>public</code>.
                        #}
                         * 
                            #if (javadocForManual) {
                                * @since    notavaCC 1.0
                            #}
                         */
                    #}
                    #topTypeNamespace.add(type.getName(), type);
                    >>[[ scope ]]<< static interface >>[[ type.getName() ]]<< extends >>[[ extendedTypes ]]<< {
                        #TypeNonterminal nonterminal = type.getTypeNonterminal();
                        #if (nonterminal != null) {
                            #if (manualLocale.equals("ja")) {
                                /**
                                 * この型の ID 値。
                                #if (javadocForManual) {
                                     * このフィールドは、抽象構文木にこの型のインスタンスが含まれうる場合に出力されます。
                                #}
                                 * @see     $Parser#$createNode
                                 * @since   notavaCC 1.0
                                 */
                            #} else {
                                /**
                                 * The ID value of the type.
                                #if (javadocForManual) {
                                     * This field is generated if the instance of this type may be included by the abstract syntx trees.
                                #}
                                 * @see     $Parser#$createNode
                                 * @since   notavaCC 1.0
                                 */
                            #}
                            public static final int $ID = >>[[ tables.getIndexOfSymbol(nonterminal) ]]<<;
                        #}
                        #if (useMyList) {
                            public static class ConcreteList extends java.util.LinkedList<>>[[ type.getName() ]]<<> {
                                private static final long serialVersionUID = 0;
                                public ConcreteList() {
                                }
                                public ConcreteList(java.util.Collection<? extends >>[[ type.getName() ]]<<> c) {
                                    super(c);
                                }
                            }
                        #}
                        #Iterator it2 = type.getDeclaredFields().iterator();
                        #while (it2.hasNext()) {
                            #Field field = (Field) it2.next();
                            #java.lang.String typeName = getTypeName(field.getType());
                            #java.lang.String name = field.getName();
                            #if (manualLocale.equals("ja")) {
                                #if (javadocForManual) {
                                    /**
                                     * <code>>>[[ encodeForHTML(name) ]]<<</code> でラベル付けされる子を返します。
                                     * <p>
                                     * ラベル付けされる子が高々１つしかない場合、このメソッドの戻り値の型は、ラベル付けされる子に共通の型で最も限定的なものになります。このとき、ラベル付けされる子が無ければ、戻り値は <code>null</code> になります。
                                     * <p>
                                     * ラベル付けされる子が複数になる可能性がある場合、このメソッドの戻り値の型は、ラベル付けされる子に共通の型で最も限定的なものの {@link java.util.List} になります。このとき、ラベル付けされる子が無ければ、戻り値は <code>isEmpty</code> もしくは長さ０の配列になります。
                                     * 
                                     * @return
                                     *      <code>>>[[ encodeForHTML(name) ]]<<</code> でラベル付けされる子。
                                     * @since   notavaCC 1.0
                                     */
                                #} else {
                                    #if (field.getType() instanceof ArrayType) {
                                        #ArrayType at = (ArrayType) field.getType();
                                        /**
                                         * @return
                                         *      {@link >>[[ encodeForHTML(getTypeName(at.getComponentType())) ]]<<} のリスト。
                                         */
                                    #} else {
                                        /**
                                         */
                                    #}
                                #}
                            #} else {
                                #if (javadocForManual) {
                                    /**
                                     * Returns the child nodes labeled by <code>>>[[ encodeForHTML(name) ]]<<</code> in the notavaCC source.
                                     * <p>
                                     * If the number of the labeled children should be zero or one, the type of the result is the most specific common type.  In this case, if there is no labeled children, the result is <code>null</code>.
                                     * <p>
                                     * If the number of the labeled children may be greater than one, the type of the result is the {@link java.util.List} of the most specific common type.  In this case, if there is no labeled children, the result <code>isEmpty</code> or is empty array rather than <code>null</code>.
                                     * 
                                     * @return
                                     *      the child nodes labeled by <code>>>[[ encodeForHTML(name) ]]<<</code>.
                                     * @since   notavaCC 1.0
                                     */
                                #} else {
                                    #if (field.getType() instanceof ArrayType) {
                                        #ArrayType at = (ArrayType) field.getType();
                                        /**
                                         * @return
                                         *      the list of {@link >>[[ encodeForHTML(getTypeName(at.getComponentType())) ]]<<}
                                         */
                                    #} else {
                                        /**
                                         */
                                    #}
                                #}
                            #}
                            #nodeNoArgumentMethodNamespace.add(name, name);
                            public >>[[ typeName ]]<< >>[[ name ]]<<();
                        #}
                    }
                #}
            #}
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * 解析対象のテキストが、何らかの誤りを持っていることをあらわします。
                #if (javadocForManual) {
                     * <p>
                     * ￢＜＞＜∪∪が出力されたコードがこの例外を発生させるのは、テキストが構文上の誤りを持っていた場合のみです。しかし、ユーザは、テキストが意味上の誤りを持っていた場合にこの例外を発生させてもかまいません。
                #}
                 * 
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * Describes a error of a text.
                #if (javadocForManual) {
                     * <p>
                     * The program generated by &#xAC;&lt;&gt;&lt;&#x222A;&#x222A; throws this exception only if the parsed text has some syntactical error.  But, the users of &#xAC;&lt;&gt;&lt;&#x222A;&#x222A; may throw this exception if it has some semantic error.
                #}
                 * 
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("ParseException", topTypeNamespace);
            public static class $ParseException extends java.lang.Exception {
                private static final long serialVersionUID = 0;
                private final java.lang.String sourceName;
                private final int index;
                private final int line;
                private final int column;
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * 例外を構築します。
                     * 
                     * @param   msg
                     *            エラーの内容を表すメッセージ。
                     * @param   sourceName
                     *            エラーが発生したソースの名前。サポートされない場合 <code>null</code>。
                     * @param   index
                     *            エラーが発生した位置のインデックス。サポートされない場合 <code>-1</code>。
                     * @param   line
                     *            エラーが発生した位置の行番号。サポートされない場合 <code>-1</code>。
                     * @param   column
                     *            エラーが発生した位置のカラム番号。サポートされない場合 <code>-1</code>。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Constructs an exception.
                     * 
                     * @param   msg
                     *            a message to describes the error.
                     * @param   sourceName
                     *            the source name where the exception is occured, or null if not supported.
                     * @param   index
                     *            the index number where the exception is occured, or -1 if not supported.
                     * @param   line
                     *            the line number where the exception is occured, or -1 if not supported.
                     * @param   column
                     *            the column number where the exception is occured, or -1 if not supported.
                     * @since   notavaCC 1.0
                     */
                #}
                public $ParseException(java.lang.String msg, java.lang.String sourceName, int index, int line, int column) {
                    super(msg);
                    this.sourceName = sourceName;
                    this.index = index;
                    this.line = line;
                    this.column = column;
                }
                #if (manualLocale.equals("ja")) {
                    /**
                     * 例外を構築します。
                     * 
                     * @param   msg
                     *            エラーの内容を表すメッセージ。
                     * @param   position
                     *            エラーが発生した位置を表すトークン。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Constructs an exception.
                     * 
                     * @param   msg
                     *            a message to describes the error.
                     * @param   position
                     *            a token that describes the position of the error.
                     * @since   notavaCC 1.0
                     */
                #}
                public $ParseException(java.lang.String msg, $Parser.$Token position) {
                    super(msg);
                    
                    java.lang.String sourceName = null;
                    int index = -1;
                    int line = -1;
                    int column = -1;
                    
                    try {
                        sourceName = position.getSourceName();
                    } catch (java.lang.UnsupportedOperationException x) {
                    }
                    try {
                        index = position.getIndex();
                    } catch (java.lang.UnsupportedOperationException x) {
                    }
                    try {
                        line = position.getLine();
                    } catch (java.lang.UnsupportedOperationException x) {
                    }
                    try {
                        column = position.getColumn();
                    } catch (java.lang.UnsupportedOperationException x) {
                    }
                    
                    this.sourceName = sourceName;
                    this.index = index;
                    this.line = line;
                    this.column = column;
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * エラーが発生したソースの名前を返します。
                     * 
                     * @return  エラーが発生したソースの名前。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the source name where the exception is occured.
                     * 
                     * @return  the source name where the exception is occured.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public final java.lang.String getSourceName() throws java.lang.UnsupportedOperationException {
                    if (sourceName == null)
                        throw new java.lang.UnsupportedOperationException();
                    return sourceName;
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * エラーが発生したインデックスを返します。
                     * 
                     * @return  エラーが発生したインデックス。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the index number where the exception is occured.
                     * 
                     * @return  the index number where the exception is occured.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public final int getIndex() throws java.lang.UnsupportedOperationException {
                    if (index == -1)
                        throw new java.lang.UnsupportedOperationException();
                    return index;
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * エラーが発生した行番号を返します。
                     * 
                     * @return  エラーが発生した行番号。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the line number where the exception is occured.
                     * 
                     * @return  the line number where the exception is occured.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public final int getLine() throws java.lang.UnsupportedOperationException {
                    if (line == -1)
                        throw new java.lang.UnsupportedOperationException();
                    return line;
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * エラーが発生したカラム番号を返します。
                     * 
                     * @return  エラーが発生したカラム番号。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the column number where the exception is occured.
                     * 
                     * @return  the column number where the exception is occured.
                     * @throws   java.lang.UnsupportedOperationException
                     *      if the method is not suported.
                     * @since   notavaCC 1.0
                     */
                #}
                public final int getColumn() throws java.lang.UnsupportedOperationException {
                    if (column == -1)
                        throw new java.lang.UnsupportedOperationException();
                    return column;
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * エラーの位置をあらわす、GNU Compiler Collection のスタイルの文字列を得ます。
                     * <p>
                     * このメソッドは、<code>getSourceName()</code> がサポートされない場合、{@link java.lang.UnsupportedOperationException} を発生させます。そうでない場合、<code>getLine()</code>がサポートされなければ、<code>getSourceName()</code> を返します。そうでない場合、<code>getColumn()</code>がサポートされなければ、<code>getSourceName() + ":" + getLine()</code> を返します。そうでない場合、<code>getSourceName() + ":" + getLine() + ":" + getColumn()</code> を返します。
                     * 
                     * @return  エラーの位置をあらわす、GNU Compiler Collection のスタイルの文字列。
                     * @throws   java.lang.UnsupportedOperationException
                     *      サポートされない場合。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Returns the GNU-Compiler-Collection-style string that describes the position of the error.
                     * <p>
                     * This method throws an UnsupportedOperationException if <code>getSourceName()</code> is not supported.
                     * Otherwise, it returns <code>getSourceName()</code> if <code>getLine()</code> is not supported.
                     * Otherwise, it returns <code>getSourceName() + ":" + getLine()</code> if <code>getColumn()</code> is not supported.
                     * Otherwise, it returns <code>getSourceName() + ":" + getLine() + ":" + getColumn()</code>.
                     * 
                     * @return
                     *      the GCC-style string describing the position of the error.
                     * @throws   UnsupportedOperationException
                     *      getSourceName() was not suppoted.
                     * @since   notavaCC 1.0
                     */
                #}
                public final java.lang.String getGCCStylePositionString() throws java.lang.UnsupportedOperationException {
                    if (line == -1) {
                        return getSourceName();
                    } else if (column == -1) {
                        return getSourceName() + ":" + line;
                    } else {
                        return getSourceName() + ":" + line + ":" + column;
                    }
                }
                
                #if (manualLocale.equals("ja")) {
                    /**
                     * ￢＜＞＜∪∪へ入力された文法が曖昧である場合に発生します。
                     * @since   notavaCC 1.0
                     */
                #} else {
                    /**
                     * Indicates the source of the notavaCC, which generates this Java program, has ambiguity.
                     * @since   notavaCC 1.0
                     */
                #}
                public static class AmbiguousGrammarError extends $AssertionError {
                    private static final long serialVersionUID = 0;
                    public AmbiguousGrammarError(java.lang.String message) {
                        super(message);
                    }
                }
            } // class $ParseException
            
            
            #if (manualLocale.equals("ja")) {
                /**
                 * デフォルトの実装を保持する名前空間です。
                 * 
                 * @since   notavaCC 1.0
                 */
            #} else {
                /**
                 * A namespace for the default implementations.
                 * 
                 * @since   notavaCC 1.0
                 */
            #}
            #reserve("Default", topTypeNamespace);
            public static interface $Default {
                #{
                    #if (manualLocale.equals("ja")) {
                        /**
                         * {@link $Parser.$Node} の簡単な実装を提供します。
                         * 
                         * @since   notavaCC 1.0
                         */
                    #} else {
                        /**
                         * Provides a simple implementation of {@link $Parser.$Node}.
                         * 
                         * @since   notavaCC 1.0
                         */
                    #}
                    public static abstract class $Node implements $Parser.$Node {
                        #if (targetVersion >= 1.5) {
                            private static final java.util.List<$Parser.$Node> EMPTY_LIST = new java.util.ArrayList<$Parser.$Node>();
                        #}
                        
                        private java.util.List$<$Parser.$Node> childNodes;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * インスタンスを構築します。
                             * <p>
                             * <code>childNodes</code> は、{@link #$getChildNodes()} の戻り値として、そのまま使われます。
                             * 
                             * @param   childNodes
                             *            子のノードのリスト。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constucts an instance.
                             * <p>
                             * The returned list of {@link #$getChildNodes()} is the <code>childNodes</code> rather than a copy of it.
                             * 
                             * @param   childNodes
                             *            a list of child nodes.
                             * @since   notavaCC 1.0
                             */
                        #}
                        public $Node(java.util.List$<$Parser.$Node> childNodes) {
                            this.childNodes = childNodes;
                        }
                        public java.util.List$<$Parser.$Node> $getChildNodes() {
                            return this.childNodes;
                        }
                        
                        #if (ambiguitySupport) {
                            public $Parser.$AmbiguityAbsorber $getWrappedAmbiguityAbsorber() {
                                return null;
                            }
                        #}
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * 子のノード <code>oldChild</code> を <code>newChild</code> で置き換えます。
                             * <p>
                             * このメソッドは、
                             * <code>$getChildNodes()</code> の中の <code>oldChild</code> を <code>newChild</code> で置き換えます。
                            #if (javadocForManual) {
                                 * また、{@link $Parser.$Type#$label()} によって返されるノードも置き換えます。
                            #} else {
                                 * また、{@link $Parser.$Node} のサブインターフェイスで定義される、特定の子を返すメソッドによって返されるノードも置き換えます。
                            #}
                             * 
                             * @param   oldChild
                             *            交換される子。
                             * @param   newChild
                             *            <code>oldChild</code> を置き換える新しい子。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Replaces the child node <code>oldChild</code> with <code>newChild</code>.
                             * <p>
                             * This method
                             * replaces <code>oldChild</code> in <code>$getChildNodes()</code> with <code>newChild</code>.
                            #if (javadocForManual) {
                                 * Also, this method replaces the values returned by {@link $Parser.$Type#$label()}.
                            #} else {
                                 * Also, this method replaces the values returned by the methods that are defined in the subinterfaces of {@link $Parser.$Node} and that return particular children.
                            #}
                             * 
                             * @param   oldChild
                             *            a replaced child node.
                             * @param   newChild
                             *            a child node to replace <code>oldChild</code>.
                             * @since   notavaCC 1.0
                             */
                        #}
                        #reserve("replaceChild", nodeMethodNamespace);
                        public void $replaceChild($Parser.$Node oldChild, $Parser.$Node newChild) {
                            $replaceAll($getChildNodes(), oldChild, newChild);
                        }
                        
                        public java.lang.String toString() {
                            java.io.StringWriter buffer = new java.io.StringWriter();
                            java.io.PrintWriter writer = new java.io.PrintWriter(buffer);
                            
                            $Parser.$Node node = this;
                            while (!(node instanceof $Parser.$Token)) {
                                if (node.getChildNodes().isEmpty()) {
                                    node = null;
                                    break;
                                }
                                node = ($Parser.$Node) node.getChildNodes().get(0);
                            }
                            if (node != null) {
                                $Parser.$Token token = ($Parser.$Token) node;
                                try {
                                    writer.print(token.getSourceName() + ":");
                                    writer.print(token.getLine() + ":");
                                    writer.println(token.getOriginalImage());
                                } catch (java.lang.UnsupportedOperationException x) {
                                }
                            }
                            
                            toString(this, 0, writer);
                            writer.close();
                            if (writer.checkError())
                                throw new java.lang.RuntimeException();
                            return buffer.toString();
                        }
                        private static void toString($Parser.$Node node, int indent, java.io.PrintWriter writer) {
                            for (int i = 0; i < indent; i++)
                                writer.print("  ");
                            
                            if (node == null) {
                                writer.println("null");
                            } else if (node instanceof $Token) {
                                $Token token = ($Token) node;
                                writer.println(token);
                                //writer.println("token: " + token.getOriginalImage());
                            } else {
                                java.lang.String binaryName = node.getClass().getName();
                                binaryName = binaryName.substring(binaryName.lastIndexOf(".") + 1);
                                binaryName = binaryName.substring(binaryName.lastIndexOf("$") + 1);
                                writer.println(binaryName);
                                indent++;
                                java.util.Iterator it = node.$getChildNodes().iterator();
                                while (it.hasNext()) {
                                    $Parser.$Node n = ($Parser.$Node) it.next();
                                    toString(n, indent, writer);
                                }
                            }
                        }
                    } // $Default.$Node
                    
                    #if (manualLocale.equals("ja")) {
                        /**
                         * {@link $Parser.$Token} の簡単な実装を提供します。
                         * 
                         * @since   notavaCC 1.0
                         */
                    #} else {
                        /**
                         * Provides a simple implementation of {@link $Parser.$Token}.
                         * 
                         * @since   notavaCC 1.0
                         */
                    #}
                    public static class $Token extends $Default.$Node implements $Parser.$Token {
                        private final int symbolID;
                        private final java.lang.String image;
                        private final java.lang.String originalImage;
                        private final java.lang.String sourceName;
                        private final int index;
                        private final int line;
                        private final int column;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * トークンを構築します。
                             * <p>
                             * このコンストラクタで構築したトークンでは、<code>getSourceName</code>、<code>getIndex</code>、<code>getLine</code>、<code>getColumn</code> はサポートされません。
                             * 
                             * @param   symbolID
                             *      <code>getSymbolID</code> で返される値。
                             * @param   image
                             *      <code>getImage</code> と <code>getOriginalImage</code> で返される値。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constructs a token.
                             * <p>
                             * 
                             * @param   symbolID
                             *      the symbol ID of the token, returned by <code>getSymbolID</code>.
                             * @param   image
                             *      the image string, returned by <code>getImage</code> and <code>getOriginalImage</code>.
                             * @since   notavaCC 1.0
                             */
                        #}
                        public $Token(int symbolID, java.lang.String image) {
                            this(symbolID, image, image, null, -1, -1, -1);
                        }
                        
                        #if (manualLocale.equals("ja")) {
                            /**
                             * トークンを構築します。
                             * <p>
                             * 
                             * @param   symbolID
                             *      <code>getSymbolID</code> で返される値。
                             * @param   image
                             *      <code>getImage</code> で返される値。
                             * @param   originalImage
                             *      <code>getOriginalImage</code> で返される値。
                             * @param   sourceName
                             *      <code>getSourceName</code> で返される値。<code>getSourceName</code> がサポートされない場合 <code>null</code>。
                             * @param   index
                             *      <code>getIndex</code> で返される値。<code>getIndex</code> がサポートされない場合 <code>-1</code>。
                             * @param   line
                             *      <code>getLine</code> で返される値。<code>getLine</code> がサポートされない場合 <code>-1</code>。
                             * @param   column
                             *      <code>getColumn</code> で返される値。<code>getColumn</code> がサポートされない場合 <code>-1</code>。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constructs a token.
                             * <p>
                             * 
                             * @param   symbolID
                             *      the symbol ID of the token, returned by <code>getSymbolID</code>.
                             * @param   image
                             *      the image string, returned by <code>getImage</code>.
                             * @param   originalImage
                             *      the image string, returned by <code>getOriginalImage</code>.
                             * @param   sourceName
                             *      the source name returned by <code>getSourceName</code>, or <code>null</code> if not supported.
                             * @param   index
                             *      the index returned by <code>getIndex</code>, or <code>-1</code> if not supported.
                             * @param   line
                             *      the line number returned by <code>getLine</code>, or <code>-1</code> if not supported.
                             * @param   column
                             *      the column number returned by <code>getColumn</code>, or <code>-1</code> if not supported.
                             * @since   notavaCC 1.0
                             */
                        #}
                        public $Token(int symbolID, java.lang.String image, java.lang.String originalImage, java.lang.String sourceName, int index, int line, int column) {
                            #if (targetVersion >= 1.5)
                                super($Parser.$Default.$Node.EMPTY_LIST);
                            #else
                                super(java.util.Collections.EMPTY_LIST);
                            this.symbolID = symbolID;
                            this.image = image;
                            this.originalImage = originalImage;
                            this.sourceName = sourceName;
                            this.index = index;
                            this.line = line;
                            this.column = column;
                        }
                        #if (manualLocale.equals("ja")) {
                            /**
                             * トークンを構築します。
                             * <p>
                             * 
                             * @param   symbolID
                             *      <code>getSymbolID</code> で返される値。
                             * @param   image
                             *      <code>getImage</code> で返される値。
                             * @param   originalImage
                             *      <code>getOriginalImage</code> で返される値。
                             * @param   token
                             *      <code>getSourceName</code>、<code>getIndex</code>、<code>getLine</code>、<code>getColumn</code> で返される値を与えるトークン。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constructs a token.
                             * <p>
                             * 
                             * @param   symbolID
                             *      the symbol ID of the token, returned by <code>getSymbolID</code>.
                             * @param   image
                             *      the image string, returned by <code>getImage</code>.
                             * @param   originalImage
                             *      the image string, returned by <code>getOriginalImage</code>.
                             * @param   token
                             *      a token that gives the returned value of <code>getSourceName</code>, <code>getIndex</code>, <code>getLine</code> and <code>getColumn</code>.
                             * @since   notavaCC 1.0
                             */
                        #}
                        public $Token(int symbolID, java.lang.String image, java.lang.String originalImage, $Parser.$Token token) {
                            #if (targetVersion >= 1.5)
                                super($Parser.$Default.$Node.EMPTY_LIST);
                            #else
                                super(java.util.Collections.EMPTY_LIST);
                            java.lang.String sourceName = null;
                            int index = -1;
                            int line = -1;
                            int column = -1;
                            try {
                                sourceName = token.getSourceName();
                            } catch (java.lang.UnsupportedOperationException x) {
                            }
                            try {
                                index = token.getIndex();
                            } catch (java.lang.UnsupportedOperationException x) {
                            }
                            try {
                                line = token.getLine();
                            } catch (java.lang.UnsupportedOperationException x) {
                            }
                            try {
                                column = token.getColumn();
                            } catch (java.lang.UnsupportedOperationException x) {
                            }
                            
                            this.symbolID = symbolID;
                            this.image = image;
                            this.originalImage = originalImage;
                            this.sourceName = sourceName;
                            this.index = index;
                            this.line = line;
                            this.column = column;
                        }
                        public int getSymbolID() {
                            return symbolID;
                        }
                        public boolean isWhite() {
                            return >>[[ tables.firstIndexOfWhiteTerminals() ]]<< <= symbolID && symbolID < >>[[ tables.superiorIndexOfWhiteTerminals() ]]<<;
                        }
                        public java.lang.String getImage() {
                            return image;
                        }
                        public java.lang.String getOriginalImage() {
                            return originalImage;
                        }
                        
                        
                        public java.lang.String getSourceName() throws java.lang.UnsupportedOperationException {
                            if (sourceName == null)
                                throw new java.lang.UnsupportedOperationException();
                            return sourceName;
                        }
                        public int getIndex() throws java.lang.UnsupportedOperationException {
                            if (index == -1)
                                throw new java.lang.UnsupportedOperationException();
                            return index;
                        }
                        public int getLine() throws java.lang.UnsupportedOperationException {
                            if (line == -1)
                                throw new java.lang.UnsupportedOperationException();
                            return line;
                        }
                        public int getColumn() throws java.lang.UnsupportedOperationException {
                            if (column == -1)
                                throw new java.lang.UnsupportedOperationException();
                            return column;
                        }
                        
                        public java.lang.String toString() {
                            //$StringBuffer result = new $StringBuffer($quoted(getOriginalImage()));
                            
                            $StringBuffer result = new $StringBuffer();
                            result.append("`");
                            java.lang.String original = getOriginalImage();
                            for (int i = 0; i < original.length(); i++) {
                                char ch = original.charAt(i);
                                switch (ch) {
                                default:
                                    if (ch < ' ' || ch == '\u007F') {
                                        result.append("\\u").append($toHexCode(ch));
                                    } else {
                                        result.append(ch);
                                    }
                                    break;
                                case '\n':   result.append("\\n");      break;
                                case '\t':   result.append("\\t");      break;
                                case '\b':   result.append("\\b");      break;
                                case '\r':   result.append("\\r");      break;
                                case '\f':   result.append("\\f");      break;
                                //case '\\':   result.append("\\\\");     break;
                                //case '\'':   result.append("\\\'");     break;
                                //case '\"':   result.append("\\\"");     break;
                                }
                            }
                            result.append("'");
                            
                            if (column != -1 || line != -1) {
                                result.append(" (");
                                if (line != -1) {
                                    result.append("line ");
                                    result.append(getLine());
                                    if (column != -1)
                                        result.append(", ");
                                }
                                if (column != -1) {
                                    result.append("column ");
                                    result.append(getColumn());
                                }
                                result.append(")");
                            }
                            return result.toString();
                        }
                    } // class $Default.$Token
                    
                    #if (manualLocale.equals("ja")) {
                        /**
                         * デフォルトの {@link $Parser.$LexicalAnalyzer}。
                         * 
                         * @since   notavaCC 1.0
                         */
                    #} else {
                        /**
                         * The default {@link $Parser.$LexicalAnalyzer}.
                         * 
                         * @since   notavaCC 1.0
                         */
                    #}
                    public static class $LexicalAnalyzer extends $Parser.$LexicalAnalyzer {
                        #if (manualLocale.equals("ja")) {
                            /**
                             * コンストラクタで渡された <code>sourceName</code>。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The <code>sourceName</code> given to the constructor.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected final java.lang.String sourceName;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * コンストラクタで渡された <code>text</code>。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The <code>text</code> given to the constructor.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected final $CharSequence text;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * {@link #text} の字句解析の対象範囲の終端のインデックス（exclusive）。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The exclusive end index of the target {@link #text} region.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected final int end;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * コンストラクタで渡された <code>tabStop</code>。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The <code>tabStop</code> given to the constructor.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected final int tabStop;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * コンストラクタへ渡された <code>table</code>。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The <code>table</code> given to the constructor.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        private final int[][] table;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * コンストラクタへ渡された <code>initialState</code>。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The <code>initialState</code> given to the constructor.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        private final int initialState;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * {@linkplain #next() 次}のトークンが始まるインデックス（zero based）。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The zero-based index the {@linkplain #next() next} token begins at.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected int index;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * {@linkplain #next() 次}のトークンが始まる行番号（zero based）。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The zero-based line number the {@linkplain #next() next} token begins at.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected int line;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * {@linkplain #next() 次}のトークンが始まるカラム番号（zero based）。
                             * 
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * The zero-based column number the {@linkplain #next() next} token begins at.
                             * 
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected int column;
                        #if (manualLocale.equals("ja")) {
                            /**
                             * 字句解析器を構築します。
                             * <p>
                             * 
                             * @param   sourceName
                             *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                             * @param   text
                             *      解析すべきテキスト
                             * @param   tabStop
                             *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                             * @since    notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constructs a lexical analyzer.
                             * <p>
                             * 
                             * @param   sourceName
                             *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                             * @param   text
                             *      the text to parse.
                             * @param   tabStop
                             *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                             * @since    notavaCC 1.0
                             */
                        #}
                        public $LexicalAnalyzer(java.lang.String sourceName, $CharSequence text, int tabStop) {
                            this(sourceName, text, tabStop, 0, text.length(), 0, 0);
                        }
                        #if (manualLocale.equals("ja")) {
                            /**
                             * テキストの一部分を解析する字句解析器を構築します。
                             * <p>
                             * 
                             * @param   sourceName
                             *      解析すべきテキストの名前。これは、{@link $Parser.$Token#getSourceName()} の戻り値を設定するためだけに使用されます。
                             * @param   text
                             *      解析すべきテキスト
                             * @param   tabStop
                             *      テキスト中で、タブ文字に対して何文字ごとの位置に移動すべきか。これは、{@link $Parser.$Token#getColumn()} の戻り値を設定するためだけに使用されます。
                             * @param   start
                             *      解析対象の領域の開始位置（inclusive）。
                             * @param   end
                             *      解析対象の領域の終了位置（exclusive）。
                             * @param   line
                             *      開始位置の行番号（zero-based）。
                             * @param   column
                             *      開始位置のカラム番号（zero-based）。
                             * @since    notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Constructs a lexical analyzer that analyze a part of a text.
                             * <p>
                             * 
                             * @param   sourceName
                             *      the name of the source, used only to set the result of {@link $Parser.$Token#getSourceName()}.
                             * @param   text
                             *      the text to parse.
                             * @param   tabStop
                             *      the number that the tab character advances the cursor to the next column that is a multiple of.  <code>tabStop</code> is used only to set the result of {@link $Parser.$Token#getColumn()}.
                             * @param   start
                             *      the inclusive start index of the text region to analyze.
                             * @param   end
                             *      the exclusive end index of the text region to analyze.
                             * @param   line
                             *      the zero-based line number of the <code>start</code> index.
                             * @param   column
                             *      the zero-based column number of the <code>start</code> index.
                             * @since    notavaCC 1.0
                             */
                        #}
                        public $LexicalAnalyzer(java.lang.String sourceName, $CharSequence text, int tabStop, int start, int end, int line, int column) {
                            this(sourceName, text, tabStop, start, end, line, column, $Parser.$dfaTable, >>[[ tables.getIndexOfState(dfa.initialState()) ]]<<);
                        }
                        private $LexicalAnalyzer(java.lang.String sourceName, $CharSequence text, int tabStop, int start, int end, int line, int column, int[][] table, int initialState) {
                            this.sourceName = sourceName;
                            this.text = text;
                            this.end = end;
                            this.tabStop = tabStop;
                            this.table = table;
                            this.initialState = initialState;
                            this.index = start;
                            this.line = line;
                            this.column = column;
                        }
                        public $Parser.$Token next() throws $ParseException {
                            if (index >= end)
                                return new $Parser.$Default.$Token(EOF_TOKEN, "", "", sourceName, index, -1, -1);
                            
                            int state = initialState;
                            
                            final int startIndex = index;
                            final int startLine = line;
                            final int startColumn = column;
                            int lastHitID = -1;
                            int lastHitIndex = -1;
                            int lastHitLine = -1;
                            int lastHitColumn = -1;
                            int lastHitImageLength = 0;
                            $StringBuffer image = new $StringBuffer();
                            
                            try {
                            dfa:
                                while (index < end) {
                                    char ch = nextChar();
                                    image.append(ch);
                                    
                                    int[] map = table[state];
                                    int base = 1;
                                    int width = (map.length - 1) / 3;
                                    #java.lang.String dfaLinearSearchThreshold = environment.parameters.getProperty("JavaCodeGenerator.dfaLinearSearchThreshold", "3");
                                    while (width > >>[[ dfaLinearSearchThreshold ]]<<) {
                                        int halfWidth = width >> 1;
                                        int mid = base + halfWidth * 3;
                                        if (ch >= map[mid]) {
                                            base = mid;
                                            width -= halfWidth;
                                        } else {
                                            width = halfWidth;
                                        }
                                    }
                                    int i = 0;
                                    for (;;) {
                                        if (i++ >= width)
                                            break dfa;
                                        if (map[base] <= ch & ch <= map[base + 1]) {
                                            state = map[base + 2];
                                            break;
                                        }
                                        base += 3;
                                    }
                                    int hitID = table[state][0];
                                    if (hitID != -1) {
                                        lastHitID = hitID;
                                        lastHitIndex = index;
                                        lastHitLine = line;
                                        lastHitColumn = column;
                                        lastHitImageLength = image.length();
                                    }
                                }
                                
                                if (lastHitID == -1) {
                                    throw new $ParseException("An invalid character sequence starting with '" + text.$subSequence(startIndex, index) + "' (line " + (startLine + 1) + ", column " + (startColumn + 1) + ")", sourceName, startIndex, startLine + 1, startColumn + 1);
                                }
                            } finally {
                                index = startIndex;
                                line = startLine;
                                column = startColumn;
                            }
                            
                            $Parser.$Token result = new $Parser.$Default.$Token(lastHitID, image.substring(0, lastHitImageLength).intern(), text.$subSequence(startIndex, lastHitIndex).toString().intern(), sourceName, startIndex, startLine + 1, startColumn + 1);
                            
                            index = lastHitIndex;
                            line = lastHitLine;
                            column = lastHitColumn;
                            
                            return result;
                        } // next()
                        #if (manualLocale.equals("ja")) {
                            /**
                             * {@link #text} 中の次の文字を返します。
                             * <p>
                             * このメソッドは、{@link #index} から始まる１文字を返します。このメソッドは {@link #index}、{@link #line}、{@link #column} を、次の文字をさすように更新しなければなりません。このメソッドは、ユニコード・エスケープのように、文字列を文字として扱いたい場合にオーバーライドされます。
                             * 
                             * @return  次の文字。
                             * @throws  java.lang.IndexOutOfBoundsException
                             *          <code>index >= end</code> の場合。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Returns the next character in the {@link #text}.
                             * <p>
                             * This method returns the character starting at {@link #index} in the sequence and updates {@link #index}, {@link #line}, and {@link #column} to point the next character of the returned character.  This method is overridden in order to deal with a string as a character through the input sequence like Unicode escapes of Java.
                             * 
                             * @return  the next character in the sequence.
                             * @throws  java.lang.IndexOutOfBoundsException  if <code>index >= end</code>
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected char nextChar() throws $ParseException, java.lang.IndexOutOfBoundsException {
                            return nextExactChar();
                        }
                        #if (manualLocale.equals("ja")) {
                            /**
                             * ユニコード・エスケープをふまえて{@link #text} 中の次の文字を返します。
                             * 
                             * @return  次の文字。
                             * @throws  ParseException
                             *              無効なユニコード・エスケープが存在した場合。
                             * @throws  java.lang.IndexOutOfBoundsException
                             *              <code>index >= end</code> の場合。
                             * @since   notavaCC 1.0
                             */
                        #} else {
                            /**
                             * Returns the next character in the {@link #text} based on the Unicode escapes.
                             * 
                             * @return  the next character in the sequence.
                             * @throws  ParseException
                             *              for a illegal escape.
                             * @throws  java.lang.IndexOutOfBoundsException
                             *              if <code>index >= end</code>
                             * @since   notavaCC 1.0
                             */
                        #}
                        protected final char nextUnicodeEscapedChar() throws ParseException, java.lang.IndexOutOfBoundsException {
                            int startIndex = this.index;
                            int startLine = line;
                            int startColumn = column;
                            {
                                int index = startIndex;
                                if (text.charAt(index) == '\\' && index + 1 < end && text.charAt(index + 1) == 'u') {
                                    int count = 0;
                                    while (index >= 0 && text.charAt(index--) == '\\')
                                        count++;
                                    if (count % 2 == 0)
                                        return nextExactChar();
                                } else {
                                    return nextExactChar();
                                }
                            }
                            
                            char ch = nextExactChar();
                            $assert (ch =='\\');
                            ch = nextExactChar();
                            $assert (ch == 'u');
                            try {
                                do {
                                    ch = nextExactChar();
                                } while (ch == 'u');
                                $StringBuffer number = new $StringBuffer(4);
                                number.append(ch);
                                number.append(nextExactChar());
                                number.append(nextExactChar());
                                number.append(nextExactChar());
                                int code = java.lang.Integer.parseInt(number.toString(), 16);
                                return (char) code;
                            } catch (java.lang.NumberFormatException x) {
                            } catch (java.lang.IndexOutOfBoundsException x) {
                            }
                            throw new ParseException("illigal Unicode escape: " + text.$subSequence(startIndex, index), sourceName, startIndex, startLine, startColumn);
                        }
                        private final char nextExactChar() throws $ParseException, java.lang.IndexOutOfBoundsException {
                            if (index >= end)
                                throw new java.lang.IndexOutOfBoundsException();
                            char ch = text.charAt(index++);
                            switch (ch) {
                            case '\r':
                            case '\n':
                                if (ch == '\r') {
                                    if (index < end && text.charAt(index) == '\n')
                                        break;
                                }
                                line++;
                                column = 0;
                                break;
                            case '\t':
                                if (tabStop >= 1)
                                    column = column - (column % tabStop) + tabStop;
                                break;
                            default:
                                column++;
                                break;
                            }
                            return ch;
                        }
                    } // class $Default.$LexicalAnalyzer
                    
                    
                    #if (ambiguitySupport) {
                        public static class $AmbiguityAbsorber extends $Default.$Node implements $Parser.$AmbiguityAbsorber {
                            public $AmbiguityAbsorber(java.util.List$<$Parser.$Node> ambiguityNodes) {
                                super(ambiguityNodes);
                            }
                            public $Parser.$AmbiguityAbsorber $getWrappedAmbiguityAbsorber() {
                                return this;
                            }
                        }
                        
                        #Iterator it38 = tables.userDefinedAbsorberTypes().iterator();
                        #while (it38.hasNext()) {
                            #UserDefinedAbsorberType type = (UserDefinedAbsorberType) it38.next();
                            #Type superType = null;
                            #Iterator it39 = type.getDirectSuperTypes().iterator();
                            #while (it39.hasNext()) {
                                #Type tmp = (Type) it39.next();
                                #if (tmp instanceof UserDefinedAbsorberType || tmp instanceof AbsorberRootType) {
                                    #superType = tmp;
                                    #break;
                                #}
                            #}
                            
                            #String superTypeString;
                            #{
                                #if (superType instanceof AbsorberRootType) {
                                    #superTypeString = "$AmbiguityAbsorber";
                                #} else {
                                    #superTypeString = ((UserDefinedAbsorberType) superType).getName();
                                #}
                            #}
                            public static class >>[[ type.getName() ]]<< extends >>[[ superTypeString ]]<< implements $Parser.>>[[ type.getName() ]]<< {
                                #Iterator it33 = type.getMapNameToField().values().iterator();
                                #while (it33.hasNext()) {
                                    #Field field = (Field) it33.next();
                                    #Type fieldType = field.getType();
                                    #Field superField = superType.getField(field.getName());
                                    #if (!(field.getOwner() instanceof UserDefinedAbsorberType)) {
                                        #if (superField == null || !fieldType.equals(superField.getType())) {
                                            public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                throw new java.lang.UnsupportedOperationException();
                                            }
                                        #}
                                    #} else {
                                        #if (superField == null) {
                                            private >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<;
                                            public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                return >>[[ field.getName() ]]<<;
                                            }
                                        #} else if (!fieldType.equals(superField.getType())) {
                                            #if (useMyList && fieldType instanceof ArrayType) {
                                                #Type componentType = ((ArrayType) fieldType).getComponentType();
                                                public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                    return >>[[ getTypeName(componentType) ]]<<.ConcreteList.class.cast(super.>>[[ field.getName() ]]<<()); // TODO: Java1.5正式版で修正
                                                }
                                            #} else {
                                                public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                    return (>>[[ getTypeName(fieldType) ]]<<) super.>>[[ field.getName() ]]<<();
                                                }
                                            #}
                                        #}
                                    #}
                                #}
                                
                                #StringBuffer constructorParameters = new StringBuffer();
                                #Iterator it34 = type.getMapNameToField().values().iterator();
                                #while (it34.hasNext()) {
                                    #Field field = (Field) it34.next();
                                    #if (field.getOwner() instanceof UserDefinedAbsorberType) {
                                        #Type fieldType = field.getType();
                                        #constructorParameters.append(", ");
                                        #constructorParameters.append(getTypeName(fieldType));
                                        #constructorParameters.append(" ");
                                        #constructorParameters.append(field.getName() + "Arg");
                                    #}
                                #}
                                #StringBuffer superArguments = new StringBuffer();
                                #Iterator it32 = superType.getMapNameToField().values().iterator();
                                #while (it32.hasNext()) {
                                    #Field field = (Field) it32.next();
                                    #if (field.getOwner() instanceof UserDefinedAbsorberType) {
                                        #Type fieldType = field.getType();
                                        #superArguments.append(", ");
                                        #superArguments.append(field.getName() + "Arg");
                                    #}
                                #}
                                public >>[[ type.getName() ]]<<(java.util.List$<$Parser.$Node> ambiguityNodes>>[[ constructorParameters ]]<<) {
                                    super(ambiguityNodes>>[[ superArguments ]]<<);
                                    
                                    #Iterator it35 = type.getMapNameToField().values().iterator();
                                    #while (it35.hasNext()) {
                                        #Field field = (Field) it35.next();
                                        #if (field.getOwner() instanceof UserDefinedAbsorberType) {
                                            #Field superField = superType.getField(field.getName());
                                            #if (superField == null) {
                                                this.>>[[ field.getName() ]]<< = >>[[ field.getName() ]]<<Arg;
                                            #}
                                        #}
                                    #}
                                }
                                
                                private static $Parser.$AmbiguityAbsorber tryToCreate(java.util.List$<$Parser.$Node> nodes) {
                                    java.util.List$<$Parser.$Node> uneatenNodes = new java.util.LinkedList$<$Parser.$Node>(nodes);
                                    #Iterator it19 = type.getDeclaredFields().iterator();
                                    #while (it19.hasNext()) {
                                        #Field field = (Field) it19.next();
                                        #Type fieldType = field.getType();
                                        #if (fieldType instanceof ArrayType) {
                                            #Type componentType = ((ArrayType) fieldType).getComponentType();
                                            #if (useMyList) {
                                                >>[[ getTypeName(componentType) ]]<<.ConcreteList >>[[ field.getName() ]]<<Tmp = new >>[[ getTypeName(componentType) ]]<<.ConcreteList();
                                            #} else {
                                                java.util.List$<>>[[ getTypeName(componentType) ]]<<> >>[[ field.getName() ]]<<Tmp = new java.util.LinkedList$<>>[[ getTypeName(componentType) ]]<<>();
                                            #}
                                            {
                                                java.util.Iterator it = uneatenNodes.iterator();
                                                while (it.hasNext()) {
                                                    $Parser.$Node node = ($Parser.$Node) it.next();
                                                    if (node instanceof >>[[ getTypeName(componentType) ]]<<) {
                                                        >>[[ field.getName() ]]<<Tmp.add((>>[[ getTypeName(componentType) ]]<<) node);
                                                        it.remove();
                                                    }
                                                }
                                            }
                                        #} else {
                                            >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<Tmp = null;
                                            {
                                                java.util.Iterator it = uneatenNodes.iterator();
                                                while (it.hasNext()) {
                                                    $Parser.$Node node = ($Parser.$Node) it.next();
                                                    if (node instanceof >>[[ getTypeName(fieldType) ]]<<) {
                                                        >>[[ field.getName() ]]<<Tmp = (>>[[ getTypeName(fieldType) ]]<<) node;
                                                        it.remove();
                                                        break;
                                                    }
                                                }
                                            }
                                        #}
                                    #}
                                    if (!uneatenNodes.isEmpty())
                                        return null;
                                    
                                    #StringBuffer constructorArguments = new StringBuffer();
                                    #Iterator it36 = type.getMapNameToField().values().iterator();
                                    #while (it36.hasNext()) {
                                        #Field field = (Field) it36.next();
                                        #if (field.getOwner() instanceof UserDefinedAbsorberType) {
                                            #Type fieldType = field.getType();
                                            #constructorArguments.append(", ");
                                            #if (field.getOwner().equals(type)) {
                                                #if (fieldType instanceof ArrayType && arrayRatherThanList) {
                                                    #String variable = field.getName() + "Tmp";
                                                    #String componentType = getTypeName(((ArrayType) fieldType).getComponentType());
                                                    #constructorArguments.append("(" + componentType + "[]) " + variable + ".toArray(new " + componentType + "[" + variable + ".size()])");
                                                #} else {
                                                    #constructorArguments.append(field.getName() + "Tmp");
                                                #}
                                            #} else {
                                                #if (fieldType instanceof ArrayType) {
                                                    #Type componentType = ((ArrayType) fieldType).getComponentType();
                                                    #if (arrayRatherThanList) {
                                                        #constructorArguments.append("new " + getTypeName(componentType) + "[0]");
                                                    #} else {
                                                        #if (targetVersion >= 1.5) {
                                                            #constructorArguments.append("java.util.Collections.<" + getTypeName(componentType) + ">emptyList()");
                                                        #} else {
                                                            #constructorArguments.append("java.util.Collections.EMPTY_LIST");
                                                        #}
                                                    #}
                                                #} else {
                                                    #constructorArguments.append("null");
                                                #}
                                            #}
                                        #}
                                    #}
                                    return new >>[[ type.getName() ]]<<(nodes>>[[ constructorArguments ]]<<);
                                }
                            }
                        #}
                    #}
                    
                    #// user-defined types
                    #Set defaultClasses = new HashSet();
                    #Iterator outer = tables.userDefinedNodeTypes().iterator();
                    #while (outer.hasNext()) {
                        #Type type = (Type) outer.next();
                        #TypeNonterminal nonterminal = ((UserDefinedNodeType) type).getTypeNonterminal();
                        #if (nonterminal != null) {
                            #for (;;) {
                                #if (defaultClasses.add(type)) {
                                    #type = (Type) type.getDirectSuperTypes().iterator().next();
                                    #if (type instanceof UserDefinedNodeType)
                                        #continue;
                                #}
                                #break;
                            #}
                        #}
                    #}
                    #outer = tables.userDefinedNodeTypes().iterator();
                    #while (outer.hasNext()) {
                        #UserDefinedNodeType type = (UserDefinedNodeType) outer.next();
                        #if (defaultClasses.contains(type)) {
                            #boolean isAbstract = (type.getTypeNonterminal() == null);
                            #java.lang.String identifier = type.getName();
                            #Type superType = (Type) type.getDirectSuperTypes().iterator().next();
                            #String superTypeString;
                            #if (superType instanceof UserDefinedNodeType) {
                                #superTypeString = ((UserDefinedNodeType) superType).getName();
                            #} else {
                                #superTypeString = "$Node";
                            #}
                            #if (manualLocale.equals("ja")) {
                                /**
                                 * <code>>>[[ encodeForHTML(identifier) ]]<<</code> の実装。
                                #if (javadocForManual) {
                                     * <p>
                                     * このクラスは、<i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> で定義される、<code>$abstract</code>ではない非終端記号 $Type に対して出力されます。また、このクラスを出力させる <i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> の <i><a href="../input.html#syntaxref:TypeDefinition:supertypes" class="symbol">supertypes</a></i> に列挙される最初の <i><a href="../input.html#syntaxref:supertypes:TypeName" class="symbol">TypeName</a></i> に対しても同様のクラスが出力されます。
                                     * <p>
                                     * このクラスは、<code>$abstract</code> な <i><a href="../input.html#syntax:TypeDefinition" class="symbol">TypeDefinition</a></i> に対しては <code>abstract</code> になります。そうでない場合、<code>abstract</code> ではなくなります。
                                     * <p>
                                     * このクラスは、<i><a href="../input.html#syntaxref:TypeDefinition:supertypes" class="symbol">supertypes</a></i> が存在しない場合 {@link $Parser.$Default.$Node} のサブクラスになります。そうでない場合、<i><a href="../input.html#syntaxref:TypeDefinition:supertypes" class="symbol">supertypes</a></i> の最初の <i><a href="../input.html#syntaxref:supertypes:TypeName" class="symbol">TypeName</a></i> と同じ名前の {@link $Parser.$Default} のネストしたクラスのサブクラスになります。
                                #}
                                 * 
                                 * @since   notavaCC 1.0
                                 */
                            #} else {
                                /**
                                 * An implementation of <code>>>[[ encodeForHTML(identifier) ]]<<</code> node.
                                 * 
                                 * @since   notavaCC 1.0
                                 */
                            #}
                            public >>[[ isAbstract ? "abstract " : "" ]]<<static class >>[[ identifier ]]<< extends $Default.>>[[ superTypeString ]]<< implements >>[[ getTypeName(type) ]]<< {
                                #Iterator it7 = type.getMapNameToField().values().iterator();
                                #while (it7.hasNext()) {
                                    #Field field = (Field) it7.next();
                                    #Type fieldType = field.getType();
                                    #Field superField = superType.getField(field.getName());
                                    #if (superField == null) {
                                        private >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<;
                                        public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                            return >>[[ field.getName() ]]<<;
                                        }
                                    #} else if (!fieldType.equals(superField.getType())) {
                                        #if (useMyList && fieldType instanceof ArrayType) {
                                            #Type componentType = ((ArrayType) fieldType).getComponentType();
                                            public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                return >>[[ getTypeName(componentType) ]]<<.ConcreteList.class.cast(super.>>[[ field.getName() ]]<<()); // TODO: Java1.5正式版で修正
                                            }
                                        #} else {
                                            public >>[[ getTypeName(fieldType) ]]<< >>[[ field.getName() ]]<<() {
                                                return (>>[[ getTypeName(fieldType) ]]<<) super.>>[[ field.getName() ]]<<();
                                            }
                                        #}
                                    #}
                                #}
                                
                                #{
                                    #String childNodes = "childNodes";
                                    #while (type.getField(childNodes) != null) {
                                        #childNodes = childNodes + "_";
                                    #}
                                    private >>[[ identifier ]]<<(java.util.List$<$Parser.$Node> >>[[ childNodes ]]<<) {
                                        super(>>[[ childNodes ]]<<);
                                    }
                                #}
                                
                                #if (manualLocale.equals("ja")) {
                                    /**
                                    #if (javadocForManual) {
                                         * <code>source.$getChildNodes()</code> や <code>source.{@linkplain $Parser.$Type#$label $label()}</code> などによって与えられる子を持つようにノードを構築します。リストオブジェクトはコピーされます。
                                    #} else {
                                         * <code>source.$getChildNodes()</code> などによって与えられる子を持つようにノードを構築します。
                                        #if (arrayRatherThanList) {
                                             * 配列はコピーされます。
                                        #} else {
                                             * リストオブジェクトはコピーされます。
                                        #}
                                    #}
                                     * 
                                     * @param   source
                                     *            ソース。
                                     */
                                #} else {
                                    /**
                                    #if (javadocForManual) {
                                         * Constructs a instance with the child nodes given by <code>source.$getChildNodes()</code> and <code>source.{@linkplain $Parser.$Type#$label $label()}</code>.  <code>List</code> objects are copied.
                                    #} else {
                                         * Constructs a instance with the child nodes given by <code>source.$getChildNodes()</code> etc.
                                        #if (arrayRatherThanList) {
                                             * Array objects are copied.
                                        #} else {
                                             * <code>List</code> objects are copied.
                                        #}
                                    #}
                                     * 
                                     * @param   source
                                     *            the source.
                                     */
                                #}
                                public >>[[ identifier ]]<<(>>[[ getTypeName(type) ]]<< source) {
                                    #if (!resultTypeNarrowing && superType instanceof UserDefinedNodeType) {
                                        super(source);
                                    #} else {
                                        this(new java.util.LinkedList$<$Parser.$Node>(source.$getChildNodes()));
                                    #}
                                    
                                    #Iterator it18 = type.getMapNameToField().values().iterator();
                                    #while (it18.hasNext()) {
                                        #Field field = (Field) it18.next();
                                        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                            #Type fieldType = field.getType();
                                            #if (fieldType instanceof ArrayType) {
                                                #ArrayType arrayType = (ArrayType) fieldType;
                                                #if (arrayRatherThanList) {
                                                    >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<[source.>>[[ field.getName() ]]<<().length];
                                                    java.lang.System.arraycopy(source.>>[[ field.getName() ]]<<(), 0, >>[[ getFieldAccessString(type, field.getName()) ]]<<, 0, source.>>[[ field.getName() ]]<<().length);
                                                #} else {
                                                    #if (useMyList) {
                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<.ConcreteList(source.>>[[ field.getName() ]]<<());
                                                    #} else {
                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = new java.util.LinkedList$<>>[[ getTypeName(arrayType.getComponentType()) ]]<<>(source.>>[[ field.getName() ]]<<());
                                                    #}
                                                #}
                                            #} else {
                                                >>[[ getFieldAccessString(type, field.getName()) ]]<< = source.>>[[ field.getName() ]]<<();
                                            #}
                                        #}
                                    #}
                                }
                                
                                #if (manualLocale.equals("ja")) {
                                    /**
                                     * <code>parameters</code> で与えられる子を持つようにノードを構築します。
                                     * 
                                     * @param   parameters
                                     *            初期化パラメタ。
                                     * @param   compact
                                     *            <code>true</code> の場合、ラベル付けされていない子を除いてノードを作る。
                                     */
                                #} else {
                                    /**
                                     * Constructs a instance with the child nodes given by <code>parameters</code>.
                                     * 
                                     * @param   parameters
                                     *            the parameters for initialization.
                                     * @param   compact
                                     *            remove the children that are not labeled.
                                     */
                                #}
                                public >>[[ identifier ]]<<($Parser.$NodeInitializationParameters parameters, boolean compact) {
                                    #if (!resultTypeNarrowing && superType instanceof UserDefinedNodeType) {
                                        super(parameters, compact);
                                    #} else {
                                        this(new java.util.LinkedList$<$Parser.$Node>());
                                    #}
                                    
                                    $Parser.$Node[] childNodes = parameters.childNodes;
                                    java.util.List labelBitsList = parameters.labelBitsList;
                                    
                                    #Iterator it8 = type.getMapNameToField().values().iterator();
                                    #while (it8.hasNext()) {
                                        #Field field = (Field) it8.next();
                                        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                            #Type fieldType = field.getType();
                                            #if (fieldType instanceof ArrayType) {
                                                #ArrayType arrayType = (ArrayType) fieldType;
                                                #if (useMyList) {
                                                    >>[[ getTypeName(arrayType.getComponentType()) ]]<<.ConcreteList >>[[ field.getName() ]]<<Tmp = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<.ConcreteList();
                                                #} else {
                                                    java.util.List$<>>[[ getTypeName(arrayType.getComponentType()) ]]<<> >>[[ field.getName() ]]<<Tmp = new java.util.LinkedList$<>>[[ getTypeName(arrayType.getComponentType()) ]]<<>();
                                                #}
                                            #}
                                        #}
                                    #}
                                    
                                    #{
                                        #boolean shouldProcess = false;
                                        #if (!resultTypeNarrowing && superType instanceof UserDefinedNodeType) {
                                        #} else {
                                            #shouldProcess = true;
                                        #}
                                        #if (!shouldProcess) {
                                            #Iterator it91 = type.getDeclaredFields().iterator();
                                            #while (it91.hasNext()) {
                                                #Field field = (Field) it91.next();
                                                #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                    #shouldProcess = true;
                                                    #break;
                                                #}
                                            #}
                                        #}
                                        #if (shouldProcess) {
                                            java.util.Iterator it = labelBitsList.iterator();
                                            int size = labelBitsList.size();
                                            for (int i = 0; i < size; i++) {
                                                #if (labelWordCount == 1) {
                                                    java.lang.Integer number = (java.lang.Integer) it.next();
                                                    int no = (number == null) ? 0 : number.intValue();
                                                #} else {
                                                    java.math.BigInteger number = (java.math.BigInteger) it.next();
                                                #}
                                                #if (!resultTypeNarrowing && superType instanceof UserDefinedNodeType) {
                                                #} else {
                                                    #if (labelWordCount == 1) {
                                                        if (!compact || no != 0)
                                                            $getChildNodes().add(childNodes[i]);
                                                    #} else {
                                                        if (!compact || !number.equals(java.math.BigInteger.ZERO))
                                                            $getChildNodes().add(childNodes[i]);
                                                    #}
                                                #}
                                                #Iterator it9 = type.getDeclaredFields().iterator();
                                                #while (it9.hasNext()) {
                                                    #Field field = (Field) it9.next();
                                                    #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                        #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                        #int labelNo = tables.getIndexOfLabel(field.getName());
                                                        #if (labelNo >= 0) {
                                                            #java.lang.String match;
                                                            #if (labelWordCount == 1) {
                                                                #match = "(no & (1 << " + labelNo + ")) != 0";
                                                            #} else {
                                                                #match = "number.testBit(" + labelNo + ")";
                                                            #}
                                                            if (>>[[ match ]]<<) {
                                                                #if (field.getType() instanceof ArrayType) {
                                                                    #Type componentType = ((ArrayType) field.getType()).getComponentType();
                                                                    #String componentTypeName = getTypeName(componentType);
                                                                    #if (ambiguitySupport) {
                                                                        $Parser.$Node node = childNodes[i];
                                                                        if (!(node instanceof >>[[ componentTypeName ]]<<)) {
                                                                            #//$assert node instanceof $Parser.$AmbiguityAbsorber : node;
                                                                            #ambiguousWrapperTypes.add(componentType);
                                                                            #if (targetVersion >= 1.3)
                                                                                node = $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) node, >>[[ componentTypeName ]]<<.class);
                                                                            #else
                                                                                node = $create>>[[ getSimpleTypeName(componentType) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) node);
                                                                        }
                                                                        >>[[ field.getName() ]]<<Tmp.add((>>[[ componentTypeName ]]<<) node);
                                                                    #} else {
                                                                        >>[[ field.getName() ]]<<Tmp.add((>>[[ componentTypeName ]]<<) childNodes[i]);
                                                                    #}
                                                                #} else {
                                                                    #if (ambiguitySupport) {
                                                                        $Parser.$Node node = childNodes[i];
                                                                        if (!(node instanceof >>[[ fieldTypeName ]]<<)) {
                                                                            #//$assert node instanceof $Parser.$AmbiguityAbsorber : node;
                                                                            #ambiguousWrapperTypes.add(field.getType());
                                                                            #if (targetVersion >= 1.3)
                                                                                node = $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) node, >>[[ fieldTypeName ]]<<.class);
                                                                            #else
                                                                                node = $create>>[[ getSimpleTypeName(field.getType()) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) node);
                                                                        }
                                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) node;
                                                                    #} else {
                                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) childNodes[i];
                                                                    #}
                                                                #}
                                                            }
                                                        #}
                                                    #}
                                                #}
                                            }
                                        #}
                                    #}
                                    
                                    #Iterator it17 = type.getMapNameToField().values().iterator();
                                    #while (it17.hasNext()) {
                                        #Field field = (Field) it17.next();
                                        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                            #Type fieldType = field.getType();
                                            #if (fieldType instanceof ArrayType) {
                                                #ArrayType arrayType = (ArrayType) fieldType;
                                                #String componentString = getTypeName(arrayType.getComponentType());
                                                #if (arrayRatherThanList) {
                                                    >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ componentString ]]<<[]) >>[[ field.getName() ]]<<Tmp.toArray(new >>[[ componentString ]]<<[>>[[ field.getName() ]]<<Tmp.size()]);
                                                #} else {
                                                    >>[[ getFieldAccessString(type, field.getName()) ]]<< = >>[[ field.getName() ]]<<Tmp;
                                                #}
                                            #}
                                        #}
                                    #}
                                }
                                
                                #{
                                    #String childNodes = "childNodes";
                                    #while (type.getField(childNodes) != null) {
                                        #childNodes = childNodes + "_";
                                    #}
                                    #StringBuffer parameters = new StringBuffer();
                                    #Iterator it171 = type.getDeclaredFields().iterator();
                                    #while (it171.hasNext()) {
                                        #Field field = (Field) it171.next();
                                        #parameters.append(", " + getTypeName(field.getType()) + " " + field.getName());
                                    #}
                                    #if (parameters.length() != 0) {
                                        public >>[[ identifier ]]<<(java.util.List$<$Parser.$Node> >>[[ childNodes ]]<<>>[[ parameters.toString() ]]<<) {
                                            this(>>[[ childNodes ]]<<);
                                            
                                            #Iterator it22 = type.getDeclaredFields().iterator();
                                            #while (it22.hasNext()) {
                                                #Field field = (Field) it22.next();
                                                #if (field.getType() instanceof ArrayType) {
                                                    #ArrayType arrayType = (ArrayType) field.getType();
                                                    #if (arrayRatherThanList) {
                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<[>>[[ field.getName() ]]<<.length];
                                                        java.lang.System.arraycopy(>>[[ field.getName() ]]<<, 0, >>[[ getFieldAccessString(type, field.getName()) ]]<<, 0, >>[[ field.getName() ]]<<.length);
                                                    #} else {
                                                        #if (useMyList) {
                                                            >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<.ConcreteList(>>[[ field.getName() ]]<<);
                                                        #} else {
                                                            >>[[ getFieldAccessString(type, field.getName()) ]]<< = new java.util.LinkedList$<>>[[ getTypeName(arrayType.getComponentType()) ]]<<>(>>[[ field.getName() ]]<<);
                                                        #}
                                                    #}
                                                #} else {
                                                    >>[[ getFieldAccessString(type, field.getName()) ]]<< = >>[[ field.getName() ]]<<;
                                                #}
                                                
                                                #if (field.getType() instanceof ArrayType) {
                                                    #if (arrayRatherThanList) {
                                                        getChildNodes().addAll(java.util.Arrays.asList(>>[[ field.getName() ]]<<));
                                                    #} else {
                                                        getChildNodes().addAll(>>[[ field.getName() ]]<<);
                                                    #}
                                                #} else {
                                                    if (>>[[ field.getName() ]]<< != null)
                                                        getChildNodes().add(>>[[ field.getName() ]]<<);
                                                #}
                                            #}
                                            
                                            #Iterator it23 = type.getMapNameToField().values().iterator();
                                            #while (it23.hasNext()) {
                                                #Field field = (Field) it23.next();
                                                #if (type.getDeclaredField(field.getName()) == null) {
                                                    #if (field.getType() instanceof ArrayType) {
                                                        #ArrayType arrayType = (ArrayType) field.getType();
                                                        #if (arrayRatherThanList) {
                                                            >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<[0];
                                                        #} else {
                                                            #if (useMyList) {
                                                                >>[[ getFieldAccessString(type, field.getName()) ]]<< = new >>[[ getTypeName(arrayType.getComponentType()) ]]<<.ConcreteList();
                                                            #} else {
                                                                >>[[ getFieldAccessString(type, field.getName()) ]]<< = new java.util.LinkedList$<>>[[ getTypeName(arrayType.getComponentType()) ]]<<>();
                                                            #}
                                                        #}
                                                    #}
                                                #}
                                            #}
                                        }
                                    #}
                                #}
                                
                                public void $replaceChild($Parser.$Node oldChild, $Parser.$Node newChild) {
                                    #if (ambiguitySupport) {
                                        if (oldChild.$getWrappedAmbiguityAbsorber() != null)
                                            oldChild = oldChild.$getWrappedAmbiguityAbsorber();
                                        if (newChild.$getWrappedAmbiguityAbsorber() != null)
                                            newChild = newChild.$getWrappedAmbiguityAbsorber();
                                        
                                    #}
                                    #if (!resultTypeNarrowing) {
                                        super.$replaceChild(oldChild, newChild);
                                    #} else {
                                        $replaceAll($getChildNodes(), oldChild, newChild);
                                    #}
                                    
                                    #if (ambiguitySupport) {
                                        if (newChild instanceof $Parser.$AmbiguityAbsorber) {
                                            #Iterator it21 = type.getMapNameToField().values().iterator();
                                            #while (it21.hasNext()) {
                                                #Field field = (Field) it21.next();
                                                #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                    #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                    #if (field.getType() instanceof ArrayType) {
                                                        #Type componentType = ((ArrayType) field.getType()).getComponentType();
                                                        #String componentTypeName = getTypeName(componentType);
                                                        {
                                                            >>[[ componentTypeName ]]<< tmp;
                                                            if (newChild instanceof >>[[ componentTypeName ]]<<)
                                                                tmp = (>>[[ componentTypeName ]]<<) newChild;
                                                            else
                                                            #if (targetVersion >= 1.3)
                                                                tmp = (>>[[ componentTypeName ]]<<) $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild, >>[[ componentTypeName ]]<<.class);
                                                            #else
                                                                tmp = $create>>[[ getSimpleTypeName(componentType) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild);
                                                            #if (arrayRatherThanList) {
                                                                {
                                                                    $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, tmp);
                                                                }
                                                            #} else {
                                                                #if (useMyList) {
                                                                    $replaceAll(>>[[ componentTypeName ]]<<.ConcreteList.class.cast(>>[[ getFieldAccessString(type, field.getName()) ]]<<), oldChild, tmp);    // TODO: Java1.5正式版で修正
                                                                #} else {
                                                                    $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, tmp);
                                                                #}
                                                            #}
                                                        }
                                                    #}
                                                #}
                                            #}
                                            
                                            if (oldChild == null) {
                                                #Iterator it24 = type.getMapNameToField().values().iterator();
                                                #while (it24.hasNext()) {
                                                    #Field field = (Field) it24.next();
                                                    #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                        #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                        #if (!(field.getType() instanceof ArrayType)) {
                                                            if (>>[[ getFieldAccessString(type, field.getName()) ]]<< == null) {
                                                                >>[[ fieldTypeName ]]<< tmp;
                                                                if (newChild instanceof >>[[ fieldTypeName ]]<<)
                                                                    tmp = (>>[[ fieldTypeName ]]<<) newChild;
                                                                else
                                                                #if (targetVersion >= 1.3)
                                                                    tmp = (>>[[ fieldTypeName ]]<<) $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild, >>[[ fieldTypeName ]]<<.class);
                                                                #else
                                                                    tmp = $create>>[[ getSimpleTypeName(field.getType()) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild);
                                                                >>[[ getFieldAccessString(type, field.getName()) ]]<< = tmp;
                                                            }
                                                        #}
                                                    #}
                                                #}
                                            #if (ambiguitySupport) {
                                                } else if (oldChild instanceof $Parser.$AmbiguityAbsorber) {
                                                    #Iterator it10 = type.getMapNameToField().values().iterator();
                                                    #while (it10.hasNext()) {
                                                        #Field field = (Field) it10.next();
                                                        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                            #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                            #if (!(field.getType() instanceof ArrayType)) {
                                                                if (oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<) || >>[[ getFieldAccessString(type, field.getName()) ]]<< != null && oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<.$getWrappedAmbiguityAbsorber())) {
                                                                    >>[[ fieldTypeName ]]<< tmp;
                                                                    if (newChild instanceof >>[[ fieldTypeName ]]<<)
                                                                        tmp = (>>[[ fieldTypeName ]]<<) newChild;
                                                                    else
                                                                    #if (targetVersion >= 1.3)
                                                                        tmp = (>>[[ fieldTypeName ]]<<) $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild, >>[[ fieldTypeName ]]<<.class);
                                                                    #else
                                                                        tmp = $create>>[[ getSimpleTypeName(field.getType()) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild);
                                                                    >>[[ getFieldAccessString(type, field.getName()) ]]<< = tmp;
                                                                }
                                                            #}
                                                        #}
                                                    #}
                                            #}
                                            } else {
                                                #Iterator it10 = type.getMapNameToField().values().iterator();
                                                #while (it10.hasNext()) {
                                                    #Field field = (Field) it10.next();
                                                    #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                        #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                        #if (!(field.getType() instanceof ArrayType)) {
                                                            if (oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<)) {
                                                                >>[[ fieldTypeName ]]<< tmp;
                                                                if (newChild instanceof >>[[ fieldTypeName ]]<<)
                                                                    tmp = (>>[[ fieldTypeName ]]<<) newChild;
                                                                else
                                                                #if (targetVersion >= 1.3)
                                                                    tmp = (>>[[ fieldTypeName ]]<<) $createAmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild, >>[[ fieldTypeName ]]<<.class);
                                                                #else
                                                                    tmp = $create>>[[ getSimpleTypeName(field.getType()) ]]<<AmbiguousWrapper(($Parser.$AmbiguityAbsorber) newChild);
                                                                >>[[ getFieldAccessString(type, field.getName()) ]]<< = tmp;
                                                            }
                                                        #}
                                                    #}
                                                #}
                                            }
                                            return;
                                        }
                                    #}
                                    
                                    
                                    #Iterator it28 = type.getMapNameToField().values().iterator();
                                    #while (it28.hasNext()) {
                                        #Field field = (Field) it28.next();
                                        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                            #java.lang.String fieldTypeName = getTypeName(field.getType());
                                            #if (field.getType() instanceof ArrayType) {
                                                #Type componentType = ((ArrayType) field.getType()).getComponentType();
                                                #if (arrayRatherThanList) {
                                                    {
                                                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);
                                                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $assert !java.util.Arrays.asList(>>[[ getFieldAccessString(type, field.getName()) ]]<<).contains(oldChild);
                                                    }
                                                #} else {
                                                    #if (useMyList) {
                                                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $replaceAll(>>[[ getTypeName(componentType) ]]<<.ConcreteList.class.cast(>>[[ getFieldAccessString(type, field.getName()) ]]<<), oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);    // TODO: Java1.5正式版で修正
                                                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $assert !>>[[ getFieldAccessString(type, field.getName()) ]]<<.contains(oldChild);
                                                    #} else {
                                                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);
                                                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                                            $assert !>>[[ getFieldAccessString(type, field.getName()) ]]<<.contains(oldChild);
                                                    #}
                                                #}
                                            #}
                                        #}
                                    #}
                                    
                                    if (oldChild == null) {
                                        #Iterator it29 = type.getMapNameToField().values().iterator();
                                        #while (it29.hasNext()) {
                                            #Field field = (Field) it29.next();
                                            #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                #if (!(field.getType() instanceof ArrayType)) {
                                                    if (>>[[ getFieldAccessString(type, field.getName()) ]]<< == null)
                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                                #}
                                            #}
                                        #}
                                    #if (ambiguitySupport) {
                                        } else if (oldChild instanceof $Parser.$AmbiguityAbsorber) {
                                            #Iterator it10 = type.getMapNameToField().values().iterator();
                                            #while (it10.hasNext()) {
                                                #Field field = (Field) it10.next();
                                                #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                    #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                    #if (!(field.getType() instanceof ArrayType)) {
                                                        if (oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<) || >>[[ getFieldAccessString(type, field.getName()) ]]<< != null && oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<.$getWrappedAmbiguityAbsorber()))
                                                            >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                                    #}
                                                #}
                                            #}
                                    #}
                                    } else {
                                        #Iterator it10 = type.getMapNameToField().values().iterator();
                                        #while (it10.hasNext()) {
                                            #Field field = (Field) it10.next();
                                            #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                                #java.lang.String fieldTypeName = getTypeName(field.getType());
                                                #if (!(field.getType() instanceof ArrayType)) {
                                                    if (oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<))
                                                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                                #}
                                            #}
                                        #}
                                    }
                                }
                                
                                #//public void $replaceChild($Parser.$Node oldChild, $Parser.$Node newChild) {
                                #//    // TODO: If newChild is an ambiguity absorber, wrap it.
                                #//    
                                #//    #if (!resultTypeNarrowing) {
                                #//        super.$replaceChild(oldChild, newChild);
                                #//    #} else {
                                #//        $replaceAll($getChildNodes(), oldChild, newChild);
                                #//    #}
                                #//    #Iterator it10 = type.getMapNameToField().values().iterator();
                                #//    #while (it10.hasNext()) {
                                #//        #Field field = (Field) it10.next();
                                #//        #if (resultTypeNarrowing || superType.getField(field.getName()) == null) {
                                #//            #java.lang.String fieldTypeName = getTypeName(field.getType());
                                #//            #if (field.getType() instanceof ArrayType) {
                                #//                #Type componentType = ((ArrayType) field.getType()).getComponentType();
                                #//                #if (arrayRatherThanList) {
                                #//                    {
                                #//                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);
                                #//                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $assert (!java.util.Arrays.asList(>>[[ getFieldAccessString(type, field.getName()) ]]<<).contains(oldChild));
                                #//                    }
                                #//                #} else {
                                #//                    #if (useMyList) {
                                #//                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $replaceAll(>>[[ getTypeName(componentType) ]]<<.ConcreteList.class.cast(>>[[ getFieldAccessString(type, field.getName()) ]]<<), oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);    // TODO: Java1.5正式版で修正
                                #//                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $assert (!>>[[ getFieldAccessString(type, field.getName()) ]]<<.contains(oldChild));
                                #//                    #} else {
                                #//                        if (newChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $replaceAll(>>[[ getFieldAccessString(type, field.getName()) ]]<<, oldChild, (>>[[ getTypeName(componentType) ]]<<) newChild);
                                #//                        else if (oldChild instanceof >>[[ getTypeName(componentType) ]]<<)
                                #//                            $assert (!>>[[ getFieldAccessString(type, field.getName()) ]]<<.contains(oldChild));
                                #//                    #}
                                #//                #}
                                #//            #} else {
                                #//                if (oldChild == null) {
                                #//                    if (>>[[ getFieldAccessString(type, field.getName()) ]]<< == null)
                                #//                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                #//                } else
                                #//                #if (ambiguitySupport) {
                                #//                    if (oldChild instanceof $Parser.$AmbiguityAbsorber) {
                                #//                        if (>>[[ getFieldAccessString(type, field.getName()) ]]<< != null && oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<.$getWrappedAmbiguityAbsorber()))
                                #//                            >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                #//                    } else
                                #//                #}
                                #//                {
                                #//                    if (oldChild.equals(>>[[ getFieldAccessString(type, field.getName()) ]]<<))
                                #//                        >>[[ getFieldAccessString(type, field.getName()) ]]<< = (>>[[ fieldTypeName ]]<<) newChild;
                                #//                }
                                #//            #}
                                #//        #}
                                #//    #}
                                #//}
                                
                            } // class >>[[ identifier ]]<<
                        #}
                    #}
                    
                    
                #}
            } // $Default
            
            #if (ambiguitySupport) {
                #if (targetVersion >= 1.3) {
                    #allocate("createAmbiguousWrapper", topMethodNamespace);
                    private static $Node $createAmbiguousWrapper(final $Parser.$AmbiguityAbsorber ambiguousNode, final java.lang.Class interfaceType) {
                        return ($Node) java.lang.reflect.Proxy.newProxyInstance($Parser.class.getClassLoader(), new java.lang.Class[] { interfaceType }, new java.lang.reflect.InvocationHandler() {
                            public java.lang.Object invoke(java.lang.Object proxy, java.lang.reflect.Method method, java.lang.Object[] args) throws java.lang.Throwable {
                                if (method.getName().equals(>>[[ quoted(getReplacement("$getWrappedAmbiguityAbsorber")) ]]<<) && (args == null || args.length == 0)) {
                                    return ambiguousNode;
                                } else if (method.getName().equals(>>[[ quoted("toString") ]]<<) && (args == null || args.length == 0)) {
                                    return "AmbiguousWrapper wrapping " + ambiguousNode.toString();
                                #/*
                                } else if (method.getName().equals(>>[[ quoted(getReplacement("$getChildNodes")) ]]<<)) {
                                    return ambiguousNode.$getChildNodes();
                                } else if (method.getDeclaringClass().equals(java.lang.Object.class)) {
                                    return method.invoke(ambiguousNode, args);
                                #*/
                                } else {
                                    throw new java.lang.UnsupportedOperationException(interfaceType.getName() + "#" + method.getName() + " by " + ambiguousNode.getClass());
                                }
                            }
                        });
                    }
                #} else {
                    #Iterator it30 = ambiguousWrapperTypes.iterator();
                    #while (it30.hasNext()) {
                        #Type type = (Type) it30.next();
                        #allocate("create" + getSimpleTypeName(type) + "AmbiguousWrapper", topMethodNamespace);
                        private static >>[[ getTypeName(type) ]]<< $create>>[[ getSimpleTypeName(type) ]]<<AmbiguousWrapper(final $Parser.$AmbiguityAbsorber node) {
                            return new >>[[ getTypeName(type) ]]<<() {
                                public $Parser.$AmbiguityAbsorber $getWrappedAmbiguityAbsorber() {
                                    return node;
                                }
                                public java.util.List$<$Parser.$Node> $getChildNodes() {
                                    throw new java.lang.UnsupportedOperationException();
                                }
                                #Iterator it10 = type.getMapNameToField().values().iterator();
                                #while (it10.hasNext()) {
                                    #Field field = (Field) it10.next();
                                    public >>[[ getTypeName(field.getType()) ]]<< >>[[ field.getName() ]]<<() {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                #}
                                public java.lang.String toString() {
                                    return "AmbiguousWrapper wrapping " + node.toString();
                                }
                                #if (type.equals(TokenType.INSTANCE)) {
                                    public int getSymbolID() {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public boolean isWhite() {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public java.lang.String getImage() {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public java.lang.String getOriginalImage() {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public java.lang.String getSourceName() throws java.lang.UnsupportedOperationException {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public int getIndex() throws java.lang.UnsupportedOperationException {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public int getLine() throws java.lang.UnsupportedOperationException {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                    public int getColumn() throws java.lang.UnsupportedOperationException {
                                        throw new java.lang.UnsupportedOperationException();
                                    }
                                #}
                            };
                        }
                    #}
                #}
            #}
            
            #{
                #if (labelWordCount == 1) {
                    #addReplacement("$LabelInteger", "java.lang.Integer");
                    #addReplacement("$LabelInt", "int");
                    #addReplacement("$LABEL_ZERO", "0");
                #} else {
                    #addReplacement("$LabelInteger", "java.math.BigInteger");
                    #addReplacement("$LabelInt", "java.math.BigInteger");
                    #addReplacement("$LABEL_ZERO", "java.math.BigInteger.ZERO");
                #}
                
                #allocate("Link", topTypeNamespace);
                private static class $Link$<T> {
                    #if (targetVersion <= 1.4) {
                        public java.lang.Object value;
                        public $Link$<T> next;
                        public $Link(java.lang.Object value, $Link$<T> next) {
                            this.value = value;
                            this.next = next;
                        }
                        public void addLink(java.lang.Object value) {
                            this.next = new $Link$<T>(value, this.next);
                        }
                    #} else {
                        public T value;
                        public $Link$<T> next;
                        public $Link(T value, $Link$<T> next) {
                            this.value = value;
                            this.next = next;
                        }
                        public void addLink(T value) {
                            this.next = new $Link$<T>(value, this.next);
                        }
                    #}
                    private java.lang.Object[] toArray($Context context) {
                        $Link cursor = this;
                        for (; cursor != null; cursor = cursor.next) {
                            context.tmpList.add(cursor.value);
                        }
                        java.lang.Object[] result = context.tmpList.toArray(new java.lang.Object[context.tmpList.size()]);
                        context.tmpList.clear();
                        return result;
                    }
                    #if (environment.metadebugLight) {
                        public java.lang.String toString() {
                            $StringBuffer result = new $StringBuffer("{ ");
                            $Link$<T> link = this;
                            for (; link != null; link = link.next) {
                                if (link != this)
                                    result.append(", ");
                                result.append(link.value);
                            }
                            result.append(" }");
                            return result.toString();
                        }
                        public static $Link$<java.lang.Object> toLink(java.lang.Object o) {
                            if (o instanceof $ObjectLink)
                                return ($ObjectLink) o;
                            return new $ObjectLink(o, null);
                        }
                    #}
                }
                #if (targetVersion >= 1.5) {
                    #allocate("ObjectLink", topTypeNamespace);
                    private static class $ObjectLink extends $Link$<java.lang.Object> {
                        public $ObjectLink(java.lang.Object value, $Link$<java.lang.Object> next) {
                            super(value, next);
                        }
                    }
                #} else {
                    #addReplacement("$ObjectLink", getReplacement("$Link"));
                #}
                #if (targetVersion >= 1.5) {
                    #allocate("ObjectList", topTypeNamespace);
                    private static class $ObjectList extends java.util.LinkedList$<java.lang.Object> {
                        private static final long serialVersionUID = 0;
                    }
                #} else {
                    #addReplacement("$ObjectList", "java.util.LinkedList");
                #}
                #if (targetVersion >= 1.5) {
                    #allocate("ObjectSet", topTypeNamespace);
                    private static class $ObjectSet extends $LinkedHashSet$<java.lang.Object> {
                        private static final long serialVersionUID = 0;
                    }
                #} else {
                    #addReplacement("$ObjectSet", getReplacement("$LinkedHashSet"));
                #}
                #allocate("StateVertex", topTypeNamespace);
                private static final class $StateVertex extends $Link$<$SymbolVertex> {
                    public int stateNo;
                    public $Link$<$Token> whiteTokenHolder;
                    #if (environment.metadebugLight) {
                        public int generation;
                    #}
                    public $StateVertex(int stateNo, int generation) {
                        super(null, null);
                        this.stateNo = stateNo;
                        $assert generation == 0;
                        #if (environment.metadebugLight) {
                            this.generation = generation;
                        #}
                    }
                    public $StateVertex($SymbolVertex successor, int stateNo, int generation) {
                        super(successor, null);
                        this.stateNo = stateNo;
                        #if (environment.metadebugLight) {
                            this.generation = generation;
                        #}
                    }
                    public void addSuccessor($SymbolVertex successor) {
                        super.addLink(successor);
                    }
                    public $Link$<$SymbolVertex> getSuccessors() {
                        $assert value != null;
                        return this;
                    }
                    #if (environment.metadebugLight) {
                        public java.lang.String toString() {
                            return "(" + $lrStateIDToName[stateNo] + "@" + generation + ")";
                        }
                    #}
                }
                #allocate("SymbolVertex", topTypeNamespace);
                private static final class $SymbolVertex extends $Link$<$StateVertex> {
                    #if (environment.metadebugLight) {
                        public static int numberOfObjects;
                        public int idNumber = numberOfObjects++;
                    #}
                    public java.lang.Object symbol;     // $ConcreteSyntaxTree, $Node, or $ObjectLink of $ConcreteSyntaxTree for absorbers.
                    public $SymbolVertex($StateVertex successor, java.lang.Object symbol) {
                        super(successor, null);
                        this.symbol = symbol;
                    }
                    public void addSuccessor($StateVertex successor) {
                        super.addLink(successor);
                    }
                    public $Link$<$StateVertex> getSuccessors() {
                        $assert value != null;
                        return this;
                    }
                    #if (environment.metadebugLight) {
                        public java.lang.String toString() {
                            $StringBuffer result = new $StringBuffer("[#");
                            result.append(idNumber);
                            result.append(" ");
                            $Link$<java.lang.Object> link = $ObjectLink.toLink(symbol);
                            for (; link != null; link = link.next) {
                                if (link.value instanceof $ConcreteSyntaxTree) {
                                    $ConcreteSyntaxTree cst = ($ConcreteSyntaxTree) link.value;
                                    result.append(cst);
                                } else {
                                    $Node node = ($Node) link.value;
                                    if (node instanceof Token) {
                                        result.append(node);
                                    } else {
                                        java.lang.String name = node.getClass().getName();
                                        result.append(name.substring(name.lastIndexOf('$') + 1));
                                    }
                                }
                                if (link.next != null)
                                    result.append(", ");
                            }
                            result.append("]");
                            return result.toString();
                        }
                    #}
                }
                #allocate("Chore", topTypeNamespace);
                private static final class $Chore {
                    $StateVertex state;
                    $SymbolVertex successor;
                    int reductionID;
                    public $Chore($StateVertex state, int reductionID) {
                        this(state, null, reductionID);
                    }
                    public $Chore($StateVertex state, $SymbolVertex successor, int reductionID) {
                        this.state = state;
                        this.successor = successor;
                        this.reductionID = reductionID;
                        
                        int[] reduction = $Parser.$reductionTable[reductionID];
                        $assert (successor == null) == (reduction.length == 1);
                        // successor == null  <==>  the reduction is an empty reduction.
                    }
                    #if (environment.metadebugLight) {
                        public java.lang.String toString() {
                            int[] reduction = $Parser.$reductionTable[reductionID];
                            return $symbolIDToName[reduction[0]] + "<--" + ((reduction.length - 1) / >>[[ labelWordCount ]]<<) + "--" + successor + "<-" + state;
                        }
                    #}
                }
                #allocate("ConcreteSyntaxTree", topTypeNamespace);
                private static final class $ConcreteSyntaxTree {
                    #if (environment.metadebugLight) {
                        private $Parser parser;
                    #}
                    public int reductionID;
                    public java.lang.Object[] children;
                                // odd: same as $SymbolVertex#symbol. even: $Link of white tokens (maybe null).
                                // or, null if this has an empty reduction.
                    public $Node node;
                    public $ConcreteSyntaxTree($Context context, int reductionID, java.lang.Object[] children) throws $ParseException {
                        #if (environment.metadebugLight) {
                            this.parser = context.parser;
                        #}
                        this.reductionID = reductionID;
                        this.children = children;
                    }
                    public $Node createNode($Context context) throws $ParseException {
                        if (node == null) {
                            int[] reduction = context.parser.$reductionTable[reductionID];
                            if (reduction[0] < >>[[ tables.superiorIndexOfTypeNonterminals() ]]<<) {
                                node = context.parser.$createNode(context, this);
                                #if (environment.metadebug) {
                                    //java.lang.System.err.println("createNode: " + $symbolIDToName[reduction[0]] + " <---- " + java.util.Arrays.asList(children));
                                #}
                            }
                        }
                        return node;
                    }
                    #if (environment.metadebugLight) {
                        public java.lang.String toString() {
                            int[] reduction = parser.$reductionTable[reductionID];
                            //return "CST(" + $symbolIDToName[reduction[0]] + ")";
                            if (children == null) {
                                return "CST(" + $symbolIDToName[reduction[0]] + ")" + null;
                            } else {
                                return "CST(" + $symbolIDToName[reduction[0]] + ")" + java.util.Arrays.asList(children);
                            }
                        }
                    #}
                }
                #allocate("AbsorbKey2", topTypeNamespace);
                private static final class $AbsorbKey2 {
                    public java.util.Set$<java.lang.Object> states;
                    public int symbolID;
                    public $AbsorbKey2(java.util.Set$<java.lang.Object> states, int symbolID) {
                        this.states = states;
                        this.symbolID = symbolID;
                    }
                    public int hashCode() {
                        return states.hashCode() + symbolID;
                    }
                    public final boolean equals(java.lang.Object rhs) {
                        if (rhs instanceof $AbsorbKey2)
                            return equals(($AbsorbKey2) rhs);
                        return false;
                    }
                    public final boolean equals($AbsorbKey2 rhs) {
                        return states.equals(rhs.states) && symbolID == rhs.symbolID;
                    }
                }
                
                #allocate("binarySearch", topMethodNamespace);
                private static int $binarySearchLRTable(int[] table, int key) {
                    // The 'table' is list of key and value, and additional Integer.MAX_VALUE.  Even indexes are keys and odd indexes are values.
                    // The result is the minimum index of the key that is grater or equals to 'key'.
                    int base = 0;
                    int width = (table.length + 1) >> 1;
                    while (width > 0) {
                        // The result should be in base..(base+width*2)
                        int halfWidth = width >> 1;
                        int index = base + (width & ~1);    // base + halfWidth * 2
                        if (table[index] < key) {
                            // The result should be in index+2..(base+width*2)
                            base = index + 2;
                            width = width - halfWidth - 1;
                        } else {
                            // The result should be in base..index
                            width = halfWidth;
                        }
                    }
                    return base;
                }
                /*#allocate("getDirective", topMethodNamespace);
                private int $getDirective(int[] symbolIDToDirective, int symbolID) {
                    #assert !multipleShiftEntry;
                    int index = $binarySearch(symbolIDToDirective, symbolID);
                    if (symbolIDToDirective[index] == symbolID) {
                        return symbolIDToDirective[index + 1];
                    } else {
                        return -1;
                    }
                }*/
                
                #allocate("Context", topTypeNamespace);
                private static final class $Context {
                    public $Context($Parser parser) {
                        this.parser = parser;
                    }
                    public final $Parser parser;
                    public java.util.Map$<java.lang.Integer, $StateVertex> states = new $LinkedHashMap$<java.lang.Integer, $StateVertex>();
                    public final java.util.Set$<$StateVertex> unbrokenStates = new $LinkedHashSet$<$StateVertex>();
                    public final java.util.List$<$Chore> chores = new java.util.ArrayList$<$Chore>();
                    public final java.util.Map$<$AbsorbKey2, $SymbolVertex> absorbMap2 = new $LinkedHashMap$<$AbsorbKey2, $SymbolVertex>();
                    public int generation;
                    public $Link$<$Token> whiteTokenHolder;
                    public int nextTokenID;
                    public $Token nextToken;
                    public final $NodeInitializationParameters parameters = new $NodeInitializationParameters();
                    public final java.util.List$<java.lang.Object> tmpList = new java.util.ArrayList$<java.lang.Object>();
                    public final java.util.Map$<java.lang.Object, java.lang.Object> tmpMap = new $LinkedHashMap$<java.lang.Object, java.lang.Object>();
                }
                #allocate("parse", topMethodNamespace);
                private $Node $parse($LexicalAnalyzer analyzer, int initialStateNo) throws $ParseException {
                    final int labelWordCount = >>[[ labelWordCount ]]<<;
                    final int[][] lrTable = $Parser.$lrTable;
                    final int[][] reductionTable = $Parser.$reductionTable;
                    
                    // build subtrees
                    $Context context = new $Context(this);
                    $StateVertex firstStateVertex = new $StateVertex(initialStateNo, 0);
                    {
                        context.whiteTokenHolder = new $Link$<$Token>(null, null);
                        firstStateVertex.whiteTokenHolder = context.whiteTokenHolder;
                        context.states.put(new java.lang.Integer(firstStateVertex.stateNo), firstStateVertex);
                        
                        java.util.Map$<java.lang.Integer, $StateVertex> nextStates = new $LinkedHashMap$<java.lang.Integer, $StateVertex>();
                        
                    tokenLoop:
                        for (context.generation = 0;; context.generation++) {
                            $assert context.unbrokenStates.isEmpty();
                            context.unbrokenStates.addAll(context.states.values());
                            {
                                $assert context.whiteTokenHolder.next == null;
                                
                                #if (true) {
                                    $Link$<$Token> whiteTokensAppendPosition = context.whiteTokenHolder;
                                    for (;;) {
                                        context.nextToken = analyzer.next();
                                        context.nextTokenID = context.nextToken.getSymbolID();
                                        if (context.nextToken.isWhite()) {
                                            whiteTokensAppendPosition.next = new $Link$<$Token>(context.nextToken, null);
                                            whiteTokensAppendPosition = whiteTokensAppendPosition.next;
                                        } else {
                                            break;
                                        }
                                    }
                                #} else {
                                    // 順序を逆にする場合
                                    for (;;) {
                                        context.nextToken = analyzer.next();
                                        context.nextTokenID = context.nextToken.getSymbolID();
                                        if (context.nextToken.isWhite()) {
                                            context.whiteTokenHolder.addLink(context.nextToken);
                                        } else {
                                            break;
                                        }
                                    }
                                #}
                            }
                            
                            #if (environment.metadebug) {
                                java.lang.System.err.println("** generation = " + context.generation);
                                java.lang.System.err.println("nextTokenID = " + $symbolIDToName[context.nextTokenID] + ", nextToken = " + context.nextToken);
                                java.lang.System.err.println("whiteTokens = " + context.whiteTokenHolder.next);
                                java.lang.System.err.println("states = " + context.states.values());
                            #}
                            
                        reduceLoop:
                            while (!context.unbrokenStates.isEmpty() || !context.chores.isEmpty()) {
                                if (!context.unbrokenStates.isEmpty()) {
                                    java.util.Iterator$<$StateVertex> it = context.unbrokenStates.iterator();
                                    $StateVertex stateToBreak = ($StateVertex) it.next();
                                    it.remove();
                                    it = null;
                                    $breakState(context, null, stateToBreak);
                                } else {
                                    java.util.Iterator$<$Chore> it = context.chores.iterator();
                                    $Chore chore = ($Chore) it.next();
                                    it.remove();
                                    it = null;
                                    $chopChore(context, chore);
                                }
                            }
                            context.absorbMap2.clear();
                            
                            $Link$<$Token> oldWhiteTokenHolder = context.whiteTokenHolder;
                            context.whiteTokenHolder = new $Link$<$Token>(null, null);
                            java.util.Iterator$<$StateVertex> it25 = context.states.values().iterator();
                        shiftLoop:
                            while (it25.hasNext()) {
                                $StateVertex currentState = ($StateVertex) it25.next();
                                
                                final int[] tokenIDToNextState = lrTable[currentState.stateNo * 3 + 0];
                                int index = $binarySearchLRTable(tokenIDToNextState, context.nextTokenID);
                                if (tokenIDToNextState[index] == context.nextTokenID) {
                                    if (context.generation != 0) {
                                        $Link$<$SymbolVertex> it = currentState.getSuccessors();
                                        do {
                                            $SymbolVertex prevSymbolVertex = ($SymbolVertex) it.value;
                                            if (!(prevSymbolVertex.symbol instanceof $Node)) {
                                                if (prevSymbolVertex.symbol instanceof $ConcreteSyntaxTree) {
                                                    $ConcreteSyntaxTree tree = ($ConcreteSyntaxTree) prevSymbolVertex.symbol;
                                                    $Node node = tree.createNode(context);
                                                    if (node != null)
                                                        prevSymbolVertex.symbol = node;
                                                } else {
                                                    // TODO: 実装
                                                    //$Node node = $createNode(context, ($ObjectLint) prevSymbolVertex.symbol);
                                                    //if (node != null)
                                                    //    prevSymbolVertex.symbol = node;
                                                }
                                            }
                                            it = it.next;
                                        } while (it != null);
                                    }
                                    
                                    for (; tokenIDToNextState[index] == context.nextTokenID; index += 2) {
                                        int nextStateNo = tokenIDToNextState[index + 1];
                                        java.lang.Integer nextStateNoInteger = new java.lang.Integer(nextStateNo);
                                        $StateVertex nextState = ($StateVertex) nextStates.get(nextStateNoInteger);
                                        if (nextState == null) {
                                            $SymbolVertex symbolVertex = new $SymbolVertex(currentState, context.nextToken);
                                            nextState = new $StateVertex(symbolVertex, nextStateNo, context.generation + 1);
                                            nextState.whiteTokenHolder = context.whiteTokenHolder;
                                            nextStates.put(nextStateNoInteger, nextState);
                                            #if (environment.metadebug) {
                                                java.lang.System.err.println("shift: " + currentState + " <- " + symbolVertex + "(new) <- " + nextState + "(new)");
                                            #}
                                        } else {
                                            $assert nextState.getSuccessors().next == null; // i.e. size == 1
                                            $SymbolVertex symbolVertex = ($SymbolVertex) nextState.getSuccessors().value;
                                            symbolVertex.addSuccessor(currentState);
                                            #if (environment.metadebug) {
                                                java.lang.System.err.println("shift: " + currentState + " <- " + symbolVertex + "(already exists) <- " + nextState + "(already exists)");
                                            #}
                                        }
                                        #assert !multipleShiftEntry;
                                        break;
                                    }
                                }
                            }
                            
                            if (nextStates.isEmpty()) {
                                context.whiteTokenHolder = oldWhiteTokenHolder;
                                #if (environment.metadebug) {
                                    java.lang.System.err.println("error: nextStates.isEmpty()");
                                #}
                                $error(context, analyzer);
                                continue tokenLoop;
                            }
                            
                            if (context.nextTokenID == EOF_TOKEN) {
                                context.states = nextStates;
                                break tokenLoop;
                            }
                            
                            context.states.clear();
                            java.util.Map$<java.lang.Integer, $StateVertex> oldStates = context.states;
                            java.util.Map$<java.lang.Integer, $StateVertex> oldNextStates = nextStates;
                            context.states = oldNextStates;
                            nextStates = oldStates;
                            //java.lang.System.gc();
                        }
                    }
                    
                    $assert context.states.size() == 1;
                    $StateVertex sinkVertex = ($StateVertex) context.states.values().iterator().next();
                    
                    $assert sinkVertex.getSuccessors().next == null;
                    $SymbolVertex eofVertex = ($SymbolVertex) sinkVertex.getSuccessors().value;
                    
                    $assert eofVertex.getSuccessors().next == null;
                    $StateVertex lastStateVertex = ($StateVertex) eofVertex.getSuccessors().value;
                    
                    if (lastStateVertex.getSuccessors().next != null) {
                        throw new $ParseException.AmbiguousGrammarError(>>[[ quoted(environment.processingFile.getName()) ]]<< + " is ambiguous.  Use --debug option to build a verbose parser.");
                    }
                    $SymbolVertex rootSymbolVertex = ($SymbolVertex) lastStateVertex.getSuccessors().value;
                    
                    $assert rootSymbolVertex.getSuccessors().next == null;
                    $assert firstStateVertex == ($StateVertex) rootSymbolVertex.getSuccessors().value;
                    
                    $ObjectLink rootLink = ($ObjectLink) rootSymbolVertex.symbol;
                    $ConcreteSyntaxTree rootCST = ($ConcreteSyntaxTree) rootLink.value;
                    
                    return $createNode(context, firstStateVertex.whiteTokenHolder.next, rootCST, lastStateVertex.whiteTokenHolder.next);
                }
                #allocate("REDUCE_ALL", topFieldNamespace);
                private final static int $ERROR_REPORT_MODE = -1;
                #allocate("error", topMethodNamespace);
                private void $error($Context context, $LexicalAnalyzer analyzer) throws $ParseException {
                    final int[][] lrTable = $Parser.$lrTable;
                    
                    $assert context.unbrokenStates.isEmpty();
                    context.unbrokenStates.addAll(context.states.values());
                    
                    $Token errorNextToken = context.nextToken;
                    context.nextTokenID = $ERROR_REPORT_MODE;
                    context.nextToken = null;
                    
                tokenLoop:
                    while (!context.unbrokenStates.isEmpty() || !context.chores.isEmpty()) {
                        if (!context.unbrokenStates.isEmpty()) {
                            java.util.Iterator$<$StateVertex> it = context.unbrokenStates.iterator();
                            $StateVertex stateToBreak = ($StateVertex) it.next();
                            it.remove();
                            it = null;
                            $breakState(context, null, stateToBreak);
                        } else {
                            java.util.Iterator$<$Chore> it = context.chores.iterator();
                            $Chore chore = ($Chore) it.next();
                            it.remove();
                            it = null;
                            $chopChore(context, chore);
                        }
                    }
                    context.absorbMap2.clear();
                    
                    final java.util.Set$<java.lang.Integer> nextTokens = new $LinkedHashSet$<java.lang.Integer>();
                    
                    java.util.Iterator$<$StateVertex> it25 = context.states.values().iterator();
                shiftLoop:
                    while (it25.hasNext()) {
                        $StateVertex currentState = ($StateVertex) it25.next();
                        
                        final int[] tokenIDToNextState = lrTable[currentState.stateNo * 3 + 0];
                        int index = 0;
                        for (; tokenIDToNextState[index] != $TABLE_SENTINEL; index += 2) {
                            nextTokens.add(new java.lang.Integer(tokenIDToNextState[index]));
                        }
                    }
                    
                    int[] selectableTokens = new int[nextTokens.size()];
                    int index = 0;
                    java.util.Iterator it11 = nextTokens.iterator();
                    while (it11.hasNext()) {
                        java.lang.Integer i = (java.lang.Integer) it11.next();
                        selectableTokens[index++] = i.intValue();
                    }
                    $error(analyzer, errorNextToken, selectableTokens);
                }
                private static void $error($LexicalAnalyzer analyzer, $Token errorNextToken, int[] selectableNextTokens) throws $ParseException {
                    java.io.StringWriter buffer = new java.io.StringWriter();
                    java.io.PrintWriter writer = new java.io.PrintWriter(buffer);
                    
                    if (errorNextToken.getSymbolID() != EOF_TOKEN) {
                        writer.println("The token " + errorNextToken + " should be one of the following.");
                    } else {
                        writer.println("The EOF should be one of the following.");
                    }
                    
                    final java.lang.String[] tokenStrings = {
                        #{
                            #Symbol[] symbols = tables.symbols();
                            #for (int i = 0; i < tables.superiorIndexOfTerminals(); i++) {
                                #Symbol symbol = symbols[i];
                                >>[[ quoted(symbol.toString()) ]]<<,
                            #}
                        #}
                    };
                    
                    for (int i = 0; i < selectableNextTokens.length; i++) {
                        int selectableNextToken = selectableNextTokens[i];
                        writer.print("        ");
                        writer.println(tokenStrings[selectableNextToken]);
                    }
                    
                    writer.close();
                    if (writer.checkError())
                        throw new java.lang.RuntimeException();
                    throw new $ParseException(buffer.toString(), errorNextToken);
                }
                
                #allocate("breakState", topMethodNamespace);
                private void $breakState($Context context, $SymbolVertex successorRestriction, $StateVertex state) {
                    final int[][] lrTable = $Parser.$lrTable;
                    
                    #if (environment.metadebug) {
                        java.lang.System.err.println("breakState: state = " + state + ", successorRestriction = " + successorRestriction);
                    #}
                    
                    int[] tokenIDToReductionID = lrTable[state.stateNo * 3 + 2];
                    if (context.nextTokenID != $ERROR_REPORT_MODE) {
                        int index = $binarySearchLRTable(tokenIDToReductionID, context.nextTokenID);
                        for (; tokenIDToReductionID[index] == context.nextTokenID; index += 2) {
                            final int reductionID = tokenIDToReductionID[index + 1];
                            $breakState(context, successorRestriction, state, reductionID);
                        }
                    } else {
                        java.util.Set$<java.lang.Integer> reductionIDs = new java.util.TreeSet$<java.lang.Integer>();
                        int index = 0;
                        for (; tokenIDToReductionID[index] != $TABLE_SENTINEL; index += 2) {
                            final int reductionID = tokenIDToReductionID[index + 1];
                            reductionIDs.add(new java.lang.Integer(reductionID));
                        }
                        java.util.Iterator it31 = reductionIDs.iterator();
                        while (it31.hasNext()) {
                            java.lang.Integer reductionID = (java.lang.Integer) it31.next();
                            $breakState(context, successorRestriction, state, reductionID.intValue());
                        }
                    }
                    #if (environment.metadebug) {
                        java.lang.System.err.println("breakState: end");
                    #}
                }
                private void $breakState($Context context, $SymbolVertex successorRestriction, $StateVertex state, int reductionID) {
                    final int[][] reductionTable = $Parser.$reductionTable;
                    final int[] reduction = reductionTable[reductionID];
                    
                    #if (environment.metadebug) {
                        java.lang.System.err.println("breakState: reduction = " + $symbolIDToName[reduction[0]] + ", length = " + ((reduction.length - 1) / >>[[ labelWordCount ]]<<));
                    #}
                    if (reduction.length == 1) {
                        if (successorRestriction == null) {
                            $Chore chore = new $Chore(state, reductionID);
                            context.chores.add(chore);
                            #if (environment.metadebug) {
                                java.lang.System.err.println("breakState: add chore " + chore);
                            #}
                        }
                    } else {
                        if (successorRestriction == null) {
                            $Link$<$SymbolVertex> it1 = state.getSuccessors();
                            do {
                                $SymbolVertex successor = ($SymbolVertex) it1.value;
                                $Chore chore = new $Chore(state, successor, reductionID);
                                context.chores.add(chore);
                                #if (environment.metadebug) {
                                    java.lang.System.err.println("breakState: add chore " + chore);
                                #}
                            } while ((it1 = it1.next) != null);
                        } else {
                            $SymbolVertex successor = successorRestriction;
                            $Chore chore = new $Chore(state, successor, reductionID);
                            context.chores.add(chore);
                            #if (environment.metadebug) {
                                java.lang.System.err.println("breakState: add chore " + chore);
                            #}
                        }
                    }
                }
                #allocate("chopChore", topMethodNamespace);
                private void $chopChore($Context context, $Chore chore) throws $ParseException {
                    final int labelWordCount = >>[[ labelWordCount ]]<<;
                    $assert chore.state.whiteTokenHolder == context.whiteTokenHolder;
                    
                    #if (environment.metadebug) {
                        java.lang.System.err.println("chopChore: chore = " + chore);
                    #}
                    
                    int reductionID = chore.reductionID;
                    if (chore.successor == null) {
                        $StateVertex turningPoint = chore.state;
                        $addLink(context, turningPoint, reductionID, null);
                    } else {
                        $StateVertex state = chore.state;
                        int[] reduction = $Parser.$reductionTable[reductionID];
                        int reductionSize = (reduction.length - 1) / labelWordCount;
                        $assert reductionSize > 0;
                        $backChoreOdd(context, reductionID, reductionSize * 2 - 1, chore.successor, new $Link$<java.lang.Object>(state.whiteTokenHolder, null), false);
                    }
                    
                    #if (environment.metadebug) {
                        java.lang.System.err.println("chopChore: end");
                    #}
                }
                #allocate("backChoreOdd", topMethodNamespace);
                private void $backChoreOdd($Context context, int reductionID, int countOfSuccessorsToReduce, $SymbolVertex symbolVertex, $Link$<java.lang.Object> children, boolean nodeCreatable) throws $ParseException {
                    $assert countOfSuccessorsToReduce % 2 == 1;
                    #if (environment.metadebug) {
                        {
                            java.lang.System.err.print("backChore: ");
                            int[] reduction = $Parser.$reductionTable[reductionID];
                            int indent = ((reduction.length - 1) / >>[[ labelWordCount ]]<<) * 2 - countOfSuccessorsToReduce;
                            for (int i = 0; i < indent; i++)
                                java.lang.System.err.print("  ");
                            java.lang.System.err.println(symbolVertex);
                        }
                    #}
                    
                    java.lang.Object child = symbolVertex.symbol;
                    if (child instanceof $ConcreteSyntaxTree && (context.nextTokenID != $ERROR_REPORT_MODE)) {
                        $ConcreteSyntaxTree cst = ($ConcreteSyntaxTree) child;
                        if (false && cst.reductionID < >>[[ tables.superiorIndexOfSkippableInCSTReduction() ]]<<) {
                            // TODO: skip の判定が間違っているので直す。
                            child = cst.children[1];
                        } else if (false && nodeCreatable) {    // TODO: nodeCreatable をなくす。
                            $Node node = cst.createNode(context);
                            if (node != null)
                                child = node;
                        }
                    }
                    children = new $Link$<java.lang.Object>(child, children);
                    
                    if (countOfSuccessorsToReduce > 1) {
                        countOfSuccessorsToReduce--;
                        $Link$<$StateVertex> it2 = symbolVertex.getSuccessors();
                        do {
                            $StateVertex state = ($StateVertex) it2.value;
                            $backChoreEven(context, reductionID, countOfSuccessorsToReduce, state, children);
                        } while ((it2 = it2.next) != null);
                    } else {
                        $StateVertex state = ($StateVertex) symbolVertex.getSuccessors().value;
                        children = new $Link$<java.lang.Object>(state.whiteTokenHolder.next, children);
                        {
                            $Link$<$StateVertex> it2 = symbolVertex.getSuccessors();
                            do {
                                $StateVertex s = ($StateVertex) it2.value;
                                $assert s.whiteTokenHolder.next == state.whiteTokenHolder.next;
                            } while ((it2 = it2.next) != null);
                        }
                        $addLinks(context, reductionID, symbolVertex, children);
                    }
                }
                #allocate("backChoreEven", topMethodNamespace);
                private void $backChoreEven($Context context, int reductionID, int countOfSuccessorsToReduce, $StateVertex state, $Link$<java.lang.Object> children) throws $ParseException {
                    $assert countOfSuccessorsToReduce % 2 == 0;
                    #if (environment.metadebug) {
                        {
                            java.lang.System.err.print("backChore: ");
                            int[] reduction = $Parser.$reductionTable[reductionID];
                            int indent = ((reduction.length - 1) / >>[[ labelWordCount ]]<<) * 2 - countOfSuccessorsToReduce;
                            for (int i = 0; i < indent; i++)
                                java.lang.System.err.print("  ");
                            java.lang.System.err.println(state);
                        }
                    #}
                    children = new $Link$<java.lang.Object>(state.whiteTokenHolder.next, children);
                    $assert countOfSuccessorsToReduce > 0;
                    countOfSuccessorsToReduce--;
                    $Link$<$SymbolVertex> it1 = state.getSuccessors();
                    do {
                        $SymbolVertex symbolVertex = ($SymbolVertex) it1.value;
                        $backChoreOdd(context, reductionID, countOfSuccessorsToReduce, symbolVertex, children, state.whiteTokenHolder != context.whiteTokenHolder);
                    } while ((it1 = it1.next) != null);
                }
                #allocate("addLinks", topMethodNamespace);
                private void $addLinks($Context context, int reductionID, $SymbolVertex symbolVertex, $Link$<java.lang.Object> children) throws $ParseException {
                    final int[][] lrTable = $Parser.$lrTable;
                    final int[] reduction = $Parser.$reductionTable[reductionID];
                    final int nextSymbolID = reduction[0];
                    #//final boolean absorberSymbol = (nextSymbolID >= >>[[ tables.firstIndexOfAbsorberNonterminals() ]]<<);
                    final boolean absorberSymbol = (nextSymbolID < >>[[ tables.superiorIndexOfAliasNonterminals() ]]<<);
                    $assert children != null;
                    $ConcreteSyntaxTree tree = new $ConcreteSyntaxTree(context, reductionID, children == null ? null : children.toArray(context));
                    
                    final boolean aho = >>[[ environment.aho ]]<<;
                    if (!aho) {
                        if (!absorberSymbol) {
                            $Link$<$StateVertex> it2 = symbolVertex.getSuccessors();
                            do {
                                $StateVertex turningPoint = ($StateVertex) it2.value;
                                final int[] symbolIDToNextState = lrTable[turningPoint.stateNo * 3 + 1];
                                int index = $binarySearchLRTable(symbolIDToNextState, nextSymbolID);
                                #assert !multipleShiftEntry;
                                if (symbolIDToNextState[index] == nextSymbolID) {
                                    int nextStateNo = symbolIDToNextState[index + 1];
                                    $addLink(context, java.util.Collections.$<java.lang.Object>singleton(turningPoint), reductionID, tree, new java.lang.Integer(nextStateNo));
                                }
                            } while ((it2 = it2.next) != null);
                            return;
                        }
                    }
                    
                    $Link$<$StateVertex> it2 = symbolVertex.getSuccessors();
                    do {
                        $StateVertex turningPoint = ($StateVertex) it2.value;
                        final int[] symbolIDToNextState = lrTable[turningPoint.stateNo * 3 + 1];
                        int index = $binarySearchLRTable(symbolIDToNextState, nextSymbolID);
                        #assert !multipleShiftEntry;
                        if (symbolIDToNextState[index] == nextSymbolID) {
                            int nextStateNo = symbolIDToNextState[index + 1];
                            java.lang.Integer key = new java.lang.Integer(nextStateNo);
                            $ObjectSet set = ($ObjectSet) context.tmpMap.get(key);
                            if (set == null) {
                                set = new $ObjectSet();
                                context.tmpMap.put(key, set);
                            }
                            set.add(turningPoint);
                        }
                    } while ((it2 = it2.next) != null);
                    
                    java.util.Iterator it26 = context.tmpMap.entrySet().iterator();
                    while (it26.hasNext()) {
                        java.util.Map.Entry entry = (java.util.Map.Entry) it26.next();
                        java.lang.Integer nextStateNo = (java.lang.Integer) entry.getKey();
                        $ObjectSet turningPoints = ($ObjectSet) entry.getValue();
                        
                        $addLink(context, turningPoints, reductionID, tree, nextStateNo);
                    }
                    context.tmpMap.clear();
                }
                #allocate("addLink", topMethodNamespace);
                private void $addLink($Context context, $StateVertex turningPoint, int reductionID, $Link$<java.lang.Object> children) throws $ParseException {
                    final int[][] lrTable = $Parser.$lrTable;
                    final int[] reduction = $Parser.$reductionTable[reductionID];
                    final int nextSymbolID = reduction[0];
                    $assert children == null;
                    $ConcreteSyntaxTree tree = new $ConcreteSyntaxTree(context, reductionID, children == null ? null : children.toArray(context));
                    
                    final int[] symbolIDToNextState = lrTable[turningPoint.stateNo * 3 + 1];
                    int index = $binarySearchLRTable(symbolIDToNextState, nextSymbolID);
                    #assert !multipleShiftEntry;
                    if (symbolIDToNextState[index] == nextSymbolID) {
                        int nextStateNo = symbolIDToNextState[index + 1];
                        $addLink(context, java.util.Collections.$<java.lang.Object>singleton(turningPoint), reductionID, tree, new java.lang.Integer(nextStateNo));
                    }
                }
                private void $addLink($Context context, java.util.Set$<java.lang.Object> turningPoints, int reductionID, $ConcreteSyntaxTree tree, java.lang.Integer nextStateNoInteger) {
                    final int[] reduction = $Parser.$reductionTable[reductionID];
                    final int nextSymbolID = reduction[0];
                    #//final boolean absorberSymbol = (nextSymbolID >= >>[[ tables.firstIndexOfAbsorberNonterminals() ]]<<);
                    final boolean absorberSymbol = (nextSymbolID < >>[[ tables.superiorIndexOfAliasNonterminals() ]]<<);
                    
                    #if (environment.metadebug) {
                        java.lang.System.err.println("addLink: turningPoints = " + turningPoints + ", nextSymbolID = " + $symbolIDToName[nextSymbolID]);
                    #}
                    
                    $StateVertex stateVertex = ($StateVertex) context.states.get(nextStateNoInteger);
                    $SymbolVertex symbolVertex = null;  // if null, create a new sumbol vertex.
                    $AbsorbKey2 key = null;             // if not null, maintain the dictionary.
                    #if (environment.debug) {
                        #//final boolean debugCheck = (>>[[ tables.firstIndexOfUnambiguityAssertedNonterminals() ]]<< <= nextSymbolID) && (nextSymbolID < >>[[ tables.superiorIndexOfUnambiguityAssertedNonterminals() ]]<<);
                        final boolean debugCheck = false;
                    #} else {
                        final boolean debugCheck = false;
                    #}
                    final boolean aho = >>[[ environment.aho ]]<<;
                    if (aho) {
                        key = new $AbsorbKey2(turningPoints, nextSymbolID);
                        if (stateVertex == null) {
                            $assert ($SymbolVertex) context.absorbMap2.get(key) == null;
                        } else {
                            symbolVertex = ($SymbolVertex) context.absorbMap2.get(key);
                        }
                        if (debugCheck || absorberSymbol) {
                            if (symbolVertex != null && !absorberSymbol) {
                                // TODO: report
                                throw new $ParseException.AmbiguousGrammarError(>>[[ quoted(environment.processingFile.getName()) ]]<< + " is ambiguous.  Use --debug option to build a verbose parser.");
                                //key = null;
                                //symbolVertex = null;
                            }
                        }
                        if (debugCheck || absorberSymbol) {
                        } else {
                            symbolVertex = null;
                        }
                    } else {
                        if (debugCheck || absorberSymbol || context.nextTokenID == $ERROR_REPORT_MODE) {
                            key = new $AbsorbKey2(turningPoints, nextSymbolID);
                            if (stateVertex == null) {
                                $assert ($SymbolVertex) context.absorbMap2.get(key) == null;
                            } else {
                                symbolVertex = ($SymbolVertex) context.absorbMap2.get(key);
                            }
                            #if (!ambiguitySupport) {
                                if (symbolVertex != null && context.nextTokenID != $ERROR_REPORT_MODE) {
                                    // TODO: report
                                    
                                    #if (environment.debug) {
                                        java.lang.System.err.println("-------- Ambiguity detected, case 1");
                                        $debugPrint(context, "  ", symbolVertex);
                                        java.lang.System.err.println("-------- Ambiguity detected, case 2");
                                        $debugPrint(context, "  ", tree);
                                        java.lang.System.err.println("--------");
                                        throw new $ParseException.AmbiguousGrammarError(>>[[ quoted(environment.processingFile.getName()) ]]<< + " is ambiguous.");
                                    #} else {
                                        throw new $ParseException.AmbiguousGrammarError(>>[[ quoted(environment.processingFile.getName()) ]]<< + " is ambiguous.  Use --debug option to build a verbose parser.");
                                    #}
                                    
                                    //key = null;
                                    //symbolVertex = null;
                                }
                            #}
                        }
                    }
                    if (symbolVertex != null) {
                        #if (environment.metadebug) {
                            java.lang.System.err.println("addLink: add " + tree + " to existing vertex " + symbolVertex);
                        #}
                        if (context.nextTokenID != $ERROR_REPORT_MODE) {
                            (($ObjectLink) symbolVertex.symbol).addLink(tree);
                        }
                        #if (environment.metadebug) {
                            java.lang.System.err.println("addLink: result=" + symbolVertex);
                        #}
                    } else {
                        java.util.Iterator it = turningPoints.iterator();
                        symbolVertex = new $SymbolVertex(($StateVertex) it.next(), new $ObjectLink(tree, null));
                        // 注意：上の$SymbolVertex で、第二引数を tree にして、あとで addLink が必要になってから $ObjectLink を作ってはいけない。
                        // addLink されるより前に flatNode の対象になる可能性がある。
                        // TODO: シフト時に flatNode する仕様にしたので、そうではなくなった。
                        while (it.hasNext()) {
                            $StateVertex turningPoint = ($StateVertex) it.next();
                            symbolVertex.addSuccessor(turningPoint);
                        }
                        if (key != null)
                            context.absorbMap2.put(key, symbolVertex);
                        if (stateVertex == null) {
                            stateVertex = new $StateVertex(symbolVertex, nextStateNoInteger.intValue(), context.generation);
                            context.states.put(nextStateNoInteger, stateVertex);
                            stateVertex.whiteTokenHolder = context.whiteTokenHolder;
                            #if (environment.metadebug) {
                                java.lang.System.err.println("addLink: link " + turningPoints + " <- " + symbolVertex + "(new) <- " + stateVertex + "(new)");
                            #}
                            context.unbrokenStates.add(stateVertex);
                        } else {
                            $assert stateVertex.whiteTokenHolder == context.whiteTokenHolder;
                            stateVertex.addSuccessor(symbolVertex);
                            #if (environment.metadebug) {
                                java.lang.System.err.println("addLink: link " + turningPoints + " <- " + symbolVertex + "(new) <- " + stateVertex + "(already exists)");
                            #}
                            if (!context.unbrokenStates.contains(stateVertex))
                                $breakState(context, symbolVertex, stateVertex);
                        }
                    }
                    #if (environment.metadebug) {
                        java.lang.System.err.println("addLink: end");
                    #}
                }
                
                #allocate("LabeledNode", topTypeNamespace);
                private static class $LabeledNode {
                    public $LabelInt labels;
                    public $Node node;
                    public $LabeledNode($LabelInt labels, $Node node) {
                        this.labels = labels;
                        this.node = node;
                    }
                }
                
                private $Node $createNode($Context context, $ConcreteSyntaxTree cst) throws $ParseException {
                    return $createNode(context, null, cst, null);
                }
                private $Node $createNode($Context context, $Link headWhiteToken, $ConcreteSyntaxTree cst, $Link tailWhiteToken) throws $ParseException {
                    int[] reduction = context.parser.$reductionTable[cst.reductionID];
                    $assert reduction[0] < >>[[ tables.superiorIndexOfTypeNonterminals() ]]<<;
                    
                    $Link$<$Link<$LabeledNode>> nodeLists = new $Link$<$Link<$LabeledNode>>(null, null);
                    #if (whiteTokensMostLeft) {
                        nodeLists = $flatWhiteTokens(headWhiteToken, nodeLists);
                        nodeLists = $flatNodes(context, $LABEL_ZERO, cst, nodeLists);
                        if (headWhiteToken != tailWhiteToken)
                            nodeLists = $flatWhiteTokens(tailWhiteToken, nodeLists);
                    #} else  {
                        if (headWhiteToken != tailWhiteToken)
                            nodeLists = $flatWhiteTokens(headWhiteToken, nodeLists);
                        nodeLists = $flatNodes(context, $LABEL_ZERO, cst, nodeLists);
                        nodeLists = $flatWhiteTokens(tailWhiteToken, nodeLists);
                    #}
                    
                    if (nodeLists.next == null) {
                        return $createNode(context, reduction[0], ($Link$<$LabeledNode>) nodeLists.value);
                    } else {
                        #if (!ambiguitySupport) {
                            throw new $AssertionError();
                        #} else {
                            java.util.List$<$Node> nodes = new java.util.LinkedList$<$Node>();
                            do {
                                nodes.add($createNode(context, reduction[0], ($Link$<$LabeledNode>) nodeLists.value));
                                nodeLists = nodeLists.next;
                            } while (nodeLists != null);
                            $Parser.$AmbiguityAbsorber result = $createAmbiguityAbsorber(nodes);
                            if (result == null)
                                $throwAmbiguousGrammerError(reduction[0], nodes);
                            return result;
                        #}
                    }
                }
                private $Node $createNode($Context context, int symbolID, $Link$<$LabeledNode> nodeList) throws $ParseException {
                    // TODO: $createNode(symbolID, parameters) をダイレクトに nodeList を受け取るように修正する。
                    
                    java.util.List$<$LabelInteger> labelBitsList = new java.util.LinkedList$<$LabelInteger>();
                    java.util.List$<$Node> nodes = new java.util.LinkedList$<$Node>();
                    for (; nodeList != null; nodeList = nodeList.next) {
                        $LabeledNode n = ($LabeledNode) nodeList.value;
                        #if (labelWordCount == 1) {
                            labelBitsList.add(0, new java.lang.Integer(n.labels));
                        #} else {
                            labelBitsList.add(0, n.labels);
                        #}
                        nodes.add(0, n.node);
                    }
                    $Node[] childNodes = ($Node[]) nodes.toArray(new $Node[nodes.size()]);
                    context.parameters.childNodes = childNodes;
                    context.parameters.labelBitsList = labelBitsList;
                    $Node result = $createNode(symbolID, context.parameters);
                    context.parameters.childNodes = null;
                    context.parameters.labelBitsList = null;
                    return result;
                }
                
                private $Link$<$Link<$LabeledNode>> $flatNodes($Context context, $LabelInt labels, $ConcreteSyntaxTree cst) throws $ParseException {
                    return $flatNodes(context, labels, cst, new $Link$<$Link<$LabeledNode>>(null, null));
                }
                private $Link$<$Link<$LabeledNode>> $flatNodes($Context context, $LabelInt labels, $ConcreteSyntaxTree cst, $Link$<$Link<$LabeledNode>> nodeLists) throws $ParseException {
                tailRecursion:
                    for (;;) {
                        if (cst.children == null)
                            return nodeLists;
                        
                        int[] reduction = context.parser.$reductionTable[cst.reductionID];
                        #if (whiteTokensMostLeft) {
                            java.lang.Object tailWhiteToken = cst.children[cst.children.length - 1];
                        #} else {
                            java.lang.Object headWhiteToken = cst.children[0];
                        #}
                        int reductionIndex = 1;
                        
                        for (int index = 1; index < cst.children.length; index += 2) {
                            $Link whiteToken1 = ($Link) cst.children[index - 1];
                            $Link whiteToken2 = ($Link) cst.children[index + 1];
                            java.lang.Object symbol = cst.children[index];
                            $Link symbols = null;
                            if (symbol instanceof $Link) {
                                symbols = ($Link) symbol;
                                if (symbols.next == null) {
                                    symbol = symbols.value;
                                    symbols = null;
                                }
                            }
                            
                            #if (!whiteTokensMostLeft) {
                                if (whiteToken1 != whiteToken2 && whiteToken1 != headWhiteToken)
                                    nodeLists = $flatWhiteTokens(whiteToken1, nodeLists);
                            #}
                            
                            #if (labelWordCount == 1) {
                                int mergedlabelBits = reduction[reductionIndex++];
                                if ((mergedlabelBits & 1) != 0) {
                                    mergedlabelBits &= ~1;
                                    mergedlabelBits |= labels;
                                }
                            #} else {
                                final int labelWordCount = >>[[ labelWordCount ]]<<;
                                java.math.BigInteger mergedlabelBits = java.math.BigInteger.ZERO;
                                for (int k = labelWordCount - 1; k >= 0; k--)
                                    mergedlabelBits = mergedlabelBits.or(java.math.BigInteger.valueOf(0xFFFFFFFFL & reduction[reductionIndex++]).shiftLeft(k * 32));
                                if (mergedlabelBits.testBit(0)) {
                                    mergedlabelBits = mergedlabelBits.clearBit(0);
                                    mergedlabelBits = mergedlabelBits.or(labels);
                                }
                            #}
                            
                            $Node node;
                            if (symbols != null) {
                                #if (!ambiguitySupport) {
                                    throw new $AssertionError();
                                #} else {
                                    java.util.List$<$Link<$LabeledNode>> childLinkList = new java.util.LinkedList$<$Link<$LabeledNode>>();
                                    
                                    java.util.List$<$Node> ambiguousNodes = new java.util.LinkedList$<$Node>();
                                    int childSymbolID = -1;
                                    $LabelInteger childLabels = null;
                                    {
                                        $Link it = symbols;
                                        do {
                                            $ConcreteSyntaxTree subtree = ($ConcreteSyntaxTree) it.value;
                                            int[] childReduction = context.parser.$reductionTable[subtree.reductionID];
                                            $assert childSymbolID == -1 || childReduction[0] == childSymbolID;
                                            childSymbolID = childReduction[0];
                                            $Link$<$Link<$LabeledNode>> childLinks = $flatNodes(context, mergedlabelBits, subtree);
                                            
                                            $Link$<$Link<$LabeledNode>> it2 = childLinks;
                                            do {
                                                $Link$<$LabeledNode> children = ($Link$<$LabeledNode>) it2.value;
                                                childLinkList.add(children);
                                                
                                                if (children != null && children.next == null) {
                                                        // has only one element.
                                                    $LabeledNode n = ($LabeledNode) children.value;
                                                    if (childLabels == null) {
                                                        #if (labelWordCount == 1) {
                                                            childLabels = new java.lang.Integer(n.labels);
                                                        #} else {
                                                            childLabels = n.labels;
                                                        #}
                                                        if (ambiguousNodes != null)
                                                            ambiguousNodes.add(n.node);
                                                        continue;
                                                    } else {
                                                        #if (labelWordCount == 1) {
                                                            if (childLabels.intValue() == n.labels) {
                                                                if (ambiguousNodes != null)
                                                                    ambiguousNodes.add(n.node);
                                                                continue;
                                                            }
                                                        #} else {
                                                            if (childLabels.equals(n.labels)) {
                                                                if (ambiguousNodes != null)
                                                                    ambiguousNodes.add(n.node);
                                                                continue;
                                                            }
                                                        #}
                                                    }
                                                }
                                                ambiguousNodes = null;
                                            } while ((it2 = it2.next) != null);
                                            it = it.next;
                                        } while (it != null);
                                    }
                                    
                                    if (ambiguousNodes != null) {
                                        node = $createAmbiguityAbsorber(ambiguousNodes);
                                        if (node == null)
                                            $throwAmbiguousGrammerError(childSymbolID, ambiguousNodes);
                                        
                                        #if (labelWordCount == 1) {
                                            mergedlabelBits = childLabels.intValue();
                                        #} else {
                                            mergedlabelBits = childLabels;
                                        #}
                                    } else {
                                        $Link$<$Link<$LabeledNode>> newNodeLists = null;
                                        
                                        $Link$<$Link<$LabeledNode>> it = nodeLists;
                                        do {
                                            $Link$<$LabeledNode> tail = ($Link$<$LabeledNode>) it.value;
                                            
                                            java.util.Iterator$<$Link<$LabeledNode>> it37 = childLinkList.iterator();
                                            while (it37.hasNext()) {
                                                $Link$<$LabeledNode> mid = ($Link$<$LabeledNode>) it37.next();
                                                
                                                $Link$<$LabeledNode> dummy = new $Link$<$LabeledNode>(null, null);
                                                $Link$<$LabeledNode> cursor = dummy;
                                                
                                                for (; mid != null; mid = mid.next) {
                                                    cursor = cursor.next = new $Link$<$LabeledNode>(mid.value, null);
                                                }
                                                cursor.next = tail;
                                                
                                                newNodeLists = new $Link$<$Link<$LabeledNode>>(dummy.next, newNodeLists);
                                            }
                                            it = it.next;
                                        } while (it != null);
                                        
                                        nodeLists = newNodeLists;
                                        node = null;
                                    }
                                #}
                            } else {
                                if (symbol instanceof $ConcreteSyntaxTree) {
                                    $ConcreteSyntaxTree subtree = ($ConcreteSyntaxTree) symbol;
                                    node = subtree.createNode(context);
                                    if (node == null) {
                                        if (index + 2 < cst.children.length) {
                                            nodeLists = $flatNodes(context, mergedlabelBits, subtree, nodeLists);
                                        } else {
                                            // the last of the loop.
                                            $assert node == null;
                                            $assert whiteToken2 == tailWhiteToken;
                                            labels = mergedlabelBits;
                                            cst = subtree;
                                            continue tailRecursion;
                                        }
                                    }
                                } else {
                                    node = ($Node) symbol;
                                }
                            }
                            if (node != null) {
                                $LabeledNode n = new $LabeledNode(mergedlabelBits, node);
                                
                                $Link$<$Link<$LabeledNode>> it = nodeLists;
                                do {
                                    $Link$<$LabeledNode> tail = ($Link$<$LabeledNode>) it.value;
                                    it.value = new $Link$<$LabeledNode>(n, tail);
                                    it = it.next;
                                } while (it != null);
                            }
                            
                            #if (whiteTokensMostLeft) {
                                if (whiteToken2 != whiteToken1 && whiteToken2 != tailWhiteToken)
                                    nodeLists = $flatWhiteTokens(whiteToken2, nodeLists);
                            #}
                        }
                        return nodeLists;
                    }
                }
                private $Link$<$Link<$LabeledNode>> $flatWhiteTokens($Link whiteTokens, $Link$<$Link<$LabeledNode>> nodeLists) {
                    $Link link = whiteTokens;
                    for (; link != null; link = link.next) {
                        $LabeledNode n = new $LabeledNode($LABEL_ZERO, ($Token) link.value);
                        
                        $Link$<$Link<$LabeledNode>> it = nodeLists;
                        do {
                            $Link$<$LabeledNode> tail = ($Link$<$LabeledNode>) it.value;
                            it.value = new $Link$<$LabeledNode>(n, tail);
                            it = it.next;
                        } while (it != null);
                    }
                    return nodeLists;
                }
                
                #if (environment.debug) {
                    #allocate("debugPrint", topMethodNamespace);
                    private void $debugPrint($Context context, java.lang.String indent, java.lang.Object obj) {
                        if (obj instanceof $Node) {
                            $debugPrintNode(context, indent, ($Node) obj);
                        } else if (obj instanceof $ConcreteSyntaxTree) {
                            $debugPrintCST(context, indent, ($ConcreteSyntaxTree) obj);
                        } else if (obj instanceof $SymbolVertex) {
                            $SymbolVertex vertex = ($SymbolVertex) obj;
                            $debugPrint(context, indent, vertex.symbol);
                        } else {
                            $debugPrintAbsorber(context, indent, ($ObjectLink) obj);
                        }
                    }
                    private void $debugPrintNode($Context context, java.lang.String indent, $Node node) {
                        java.lang.System.err.print(indent);
                        if (node instanceof $Token) {
                            java.lang.System.err.println(node);
                        } else {
                            java.lang.System.err.println(node.getClass().getName());
                            indent = indent + "  ";
                            java.util.Iterator it20 = node.getChildNodes().iterator();
                            while (it20.hasNext()) {
                                $Node child = ($Node) it20.next();
                                $debugPrintNode(context, indent, child);
                            }
                        }
                    }
                    private void $debugPrintCST($Context context, java.lang.String indent, $ConcreteSyntaxTree cst) {
                        java.lang.String[] nodeNames = {
                            #{
                                #Symbol[] symbols = tables.symbols();
                                #for (int i = 0; i < tables.superiorIndexOfAliasNonterminals(); i++) {
                                    #Symbol symbol = symbols[i];
                                    >>[[ quoted(symbol.toString()) ]]<<,
                                #}
                            #}
                        };
                        
                        int[] reduction = context.parser.$reductionTable[cst.reductionID];
                        final int superiorIndexOfAliasNonterminals = >>[[ tables.superiorIndexOfAliasNonterminals() ]]<<;
                        if (reduction[0] < superiorIndexOfAliasNonterminals) {
                            java.lang.System.err.print(indent);
                            java.lang.System.err.println(nodeNames[reduction[0]]);
                            indent = indent + "  ";
                        }
                        if (cst.children != null) {
                            for (int i = 1; i < cst.children.length; i += 2) {
                                java.lang.Object child = cst.children[i];
                                $debugPrint(context, indent, child);
                            }
                        }
                    }
                    private void $debugPrintAbsorber($Context context, java.lang.String indent, $Link$<java.lang.Object> nodes) {
                        if (nodes.next == null) {
                            $debugPrint(context, indent, nodes.value);
                        } else {
                            java.lang.System.err.print(indent);
                            java.lang.System.err.println("an ambiguity absorber");
                            indent = indent + "  ";
                            
                            for (; nodes != null; nodes = nodes.next) {
                                java.lang.Object node = nodes.value;
                                $debugPrint(context, indent, node);
                            }
                        }
                    }
                #} // debug
            #} // parse routines.
            
            
            #if (separateOutput) {
                #changeStoreChannel(1);
                #if (packageName != null)
                    #println("package $packageName;");
                #println("class " + topLevelClassName + "Tables {");
            #}
            
            #allocate("TABLE_SENTINEL", topFieldNamespace);
            >>[[ separateOutput ? "/*private*/" : "private" ]]<< final static int $TABLE_SENTINEL = java.lang.Integer.MAX_VALUE;
            
            #{
                #// DFA table
                #final int unit = 50;
                #DFAState[] states = tables.dfaStates();
                #logger.verbose(environment.processingFile, "DFA table: {0} entries.", new Integer(states.length));
                >>[[ separateOutput ? "/*private*/" : "private" ]]<< static $final_avoid_jdk12_bug int[][] $dfaTable;
                #allocate("dfaTable", topFieldNamespace);
                static {
                    $dfaTable = new int[>>[[ states.length ]]<<][];
                    #for (int i = 0; i < states.length; i += unit) {
                        initializeDFATable>>[[ i ]]<<();
                    #}
                }
                #for (int i = 0; i < states.length; i += unit) {
                    #topMethodNamespace.add("initializeDFATable" + i, RESERVED_NAME);
                    private static void initializeDFATable>>[[ i ]]<<() {
                        #for (int j = i; j < i + unit && j < states.length; j++) {
                            #DFAState state = states[j];
                            $dfaTable[>>[[ j ]]<<] = new int[] {
                                #Terminal symbol = state.hitSymbol();
                                #if (symbol == null)
                                    -1,
                                #else
                                    >>[[ symbolToJavaWord.get(symbol) ]]<<,
                                #Iterator it = state.getMapInputToNextStates().entrySet().iterator();
                                #while (it.hasNext()) {
                                    #Map.Entry entry = (Map.Entry) it.next();
                                    #DFAInput input = (DFAInput) entry.getKey();
                                    #DFAState nextState = (DFAState) entry.getValue();
                                    >>[[ quoted(input.lower) ]]<<, >>[[ quoted(input.upper) ]]<<, >>[[ tables.getIndexOfState(nextState) ]]<<, 
                                #}
                            };
                        #}
                    }
                #}
            #}
            
            #{
                #LinkedHashMap tableCache = new LinkedHashMap(16, 0.75f, true) {
                    #protected boolean removeEldestEntry(Map.Entry eldest) {
                        #return size() > 300;
                    #}
                #};
                #// LALR(1) table
                #final int unit = 20;
                #LRState[] states = tables.lrStates();
                #logger.verbose(environment.processingFile, "LR table: {0} entries.", new Integer(states.length));
                #int tableSize = 0;
                >>[[ separateOutput ? "/*private*/" : "private" ]]<< static $final_avoid_jdk12_bug int[][] $lrTable;
                #allocate("lrTable", topFieldNamespace);
                static {
                    $lrTable = new int[>>[[ states.length ]]<< * 3][];
                    #for (int i = 0; i < states.length; i += unit) {
                        initializeLRTable>>[[ i ]]<<();
                    #}
                }
                #for (int i = 0; i < states.length; i += unit) {
                    #topMethodNamespace.add("initializeLRTable" + i, RESERVED_NAME);
                    private static void initializeLRTable>>[[ i ]]<<() {
                        #for (int j = i; j < i + unit && j < states.length; j++) {
                            #LRState state = states[j];
                            // State >>[[ tables.getIndexOfState(state) ]]<<  (>>[[ state ]]<<)
                            #Integer index;
                            #
                            #Map teminalToNextStates = tables.getMapTeminalToNextStates(state);
                            #tableSize += 4;    // address
                            #index = (Integer) tableCache.get(teminalToNextStates);
                            #if (index == null) {
                                #tableCache.put(teminalToNextStates, new Integer(j * 3 + 0));
                                $lrTable[>>[[ j ]]<< * 3 + 0] = new int[] {
                                    #Iterator it3 = teminalToNextStates.entrySet().iterator();
                                    #while (it3.hasNext()) {
                                        #Map.Entry entry = (Map.Entry) it3.next();
                                        #Symbol symbol = (Symbol) entry.getKey();
                                        #Collection nextStates = (Collection) entry.getValue();
                                        #if (!multipleShiftEntry) {
                                            #assert nextStates.size() == 1;
                                        #}
                                        #Iterator it27 = nextStates.iterator();
                                        #while (it27.hasNext()) {
                                            #LRState nextState = (LRState) it27.next();
                                            >>[[ symbolToJavaWord.get(symbol) ]]<<, >>[[ tables.getIndexOfState(nextState) ]]<<,
                                            #tableSize += 4 + 4;
                                        #}
                                    #}
                                    $TABLE_SENTINEL,
                                    #tableSize += 4;
                                };
                            #} else {
                                $lrTable[>>[[ j ]]<< * 3 + 0] = $lrTable[>>[[ index.intValue() / 3 ]]<< * 3 + >>[[ index.intValue() % 3 ]]<<];
                            #}
                            #
                            #Map nonteminalToNextStates = tables.getMapNonteminalToNextStates(state);
                            #tableSize += 4;    // address
                            #index = (Integer) tableCache.get(nonteminalToNextStates);
                            #if (index == null) {
                                #tableCache.put(nonteminalToNextStates, new Integer(j * 3 + 1));
                                $lrTable[>>[[ j ]]<< * 3 + 1] = new int[] {
                                    #Iterator it4 = nonteminalToNextStates.entrySet().iterator();
                                    #while (it4.hasNext()) {
                                        #Map.Entry entry = (Map.Entry) it4.next();
                                        #Symbol symbol = (Symbol) entry.getKey();
                                        #Collection nextStates = (Collection) entry.getValue();
                                        #if (!multipleShiftEntry) {
                                            #assert nextStates.size() == 1;
                                        #}
                                        #Iterator it27 = nextStates.iterator();
                                        #while (it27.hasNext()) {
                                            #LRState nextState = (LRState) it27.next();
                                            >>[[ symbolToJavaWord.get(symbol) ]]<<, >>[[ tables.getIndexOfState(nextState) ]]<<,
                                            #tableSize += 4 + 4;
                                        #}
                                    #}
                                    $TABLE_SENTINEL,
                                    #tableSize += 4;
                                };
                            #} else {
                                $lrTable[>>[[ j ]]<< * 3 + 1] = $lrTable[>>[[ index.intValue() / 3 ]]<< * 3 + >>[[ index.intValue() % 3 ]]<<];
                            #}
                            #
                            #Map symbolToReductions = tables.getMapSymbolToReduction(state);
                            #tableSize += 4;    // address
                            #index = (Integer) tableCache.get(symbolToReductions);
                            #if (index == null) {
                                #tableCache.put(symbolToReductions, new Integer(j * 3 + 2));
                                $lrTable[>>[[ j ]]<< * 3 + 2] = new int[] {
                                    #Iterator it5 = symbolToReductions.entrySet().iterator();
                                    #while (it5.hasNext()) {
                                        #Map.Entry entry = (Map.Entry) it5.next();
                                        #Symbol symbol = (Symbol) entry.getKey();
                                        #Set reductions = (Set) entry.getValue();
                                        #Iterator it6 = reductions.iterator();
                                        #while (it6.hasNext()) {
                                            #Reduction reduction = (Reduction) it6.next();
                                            >>[[ symbolToJavaWord.get(symbol) ]]<<, >>[[ tables.getIndexOfReduction(reduction) ]]<<,
                                            #tableSize += 4 + 4;
                                        #}
                                    #}
                                    $TABLE_SENTINEL,
                                    #tableSize += 4;
                                };
                            #} else {
                                $lrTable[>>[[ j ]]<< * 3 + 2] = $lrTable[>>[[ index.intValue() / 3 ]]<< * 3 + >>[[ index.intValue() % 3 ]]<<];
                            #}
                        #}
                    }
                #}
                #logger.verbose(environment.processingFile, "LR table: {0} bytes.", new Integer(tableSize));
            #}
            
            #{
                #// reduction table
                #Reduction[] reductions = tables.reductions();
                #logger.verbose(environment.processingFile, "{0} reductions.", new Integer(reductions.length));
                #allocate("reductionTable", topFieldNamespace);
                >>[[ separateOutput ? "/*private*/" : "private" ]]<< static final int[][] $reductionTable = {
                    #for (int i = 0; i < reductions.length; i++) {
                        #Reduction reduction = reductions[i];
                        #java.lang.StringBuffer result = new java.lang.StringBuffer();
                        #Iterator it = reduction.labelSetList().iterator();
                        #while (it.hasNext()) {
                            #Set labels = (Set) it.next();
                            #BigInteger no = BigInteger.ZERO;
                            #Iterator innest = labels.iterator();
                            #while (innest.hasNext()) {
                                #java.lang.String label = (String) innest.next();
                                #no = no.setBit(tables.getIndexOfLabel(label));
                            #}
                            #String hex = no.toString(16).toUpperCase(Locale.ENGLISH);
                            #while (hex.length() < labelWordCount * 8) {
                                #hex = '0' + hex;
                            #}
                            #for (int j = 0; j < labelWordCount; j++) {
                                #result.append("0x").append(hex.substring(j * 8, j * 8 + 8)).append(",");
                            #}
                            #result.append(" ");
                        #}
                        /*>>[[ i ]]<<*/{ >>[[ symbolToJavaWord.get(reduction.reducedSymbol()) ]]<<,  >>[[ result ]]<<},
                    #}
                };
            #}
            
            #if (separateOutput) {
                #println("}");
                #changeStoreChannel(0);
            #}
            
            #if (environment.metadebugLight) {
                #allocate("symbolIDToName", topFieldNamespace);
                private static java.lang.String[] $symbolIDToName = {
                    #for (int i = 0; i < tables.symbols().length; i++) {
                        >>[[ quoted(tables.symbols()[i].toString()) ]]<<,
                    #}
                };
                #allocate("lrStateIDToName", topFieldNamespace);
                private static java.lang.String[] $lrStateIDToName = {
                    #for (int i = 0; i < tables.lrStates().length; i++) {
                        >>[[ quoted(tables.lrStates()[i].toString()) ]]<<,
                    #}
                };
            #}
            
            #allocate("replaceAll", topMethodNamespace);
            private static $<T> void $replaceAll(java.util.List$<T> list, $Node oldValue, >>[[ targetVersion < 1.5 ? "$Node" : "T" ]]<< newValue) {
                #if (!ambiguitySupport && 1.4 <= targetVersion && targetVersion < 1.5) {
                    java.util.Collections.replaceAll(list, oldValue, newValue);
                #} else {
                    java.util.ListIterator$<T> it = list.listIterator();
                    if (oldValue == null) {
                        while (it.hasNext()) {
                            java.lang.Object old = it.next();
                            if (oldValue == old)
                                it.set(newValue);
                        }
                    } else {
                        while (it.hasNext()) {
                            java.lang.Object old = it.next();
                            #if (ambiguitySupport) {
                                if (oldValue instanceof $Parser.$AmbiguityAbsorber && oldValue.equals((($Node) old).$getWrappedAmbiguityAbsorber()))
                                    it.set(newValue);
                                else
                            #}
                            if (oldValue.equals(old))
                                it.set(newValue);
                        }
                    }
                #}
            }
            #if (arrayRatherThanList) {
                private static void $replaceAll(java.lang.Object[] array, $Node oldValue, $Node newValue) {
                    if (oldValue == null) {
                        for (int i = 0; i < array.length; i++) {
                            if (array[i] == null)
                                array[i] = newValue;
                        }
                    } else {
                        for (int i = 0; i < array.length; i++) {
                            #if (ambiguitySupport) {
                                if (oldValue instanceof $Parser.$AmbiguityAbsorber && oldValue.equals((($Node) array[i]).$getWrappedAmbiguityAbsorber()))
                                    array[i] = newValue;
                                else
                            #}
                            if (oldValue.equals(array[i]))
                                array[i] = newValue;
                        }
                    }
                }
            #}
            
            #allocate("toHexCode", topMethodNamespace);
            private static java.lang.String $toHexCode(char ch) {
                java.lang.String hex = "000" + java.lang.Integer.toString(ch, 16).toUpperCase(java.util.Locale.ENGLISH);
                hex = hex.substring(hex.length() - 4);
                return hex;
            }
            
            /*
            #allocate("quoted", topMethodNamespace);
            private static String $quoted(char ch) {
                StringBuffer quoted = new StringBuffer("\'");
                $quoted(quoted, ch);
                quoted.append("\'");
                return quoted.toString();
            }
            private static String $quoted(String str) {
                return $quoted(str, true);
            }
            private static String $quoted(String str, boolean quote) {
                StringBuffer quoted = new StringBuffer();
                if (quote)
                    quoted.append("\"");
                for (int i = 0; i < str.length(); i++) {
                    char ch = str.charAt(i);
                    $quoted(quoted, ch);
                }
                if (quote)
                    quoted.append("\"");
                return quoted.toString();
            }
            private static void $quoted(StringBuffer result, char ch) {
                switch (ch) {
                default:
                    if (0x0020 <= ch && ch <= 0x007E) {
                        result.append(ch);
                    } else {
                        result.append("\\u").append($toHexCode(ch));
                    }
                    break;
                case '\n':   result.append("\\n");      break;
                case '\t':   result.append("\\t");      break;
                case '\b':   result.append("\\b");      break;
                case '\r':   result.append("\\r");      break;
                case '\f':   result.append("\\f");      break;
                case '\\':   result.append("\\\\");     break;
                case '\'':   result.append("\\\'");     break;
                case '\"':   result.append("\\\"");     break;
                }
            }
            */
            
            #if (!root.hasProtectedConstructor() && publicParsableTypeCount == 1) {
                #if (manualLocale.equals("ja")) {
                    /**
                    #if (javadocForManual) {
                         * 引数に渡されたファイルをパーズし、構文木を出力します（サンプルプログラム）。
                         * <p>
                         * このメソッドは、<code>$parsable</code> な非終端記号が１つ存在する場合にだけ出力されます。このメソッドの仕様は、将来のバージョンでは変更になる可能性があります。 
                    #} else {
                         * 引数に渡されたファイルをパーズし結果を出力します。
                    #}
                     * 
                     * @param   args
                     *      解析対象のファイルの名前
                    #if (javadocForManual) {
                        * @since    notavaCC 1.0
                    #}
                     */
                #} else {
                    /**
                    #if (javadocForManual) {
                         * Parses the given files and prints the syntax tree (a sample program).
                         * <p>
                         * This method is generated if there is only one <code>$parsable</code> nonterminal symbol.  The specification of this method may be changed in the feature.
                    #} else {
                         * Parses the given files and prints the syntax tree.
                    #}
                     * 
                     * @param   args
                     *      the names of the files to parse.
                    #if (javadocForManual) {
                        * @since    notavaCC 1.0
                    #}
                     */
                #}
                #reserve("main", topMethodNamespace);
                public static void $main(java.lang.String[] args) {
                    try {
                        $Parser p = new $Parser();
                        for (int i = 0; i < args.length; i++) {
                            java.io.File file = new java.io.File(args[i]);
                            java.lang.System.out.println(file);
                            
                            #Iterator it12 = tables.userDefinedNodeTypes().iterator();
                            #while (it12.hasNext()) {
                                #UserDefinedNodeType type = (UserDefinedNodeType) it12.next();
                                #LRState initialState = (LRState) lrTable.symbolToInitialState().get(type.getTypeNonterminal());
                                #if (initialState != null) {
                                    java.lang.System.out.println(p.parse>>[[ type.getName() ]]<<(file, null));
                                    #break;
                                #}
                            #}
                        }
                    } catch(java.lang.Exception x) {
                        x.printStackTrace();
                    }
                }
            #}
        } // class $Parser
>>[[
        
        logger.verbosest(environment.processingFile, "checking topTypeNamespace");
        checkDuplex(topTypeNamespace);
        logger.verbosest(environment.processingFile, "checking topFieldNamespace");
        checkDuplex(topFieldNamespace);
        logger.verbosest(environment.processingFile, "checking topMethodNamespace");
        checkDuplex(topMethodNamespace);
        logger.verbosest(environment.processingFile, "checking nodeFieldNamespace");
        checkDuplex(nodeFieldNamespace);
        logger.verbosest(environment.processingFile, "checking nodeNoArgumentMethodNamespace");
        checkDuplex(nodeNoArgumentMethodNamespace);
        logger.verbosest(environment.processingFile, "checking nodeMethodNamespace");
        checkDuplex(nodeMethodNamespace);
        if (logger.hasError())
            logger.fatal();
    } // generateCode
    
    private String getFieldAccessString(UserDefinedNodeType type, String name) {
        Type superType = (Type) type.getDirectSuperTypes().iterator().next();
        if (superType.getField(name) != null) {
            return getFieldAccessStringSuper((UserDefinedNodeType) superType, name);
        } else {
            return "this." + name;
        }
    }
    private String getFieldAccessStringSuper(UserDefinedNodeType type, String name) {
        Type superType = (Type) type.getDirectSuperTypes().iterator().next();
        if (superType.getField(name) != null) {
            return getFieldAccessStringSuper((UserDefinedNodeType) superType, name);
        } else {
            return "(($Parser.$Default." + type.getName() + ") this)." + name;
        }
    }
    
    public void checkDuplex(MultiMap namespace) {
        Iterator it13 = namespace.entrySet().iterator();
        while (it13.hasNext()) {
            Map.Entry entry = (Map.Entry) it13.next();
            String name = (String) entry.getKey();
            Set reservers = (Set) entry.getValue();
            if (reservers.contains(RESERVED_NAME))
                logger.verbosest(environment.processingFile, "checking the reserved name {0}.", name);
            if (reservers.size() >= 2) {
                Iterator it14 = reservers.iterator();
                while (it14.hasNext()) {
                    Object reserver = (Object) it14.next();
                    if (reserver.equals(RESERVED_NAME)) {
                        // ignore.
                    } else if (reserver instanceof Symbol) {
                        logger.error(environment.processingFile, "The reserved name {0} is used.", reserver);
                    } else if (reserver instanceof Type) {
                        logger.error(environment.processingFile, "The reserved name {0} is used.", reserver);
                    } else {
                        logger.error(environment.processingFile, "The reserved name {0} is used.", reserver);
                    }
                }
            }
        }
    }
    
} // JavaCodeGenerator

